! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!*****************************************************************************
!
! Various routine for PSF photometry
!
!****************************************************************************
module ompsfphotometry_mod
 
   use param
   use errorHandling
 
   use omutility_mod      ! In mssllib various fits access routines
   use omconstants_mod    ! In mssllib defines various constants
   use omsourceparam_mod  ! In mssllib defines Struct SourceParameters
   use OmCoinccorr_mod       ! For coincidence-loss corrections   
   use omparameters_mod 
   use omphotometry_mod   ! In mssllib for doing aperture photometry
   use leastsquares_mod
   use optimize_mod


   implicit none


  

!*******************************************************************************
!
! The following are global variables
! 
!*******************************************************************************
    real(single), private, pointer :: mypsfImage(:, :) => null()

    integer, private :: imin, imax, jmin, jmax

    real(double), private , pointer, dimension(:, :) :: xdata2 => null()
    real(double), private , pointer, dimension(:, :) :: xdata => null()
    real(double), private,  pointer, dimension(:) :: ydata => null(),  &
           weight => null()

    integer(int16), private :: l0=0, diFit
    

    real(single), private :: psfxCentre, psfYcentre, sumPsfImage
   

    real(double), private :: bl(50), bu(50)
    integer(int8), private, parameter :: maxShift = 1

    integer, private :: psfNx, psfny
    integer(int8), private :: numberToFit, weightingMethod = 1
    real(single), private, pointer :: myImage(:, :) => null()
    real(single), private :: myApertureRadius, maxRadius = 0.0
    integer(int16), private, parameter :: maxNdata = 5000
    integer(int8), private :: maxNumNeighbours = 9
    real(single), private :: maxRawCountRate = 50.0
    real(double), private :: computedBackground = 0.0, std = 0.0_double
    type(gridinterpolant), save :: psfInterpolant
  
 
    integer(int8) :: maxIter = 3_int8 
    real(single), private :: time, currentTime
    real(single), private, parameter :: maxTime = 5.0
    logical, private :: displayResults  = .true.
contains

!********************************************************
!
!
!
!*******************************************************
subroutine setMaxIter( num )
    integer(int8), intent(in) :: num

    maxIter = num

end subroutine

!********************************************************
!
!
!
!*******************************************************
subroutine setMaxNumNeighbours( num )
    integer, intent(in) :: num

    maxNumNeighbours = num

end subroutine setMaxNumNeighbours

!********************************************************
!
!
!
!*******************************************************
subroutine setMaxRawCountRate( rate )
    real(single), intent(in) :: rate

    maxRawCountRate = rate

end subroutine setMaxRawCountRate



!********************************************************
!
!
!
!*******************************************************
subroutine setweightingmethod( method )
    integer, intent(in) :: method

    weightingMethod = method

end subroutine setweightingmethod


!*****************************************************************
!
! Loads the image and creates the PSF interpolant using the CAL
! PSF image
!
!*****************************************************************
subroutine initialisePSFPhotometry(image, maxRawCountRateIn)
    real(single), pointer :: image(:, :)
    real(single), intent(in), optional :: maxRawCountRateIn 
  

    if(present(maxRawCountRateIn)) then

        call setMaxRawCountRate( maxRawCountRateIn )

    end if

  

    
    myImage => image



    myApertureRadius = 12.0


    !**************************************************
    ! Create the psf image
    !
    !*************************************************
    call CreatethePsfImage
    
end subroutine initialisePSFPhotometry




!**************************************************************************
!
!
!
!
!**************************************************************************
subroutine CreatePSFImage21test(psfNx, psfNy, psfimageOut, xPos, yPos, &
                          ratio, bin, res,totalCounts,           &
                          psfGridFitOut, psfsplinefit, bilinearData)


    real (single), optional, pointer :: psfImageOut(:, :) ! Pointer to the psf image

    real(double), intent(in) :: xPos, yPos, ratio
    integer (int32), intent(inout) :: psfnx, psfny
    integer (int32), intent(in) :: bin
    integer, optional, intent(in) :: res
    real(single), intent(inout), optional :: totalCounts
    real(double) :: x1, y1    
    type(GridInterpolant), optional, intent(out) :: psfGridFitOut
   
    type(TwoDSpline), optional, intent(out) :: psfSplineFit    
    type(BilinearStruct), optional, intent(inout) :: bilinearData 

    real(single), pointer :: x(:) => null(), y(:) => null(), height(:, :) => null() 
    real (single), pointer :: psfImage(:, :) => null() ! Pointer to the psf image


    real(double) :: xCentre, yCentre, f, f1, f2, f3, f4, fCentre
    integer :: i, j, m
    integer :: xRes, yRes, nx, ny  
    type(GridInterpolant) :: psfGridFit1

    real(single), dimension(:, :), pointer :: psf => null() ! To hold the point-spread function

    type(SourceParameters) :: source
  
 
    !******************************************************
    ! Create the PSF  -image map
    !*****************************************************

    if (present(res)) then
        xres = res
        yres = res
    else
        xres = 1
        yres = 1
    end if


    !*******************************************
    !
    ! Get the point-spread function from cal
    !
    !*******************************************             
    call CAL_omGetPSFmap(xpos, ypos, &
                psfNx, psfNx, psfNy, psfNy,  &
                xres, yres, dble(ratio), psf)
 


  

    !*************************************
    ! Normalize so that central peak is 1
    !*************************************
    f = maxval(psf)    
    psf = psf / f
  
          
    psfnx = size(psf, 1)
    psfny = size(psf, 2)


    ! Construct a grid-interpolant

    m = psfnx * psfny
    x => allocateSingle(1, psfnx, " x")
    y => allocateSingle(1, psfny, " y")

    xCentre = dble(psfnx) / 2.0_double + 0.5_double
    yCentre = dble(psfny) / 2.0_double + 0.5_double


    source%xpos = xcentre ! NB centres are at the middle of the pixel, pixel(0, 0) = 1, 1
    source%ypos = ycentre



    do i = 1, psfnx
        x(i) = dble(i)
    end do

    do j = 1, psfny
        y(j) = dble(j) 
    end do


    call GridInterpolantConstruct(psfnx, psfny, x, y, psf, psfGridFit1)


    fCentre = GridInterpolate(xCentre, yCentre, psfGridFit1)


    if(xres .gt. 1) then
        nx = psfnx / xres
        ny = psfny / yres
     else
        nx = psfnx
        ny = psfny
    end if


    if (bin .eq. 1) then ! No binning


     
        if (present(totalCounts)) then
            if (totalCounts .gt. 0.0) then
                f = sum(psf)

                psf = psf * totalCounts  / f

            else
                f = maxval(psf)
        
                psf = psf * abs(totalCounts) / f
                totalCounts = sum(psf)

            end if
        end if


        if (present(psfimageout)) then
            psfImageOut => allocate2dSingle(0, psfnx - 1, 0, psfny - 1, "psfimageout")
        
            psfImageout = psf


        end if 


        if (present(psfSplineFit)) then    

            call ConstructSpline(psfnx, psfny, x, y, psf, psfSplineFit)

            fCentre = interpolateSpline(0.0_double, 0.0_double, psfSplineFit)
 !           write(*, *) "Interpolated spline counts at centre=", fCentre

        end if


!write(*, *) "ompsf line 704"


        if(present(psfGridFitout)) then

!write(*, *) "ompsf line 709, xcentre=", xcentre, " ycentre=", ycentre
!write(*, *) "psfnx=", psfnx, " psfny=", psfny


            call GridInterpolantDestruct(psfGridFit1)
            do i = 1, psfnx
                x(i) = real(i) - xCentre
            end do

            do j = 1, psfny

                y(j) = real(j) - yCentre
            end do


            call GridInterpolantConstruct(psfnx, psfny, x, y, psf, psfGridFitOut)


!write(*, *) "Line 755  CreatePSFImage21- gridfit - interpolated value at 0, 0=",  &
!   gridinterpolate(0.0_double, 0.0_double, psfGridFitout)
!write(*, *) "Line 755 interpolated value at -1, -1=",  gridinterpolate(-1.0_double, -1.0_double, psfGridFitout)
!write(*, *) "Line 755 interpolated value at 1, 1=",  gridinterpolate(1.0_double, 1.0_double, psfGridFitout)

!write(*, *) "Line 755 interpolated value at 0.5, 0.5=",  gridinterpolate(0.5_double, 0.5_double, psfGridFitout)
!write(*, *) "Line 755 interpolated value at -0.5,-0.5=",  gridinterpolate(-0.5_double, -0.5_double, psfGridFitout)

!write(*, *) "Line 755 interpolated value at 0.5, -0.5=",  gridinterpolate(0.5_double, -0.5_double, psfGridFitout)
!write(*, *) "Line 755 interpolated value at 2.0, 2.0=",  gridinterpolate(2.0_double, 2.0_double, psfGridFitout)
!write(*, *) "Line 755 interpolated value at -2.0, -2.0=",  gridinterpolate(-2.0_double, -2.0_double, psfGridFitout)

        end if 

        if(present(bilinearData)) then

            call BilinearConstruct(psf, bilinearData)
        
        end if         

        deallocate(x, y)
        nullify(x, y)
        call CAL_releaseMemory(psf)
        return

    end if ! if binning equals 1





    !******************************************
    ! Finished if binning equals 1
    !*****************************************


    psfnx = nx / 2 + 1
    psfny = ny / 2 + 1
    deallocate(x, y)
        
    x => allocateSingle(1, psfnx, " x")
    y => allocateSingle(1, psfny, " y")


    !*********************************************
    !
    ! Construct the binned image
    !
    !********************************************

    psfImage => allocate2dSingle(0, psfnx - 1, 0, psfny - 1, "psfimage")
    psfImage = 0.0


    do j = 0, psfny - 1
        y1 = dble(j * bin) + 0.5_double
        x(j + 1) = dble(j + 1) - ycentre
      
        do i = 0, psfnx - 1
            x1 = dble(i * bin) + 0.5_double
            y(i + 1) = dble(i + 1) - xcentre
            f1= max(0.0_double, gridinterpolate(x1, y1, psfGridFit1))
            f2= max(0.0_double, gridinterpolate(x1 + 1.0_double, y1, psfGridFit1))
            f3= max(0.0_double, gridinterpolate(x1, y1 + 1.0_double, psfGridFit1))
            f4= max(0.0_double, gridinterpolate(x1 + 1.0_double, y1 + 1.0_double, psfGridFit1))
           
            psfImage(i, j) = f1 + f2 + f3 + f4 
          
        end do
    end do
  
   



    if (present(totalCounts)) then ! Normalize so that psf image contains the required counts
        if (totalCounts .gt. 0.0) then
            f = sum(psfImage)

            psfImage = psfImage * totalCounts  / f

        else
            f = maxval(psfImage)
        
            psfImage = psfImage * abs(totalCounts) / f
            totalCounts = sum(psfImage)

        end if
    else

        f = maxval(psfImage)


        psfimage = psfimage / f



    end if


    xCentre = real(psfnx) / 2.0
    yCentre = real(psfny) / 2.0


    do i = 1, psfnx
        x(i) = real(i) - 0.5 - xCentre
    end do

    do j = 1, psfny

        y(j) = real(j) - 0.5 - yCentre
    end do


    if (present(psfSplineFit)) then    

        call ConstructSpline(psfnx, psfny, x, y, psfImage, psfSplineFit)

        fCentre = interpolateSpline(0.0_double, 0.0_double, psfSplineFit)
!        write(*, *) "Interpolated spline counts at centre=", fCentre

    end if
  
   
    call CAL_releaseMemory(psf)


    if (present(psfGridFitOut)) then ! User wants the grid interpolant
        call GridInterpolantDestruct(psfGridFit1)
  

        call GridInterpolantConstruct(psfnx, psfny, x, y, psfimage, psfGridFit1)

        psfgridfit = psfgridfit1

        deallocate(x, y)

        psfnx = psfnx * 10
        psfny = psfnx


        x => allocateSingle(1, psfnx, "x")
        y => allocateSingle(1, psfny, "y")
        height => allocate2dSingle(0, psfnx - 1, 0,  psfny - 1, "height")


         
        do i = 1, psfnx
            x(i) = real(i) / 10.0 - 0.5 - xCentre
    

            do j = 1, psfny

                y(j) = real(j) / 10.0  - 0.5 - yCentre

!                height(i - 1, j - 1) = max(0.0_double, GridIntegration1(dble(x(i)), dble(y(j))))        
                 height(i - 1, j - 1) = gridinterpolate(dble(x(i)), dble(y(j)), psfGridFit1)

            end do

        end do

       

      
        do i = 1, psfnx
            x(i) = real(i) / 10.0 - 0.5 - xCentre
        end do

        do j = 1, psfny
            y(j) = real(j) / 10.0  - 0.5 - yCentre
        end do



        call GridInterpolantDestruct(psfGridFit1)
         


        call GridInterpolantConstruct(psfnx, psfny, x, y, height, psfGridFitOut)

        psfnx = psfnx / 10
        psfny = psfnx
        deallocate(height)
        nullify(height)


!write(*, *) "Line 933  CreatePSFImage21- gridfit - interpolated value at 0, 0=",  &
!   gridinterpolate(0.0_double, 0.0_double, psfGridFitout)
!write(*, *) "Line 935 interpolated value at -1, -1=",  gridinterpolate(-1.0_double, -1.0_double, psfGridFitout)
!write(*, *) "Line 936 interpolated value at 1, 1=",  gridinterpolate(1.0_double, 1.0_double, psfGridFitout)

!write(*, *) "Line 938 interpolated value at 0.5, 0.5=",  gridinterpolate(0.5_double, 0.5_double, psfGridFitout)
!write(*, *) "Line 939 interpolated value at -0.5,-0.5=",  gridinterpolate(-0.5_double, -0.5_double, psfGridFitout)

!write(*, *) "Line 941 interpolated value at 0.5, -0.5=",  gridinterpolate(0.5_double, -0.5_double, psfGridFitout)
!write(*, *) "Line 942 interpolated value at 2.0, 2.0=",  gridinterpolate(2.0_double, 2.0_double, psfGridFitout)
!write(*, *) "Line 943 interpolated value at -2.0, -2.0=",  gridinterpolate(-2.0_double, -2.0_double, psfGridFitout)





!write(*, *) "line 867"



!    do i = 1, psfnx
!        x(i) = real(i) - 0.5 - xCentre
    

!        do j = 1, psfny

!            y(j) = real(j) - 0.5 - yCentre
!            write(*, *) "x=", x(i), " y=", y(j), &
!               " psfimage(", i - 1, " , ", j - 1, ")=", psfimage(i - 1, j - 1), " grif-interpolated value=", &
!               gridinterpolate(dble(x(i)), dble(y(j)), psfgridfitout)

!        end do

!    end do

!    call GridInterpolantDestruct(psfGridFit1)


    end if  ! Construct grid-interpolant
  


    if(present(bilinearData)) then ! Construct bilinear interpolant

        call BilinearConstruct(psfimage, bilinearData)
           
    end if         



    deallocate(x, y)
    nullify(x, y)


    !*****************************************
    ! May want the actual psf image
    !***************************************** 
    if (present(psfImageOut)) then
        if (associated(psfImageOut)) deallocate(psfimageout)
         
        psfImageOut => psfImage


    else
        deallocate(psfImage)
        nullify(psfImage)
    end if 


end subroutine CreatePSFImage21test



!*****************************************************
!
!
!
!
!*****************************************************
subroutine CreatethePsfImage
    real(double) :: maxPsf
   
    real(single) ::  counts
    real(single) :: radius

   
    psfnx = 16
    psfny = 16
       
     
    l0 = 14 / binx


    if(associated( myPsfImage )) then
        deallocate( myPsfImage )
        nullify( myPsfImage )
    end if



    !**********************************************************************
    ! Get the point-spread function image
    !  
    !*********************************************************************

!    call CreatePSFImage21(psfnx=psfNx, psfny=psfNy,                            & 
!                          psfimageOut=myPsfImage,                              &
!                          xpos=1024.5_double,                                  & 
!                          ypos=1024.5_double,                                  &
!                          ratio=0.5_double, bin=binx, res=1,                   &
!                          psfGridFitout = psfInterpolant)


    call CreatePSFImage21test(psfnx=psfNx, psfny=psfNy,                            & 
                          psfimageOut=myPsfImage,                              &
                          xpos=1024.5_double,                                  & 
                          ypos=1024.5_double,                                  &
                          ratio=0.5_double, bin=binx, res=1,                   &
                          psfGridFitout = psfInterpolant)


    psfnx = size(mypsfimage, 1) 
    psfny = size(mypsfimage, 2)  


    psfXcentre = real(psfnx) / 2.0 ! The x centre of the psf image
    psfYcentre = real(psfny) / 2.0 ! The y centre of the psf image


    !*******************************************************
    !
    ! Normalize psf image to a maximum height of 1.0
    !
    !*******************************************************

    maxPsf = maxval(myPsfImage) 
    mypsfImage = mypsfImage / maxPsf

    !---------------------------------------------------
    ! 2010-11-11 test increasing the radius by 0.5 pixel
    !---------------------------------------------------
    radius=12.0/real(binx)




    call GetCounts2( mypsfImage, psfXCentre, psfYCentre, radius, counts)
    
    psfXcentre = psfxcentre + 0.5
    psfYcentre = psfYcentre + 0.5


    sumPsfImage = counts



end subroutine CreatethePsfImage


!*******************************************************************
! Least-squares fit of the kind y=constant+a1*x+a2*x etc
! nterms is the number of a constants to be solved for
!
!******************************************************************
subroutine LeastSquaresFit21test(nData, xData, ydata, weight, nterms, maxIter, &
            nstd1, gradient, gradientError, constant, constantError, std1, &
            residual1)
    integer, intent(in) :: nData
    real(double) :: xData(:, :), ydata(:), weight(:)
    real(double), optional :: residual1(:)

    integer(int8), intent(in) :: nTerms
    integer(int8), intent(in), optional :: maxIter
    real(single), intent(in), optional :: nstd1
    real(double), intent(out), dimension(:) :: gradient, gradientError
    real(double), intent(out) :: constant, constantError
    real(double), optional, intent(out) :: std1
    integer :: i, j, k, n1, m, ifail, nCoeffs, nits
  
    real(double), dimension(nterms + 1) :: beta, x
    real(double), dimension(:), allocatable :: sterr, var    
    real(double), pointer :: residual(:) => null()
    real(double) :: std, nstd
    real(double) :: fit, d    




    if(present(nstd1)) then
        nstd = nstd1
    else
        nstd = 3.0
    end if
    std = 1000000.0

    residual => allocateDouble(1, nData, "residual")
    
    residual = 0.0
    std = 100000.0
    if(present(maxIter)) then
        nits = maxIter
    else     
        nits = 2
    end if
    n1 = nData
    nCoeffs = nterms + 1
 
    write(*,*) "L278 nCoeffs=", nCoeffs
    m = ncoeffs * (ncoeffs + 1)
    allocate(sterr(1 : m))  

    allocate(var(1 : m))  


   
    do j = 1, nits
        
        call startup(nCoeffs, .false.)
        call tolset  


        m = 0

        do i = 1, nData
            if (residual(i) .gt. nstd * std) cycle
            
            m = m + 1
            
            x(1) = 1.0_double
            do k = 2, nCoeffs
                x(k) = xdata(i, k - 1)
            end do
          
            call includ( weight(i), x, ydata(i))


        end do
        if (j .gt. 1 .and. (m .lt. nData / 2 .or. m  .eq. n1)) exit
        n1 = m  


        !*****************************************************************
        ! Compute the regression coefficients
        !*****************************************************************
        call regcf(beta, ncoeffs, ifail)


        call cov(ncoeffs, d, var, size(var), sterr, ifail)
!        if(ifail .ne. 0) then
!            write(*, *) "Error! ifail=", ifail
!        endif   

        var = sqrt(var)

        do i = 1, nData
            fit = beta(1)
            do k = 2, nCoeffs
                fit = fit + xdata(i, k - 1) * beta(k)
            end do  
            residual(i) = (yData(i) - fit) * weight(i)
        end do
         
        std = sqrt(sserr / dble(m - 1))
    end do 
    if(present(std1) ) then

        std1 = std
    end if

    constant = beta(1)
    constantError = sterr(1)
!    write(*, *) "Background=", constant, "+/-", constantError
    do k = 1, nterms
        gradient(k) = beta(k + 1)
        gradientError(k) = sterr(k + 1)
!        write(*, *) "Least-squares fit 21 - gradient ", k, ")=", gradient(k), "+/-", gradientError(k)
    end do

    if(present(residual1)) then
        residual1(1 : nData) = residual(1 : nData)

    end if

 
    deallocate(residual)
    nullify(residual)
    deallocate(sterr, var)

end subroutine LeastSquaresFit21test


!*************************************************************************
!
! Least-squares to the PSF to include the background and an arbitrary
! number of linear constants
! ie y=constant+a1*x+a2*x, etc
! NO FIT TO THE COORDINATES
! nSources is the number of sources 
!
!************************************************************************
subroutine PSFPhotometry1(image, nSources,source, di, isource, &
     deblendedCounts, deblendedCounts3, deblendedX, deblendedY)
    real(single), pointer :: image(:, :)
    integer(int8), intent(in) :: nSources
    type(SourceParameters) :: source(1 : nSources)
    integer(int8), intent(in) :: di
    real(double), intent(out) :: deblendedCounts, deblendedCounts3, deblendedX, deblendedY

    integer :: nx, ny
    integer :: nx1, ny1

    integer :: i0, j0, i, j, n
   
    integer, intent(in) :: isource
    logical :: printFlag

    real(double), dimension(1 : nSources) :: gradient, gradientError
    real(double), dimension(1 : nSources) :: gradient1
    !real(double) :: deblendedCounts, deblendedX, deblendedY

    !---
    real(double), dimension(1 : nSources) :: maxSrcRecalculated
    
    real(double) :: constant, constantError
    real(double) :: constant1

    real(single) :: maxRadius1, extractionRadius
    integer :: halfRadius
    integer :: di15

    real(double) :: newRate, newRateError, newSignificance, counts, backgroundCounts

    type(SourceParameters) :: sourceAux

!    integer, save :: count = 0


    real(double) :: x(1 : nSources * 2)
    integer :: ii, jj, imax1, jmax1
    real(double) :: auxWeight, coord, auxSum, imageAver, imageMax1, imageMax2
    real(double) :: naux
    real(double) :: auxCoord, newX, newY
    integer :: iminC, imaxC, jminC, jmaxC
    integer :: iminC1, imaxC1, jminC1, jmaxC1

    real(double) :: imageMaxPrev
    integer :: niter

    !---------------
    integer :: imin2, imax2, jmin2, jmax2
    !integer :: imin, imax, jmin, jmax 
    integer :: imin0, imax0, jmin0, jmax0
    real(single) :: x1, y1, dx, dy, dx1, dy1, radius
    real(single) :: minSrcDistance
    integer :: searchRange=3
    
    integer :: ti0, tj0, tcount1, maxFoundFlag
    real(single) :: dist1, dist2
    !--------------

    printFlag=.false.
    !if (iSource .eq. 1) printFlag=.true. 
    !if (iSource .eq. 4) printFlag=.true.
    !if (iSource .eq. 45) printFlag=.true.
    !if (iSource .eq. 6) printFlag=.true.
    !if (iSource .eq. 9) printFlag=.true.
    !if (iSource .eq. 12) printFlag=.true.
    !if (iSource .eq. 28) printFlag=.true.
    !if (iSource .eq. 51) printFlag=.true.
    !if (iSource .eq. 25) printFlag=.true.
    numberToFit = nSources


    diFit = di

   
    nx = ubound(myimage, 1)
    ny = ubound(myimage, 2)

    nx1 = ubound(image, 1)
    ny1 = ubound(image, 2)

    maxRadius1 = real(20 / binx)
    extractionRadius=real(12/binx)
    
    !if (printFlag) then
    !   write(*,*) "L864 nx=", nx, " ny=", ny
    !   write(*,*) "L865 nx1=", nx1, " ny1=", ny1
    !endif  

    !******************************************************************
    !
    ! Find out roughly how many pixels to be used in the fit
    !
    !******************************************************************
    n = 0


    imin = lbound(myimage, 1)
    imax = ubound(myimage, 1)!

    jmin = lbound(myimage, 2)
    jmax = ubound(myimage, 2)


    imin0=imin
    imax0=imax
    jmin0=jmin
    jmax0=jmax

    !if (printFlag) then
    !   write(*,*) "L905 included into fitting nSources=", nSources
    !   write(*,*) "L906 imin=", imin, " imax=", imax, " jmin=", jmin, " jmax=", jmax
    !endif

    maxRadius = real(di)
    halfRadius=di/2.0

    di15=di+halfRadius


    if (printFlag) then
       write(*,*) "L915 halfRadius=", halfRadius, " maxRadius1=", maxRadius1, &
            " maxRadius=",maxRadius, " extractionRadius=", extractionRadius 
    endif  

    minSrcDistance=-1.0
    searchRange=int(extractionRadius/2)
    do i = 1, nSources ! the number of sources included into the fit

       !iminC = lbound(myimage, 1)
       !imaxC = ubound(myimage, 1)

       !jminC = lbound(myimage, 2)
       !jmaxC = ubound(myimage, 2) 


       i0 = int(source(1)%xpos + 0.5)
       if (i0 .lt. 0) i0=0
       if (i0 .gt. nx) i0= nx
       j0 = int(source(1)%ypos + 0.5)
       if (j0 .lt. 0) j0=0
       if (j0 .gt. ny) j0= ny

       imagemaxPrev=myimage(i0,j0)

       iminC=i0-searchRange
       if (iminC .lt. 0) iminC=0
       imaxC=i0+searchRange
       if (imaxC .gt. nx) imaxc=nx
       jminC=j0-searchRange
       if (jminC .lt. 0) jminC=0
       jmaxC=j0+searchRange
       if (jmaxC .gt. ny) jmaxc=ny
       

       if (i .eq. 1) then

          imin=iminC
          imax=imaxC
          jmin=jminC
          jmax=jmaxC
          
          imin0=imin
          imax0=imax
          jmin0=jmin
          jmax0=jmax
          radius=extractionRadius/2.0
          do j=2,nSources
             x1=real(ii)
             y1=real(jj)
             dx1=source(j)%xpos-real(i0)
             dy1=source(j)%ypos-real(j0)
             radius=sqrt(dx1*dx1+dy1*dy1)
             if (printFlag) then
                write(*,*) "L980 j=", j, " radius=", radius, " extrRad/2=", extractionRadius/2.0
             endif
             if (radius/2.0 .le. searchRange) then
                if (printFlag) then
                   write(*,*) "L984 reducing searchRange"
                endif
                searchRange=int(radius/2.)
                iminC=i0-searchRange
                if (iminC .lt. 0) iminC=0
                imaxC=i0+searchRange
                if (imaxC .gt. nx) imaxc=nx
                jminC=j0-searchRange
                if (jminC .lt. 0) jminC=0
                jmaxC=j0+searchRange
                if (jmaxC .gt. ny) jmaxc=ny
                if (printFlag) then
                   write(*,*) "L995 iminC=", iminC, " imaxC=", imaxC, " jminC=", jminC, " jmaxC=", jmaxC
                endif
             endif
          enddo

       endif

       if (printFlag) then
          write(*,*) "---"
          write(*,*) "L1003 i=",i
          write(*,*) "L1004 i0=", i0, " j0=", j0, " imagemaxPrev=", imageMaxPrev
          write(*,*) "L1005 iminC=", iminC, " imaxC=", imaxC, " jminC=", jminC, " jmaxC=", jmaxC
       endif

       ! try to find the maximum using gradients



       if (i .eq. 1) then      

          iminC1=i0-1
          if (iminC1 .lt. 0) iminC1=0
          imaxC1=i0+1
          if (imaxC1 .gt. nx) imaxC1=nx
          jminC1=j0-1
          if (jminC1 .lt. 0) jminC1=0
          jmaxC1=j0+1
          if (jmaxC1 .gt. ny) jmaxC1=ny

          imageMax1=myimage(i0,j0)
          ti0=i0
          tj0=j0
          do tcount1=1,5 
             maxFoundFlag=-1
             !--
             do ii=iminC1,imaxC1
                do jj=jminC1,jmaxC1
                   if (ii .ne. i .and. jj .ne. j) then
                      if (myimage(ii,jj) .gt. imageMax1) then
                         imageMax1=myimage(ii,jj)
                         ti0=ii
                         tj0=jj
                         maxFoundFlag=1
                      endif
                   endif
                enddo
             enddo
             if (maxFoundFlag .gt. 0) then
                iminC1=ti0-1
                if (iminC1 .lt. 0) iminC1=0
                imaxC1=ti0+1
                if (imaxC1 .gt. nx) imaxC1=nx
                jminC1=tj0-1
                if (jminC1 .lt. 0) jminC1=0
                jmaxC1=tj0+1
                if (jmaxC1 .gt. ny) jmaxC1=ny
             else
                exit
             endif
             !--
          enddo ! tcount1=1,5
          i0=ti0
          j0=tj0
          source(i)%xpos=real(i0)
          source(i)%ypos=real(j0)
       else

          ti0=source(i)%xpos
          tj0=source(i)%ypos      
          imageMax1=myimage(ti0,tj0)
          dx1=real(ti0)-real(i0)
          dy1=real(tj0)-real(j0)
          dist1=sqrt(dx1*dx1+dy1*dy1)
          dist2=dist1

          if (printFlag) then
             write(*,*) "L1075 ti0=", ti0, " tj0=", tj0, " dist1=", dist1
          endif

          iminC1=ti0-1
          if (iminC1 .lt. 0) iminC1=0
          imaxC1=ti0+1
          if (imaxC1 .gt. nx) imaxC1=nx
          jminC1=tj0-1
          if (jminC1 .lt. 0) jminC1=0
          jmaxC1=tj0+1
          if (jmaxC1 .gt. ny) jmaxC1=ny

          do tcount1=1,5 
             maxFoundFlag=-1
             !--
             do ii=iminC1,imaxC1
                do jj=jminC1,jmaxC1
                   if (ii .ne. i .and. jj .ne. j) then
                      if (myimage(ii,jj) .gt. imageMax1) then
                         ! check whether this is the direction towards the main source
                         dx1=real(ii)-real(i0)
                         dy1=real(jj)-real(j0)
                         dist2=sqrt(dx1*dx1+dy1*dy1)
                         if (dist2 .gt. dist1) then
                            ! accept this as a next maximum
                            imageMax1=myimage(ii,jj)
                            ti0=ii
                            tj0=jj
                            maxFoundFlag=1
                         endif
                      endif
                   endif
                enddo ! jj
             enddo ! ii

             if (printFlag) then
                write(*,*) "L1111 ti0=", ti0, " tj0=", tj0, " dist1=", dist1, " dist2=", dist2, &
                     " maxFoundFlag=", maxFoundFlag, " tcount1=", tcount1
             endif

             if (maxFoundFlag .gt. 0) then
                iminC1=ti0-1
                if (iminC1 .lt. 0) iminC1=0
                imaxC1=ti0+1
                if (imaxC1 .gt. nx) imaxC1=nx
                jminC1=tj0-1
                if (jminC1 .lt. 0) jminC1=0
                jmaxC1=tj0+1
                if (jmaxC1 .gt. ny) jmaxC1=ny


                dx1=real(ti0)-real(i0)
                dy1=real(tj0)-real(j0)
                dist1=sqrt(dx1*dx1+dy1*dy1)                
                dist2=dist1
             else
                exit
             endif
             !--
          enddo

          
       endif ! i eq 1

       if (printFlag) then
          write(*,*) "L1061  gradient maximum search: ti0=", ti0, " tj0=", tj0, &
               " imageMax1=", imageMax1
       endif
       imax1=ti0
       jmax1=tj0

       !recalculate imageMax1 as an average in the vicinity of the max
       imageMax2=0.0
       naux=0.0
       do ii=ti0-1,ti0+1
          do jj=tj0-1,tj0+1
             if (ii .ge. 0 .and. ii .le. nx .and. jj .ge. 0 .and. jj .le. ny) then
                if (myimage(ii,jj) .gt. 0.0) then
                   imageMax2=imageMax2+myimage(ii,jj)
                   naux=naux+1.0                   
                endif
             endif
          enddo
       enddo
       if (naux .gt. 1.0) then
          imageMax2=imageMax2/naux
       endif

       !else
       !   ! check whether the source is really a neighbour and not the primary
       !   do ii=iminC,imaxC
       !      do jj=jminC,jmaxC
       !         x1=real(ii)
       !         y1=real(jj)
       !         dx1=x1-x(1)
       !         dy1=y1-x(2)
       !         radius=sqrt(dx1*dx1+dy1*dy1)
       !         if (radius .gt. extractionRadius/2.0) then
       !            if (myimage(ii,jj) .gt. imageMax1) then
       !               imageMax1=myimage(ii,jj)
       !               imax1=ii
       !               jmax1=jj
       !            endif
       !         endif
       !      enddo
       !   enddo
       !endif

       !maxSrcRecalculated(i)=imageMax1
       maxSrcRecalculated(i)=imageMax2

       if (printFlag) then
          write(*,*) "L991 imax1=", imax1, " jmax1=", jmax1, " imageMax1=", imageMax1, &
               " imageMaxPrev=",imageMaxPrev, " imageMax2=", imageMax2 
       endif

       iminC=imax1-extractionRadius/2
       if (iminC .lt. 0) iminC=0
       imaxC=imax1+extractionRadius/2
       if (imaxC .gt. nx) imaxC=nx
       jminC=jmax1-extractionRadius/2
       if (jminC .lt. 0) jminC=0
       jmaxC=jmax1+extractionRadius/2
       if (jmaxC .gt. ny) jmaxC=ny      
       if (printFlag) then
          write(*,*) "L988 iminC=", iminC, " imaxC=", imaxC, " jminC=", jminC, " jmaxC=",jmaxC
       endif

       if (i .eq. 1) then          
          imin0=iminC
          imax0=imaxC
          jmin0=jminC
          jmax0=jmaxC
       endif

       
       ! modify the coordinates of the source
       newX=real(imax1)
       newY=real(jmax1)

       source(i)%xpos=newX
       source(i)%ypos=newY
       if (printFlag) then
          write(*,*) "L1022 newX=", newX, " newY=", newY
       endif
       !iminC=imax1-di
       !if (iminC .lt. 0) iminC=0
       !imaxC=imax1+di
       !if (imaxC .gt. nx) imaxC=nx
       !jminC=jmax1-di
       !if (jminC .lt. 0) jminC=0
       !jmaxC=jmax1+di
       !if (jmaxC .gt. ny) jmaxC=ny      

       if (i .eq. 1) then
          imin=iminC
          imax=imaxC
          jmin=jminC
          jmax=jmaxC
       endif

       if (printFlag) then
          write(*,*) "L1016 iminC=", iminC, " imaxC=", imaxC, " jminC=", jminC, " jmaxC=", jmaxC
          write(*,*) "L1017 imin=", imin, " imax=", imax, " jmin=", jmin, " jmax=", jmax
          write(*,*) "---"
       endif

       !-----------------------------------------
       !
       !-----------------------------------------

       j = (i - 1) * 2 + 1
   

       x(j) = newX 
       x(j + 1) =  newY 
    
    end do

    if (printFlag) then
       write(*,*) "L1054 imin=", imin, " imax=", imax, " jmin=", jmin, " jmax=", jmax
       write(*,*) "L1055 imin0=", imin0, " imax0=", imax0, " jmin0=", jmin0, " jmax0=", jmax0
    endif
    !---------------------
    imax2=imax0
    imin2=imin0
    jmax2=jmax0
    jmin2=jmin0   
    if ((imax0-imin0) .lt. int(2.0*maxRadius1)) then
       ii=(int(2.0*maxRadius1+0.5)-(imax0-imin0))/2
       imax2=imax2+ii
       imin2=imin2-ii
    endif
    if ((jmax-jmin) .lt. int(2.0*maxRadius1)) then
       jj=(int(2.0*maxRadius1+0.5)-(jmax-jmin))/2
       jmax2=jmax2+jj
       jmin2=jmin2-jj
    endif
    !--------------------

    !n = (imax - imin + 1) * (jmax - jmin + 1) ! The maximum number of pixels to be used
    n = (imax2 - imin2 + 1) * (jmax2 - jmin2 + 1) 

    !write(*,*) "L1171 allocating xdata, n=", n, "<---------1"
    xdata => allocate2dDouble(1, n, 1, int(nSources, int32), "xdata")
    ydata => allocateDouble(1, n, "ydata")
    weight => allocateDouble(1, n, "weight")

    xdata2 => allocate2dDouble(1, n, 1, int(nSources, int32), "xdata2")
    !write(*,*) "L1154 xdata2 is allocated..."

    n = 0


!*****************************************************
!
! FOR TESTING
!
!*****************************************************
!myImage = 10.0
!write(*, *) "Line 435"
!constant = PSFFunction12(x) 
!write(*, *) "Line 437"



    !if (printFlag) then
    !   write(*, *) "L1048, psfnx=", psfnx, " psfny=", psfny
    !   write(*, *) "The sources:"
    !   do i = 1, nSources
    !      write(*, *) i, ") x=", source(i)%xpos, " y=", source(i)%ypos, &
    !           " raw=", source(i)%rawCountRate, " exposure=", source(i)%exposure
    !   end do
    !   write(*, *) "L1054 maxradius=", maxradius   
    !endif
! end FOR TESTING

    !***************************************************************
    !
    ! Do the least-squares fit
    !
    !**************************************************************
    !call doLeastSquaresFit(nSources, x,                    &
    !                         constant, constantError,      &
    !                         gradient, gradientError,isource, &
    !                         constant1, gradient1) 

    !
    ! Trying to change the PSF fitting algorithm 2011.11.13

!    call doLeastSquaresFit2(nSources, x, maxSrcRecalculated, maxRadius1, maxRadius1,  &
!                             constant, constantError,      &
!                             gradient, gradientError,isource, &
!                             constant1, gradient1) 

!    imin=imin0
!    imax=imax0
!    jmin=jmin0
!    jmax=jmax0
    
    call deblendSource(nSources, x, maxSrcRecalculated, extractionRadius, maxRadius1,  &
                             deblendedCounts, deblendedCounts3, deblendedX, deblendedY, &
                             constant, constantError,      &
                             gradient, gradientError,isource, &
                             constant1, gradient1) 

  
    !*******************************************************************
    !
    ! THE NEW RAW COUNT RATES FOR A 12 radius aperture
    !
    !*******************************************************************


    computedBackground = constant
    

    counts =  gradient(1) * sumPsfImage
     
    !*****************************************
    ! 2011-11-17
    !newRate =  counts / source(1)%exposure
    !newRate =  deblendedCounts / source(1)%exposure
    newRate =  deblendedCounts3 / source(1)%exposure
   
    !**********************************************


    !if (printflag) then
    !write(*,*) "L963 constant1=", constant1, " gradient1(1)=", gradient1(1)
    !write(*,*) "L964 constant =", constant, " gradient(1) =", gradient(1)
    !write(*,*) "       sumPsfImage=", sumPsfImage, " counts=", counts
       
    !endif



    backgroundCounts = constant * PI * real(144  / &
                  (binX * binX)) 
    

    newRateError = ComputeSourceRateError1(counts, backGroundCounts, source(1)%exposure)
    




    if(counts > 0.0) then
        newSignificance = counts / sqrt(counts + backgroundCounts)        
    else
        newSignificance = 0.0
    end if 
!write(*, *) "new rate=", newRate, "+/-", newRateError
!  write(*, *) "gradient error=", gradientError(1)
    if( displayResults) then
 
        write(errstr, '(a, f0.2, a, f0.2, a, f0.2, a, f0.2, a, f0.2, a, f0.2)' )  &
                "Raw countrate old=", source(1)%rawCountRate, " +-", source(1)%rawCountRateErr,  &
                ", new=", newRate, " +-", newRateError,           &
                ", significance old=", source(1)%significance, ", new=", newSignificance  
        call message(AppMsg, Verbose)
    end if

    if (printFlag) then
        write(*, '(a, f0.2, a, f0.2, a, f0.2, a, f0.2, a, f0.2, a, f0.2)' )  &
                "L1193 Raw countrate old=", source(1)%rawCountRate, " +-", source(1)%rawCountRateErr,  &
                ", new=", newRate, " +-", newRateError,           &
                ", significance old=", source(1)%significance, ", new=", newSignificance 
         write(*, '(a, f0.2, a, f0.2, a, f0.2)') "L1196 deblended Counts=", deblendedCounts, &
              " x=", deblendedX, " y=", deblendedY
         write(*,*) "-------------"
    endif
    source(1)%apertureRadius = myApertureRadius


    if( newRate > 0.0) then
        source(1)%rawCountRate = newRate
        source(1)%rawCountRateErr = newRateError
        source(1)%significance = newSignificance
        source(1)%rawBackgroundCountRate = backgroundCounts / source(1)%exposure
        source(1)%rawBackgroundCountRateErr = sqrt(max(0.0_double, backgroundCounts)) / source(1)%exposure

        !*************************************************2010-03-03 vny
        ! Implementing the coincidence-loss correction for the psf-photometry 
        ! count rates 
        !---------------------------------------------------------------------
        sourceAux=source(1)
        ! 2013-10-14 adding deadFrac to the routine PointSourceCoincLossCorr
        !call PointSourceCoincLossCorr(sourceAux, 12.0, frameTime) 
        call PointSourceCoincLossCorr(sourceAux, 12.0, frameTime, deadFrac) 
        
        source(1)%corrCountRate=sourceAux%corrCountRate
        source(1)%corrCountRateErr=sourceAux%corrCountRateErr
        source(1)%corrBackgroundCountRate=sourceAux%corrBackgroundCountRate
        source(1)%corrBackgroundCountRateErr=sourceAux%corrBackgroundCountRateErr
        source(1)%coincidenceLossCorr=sourceAux%coincidenceLossCorr
        !************************************************* 2010-03-03 vny

    else

        call warning("Negative count-rate- ignoring source")
    end if


  
   
    deallocate(xdata, xdata2, ydata, weight)
    nullify(xdata, xdata2, ydata, weight)

   

end subroutine PSFPhotometry1





!************************************************************************************
!
!
! Least-squares fit to PSF, including background, and including x and y
! for 1 or more sources
!
!
!***********************************************************************************
subroutine PSFPhotometry2(nSources, di, source, x, isource)
  

    integer(int8), intent(in) :: nSources
    integer :: nx, ny
    integer, intent(in) :: isource

    integer(int8), intent(in) :: di  

    type(SourceParameters) :: source(1 : nSources)

    real(double) :: x(:)
 
    integer(int32) :: i0, j0
    integer(int32) :: n
    integer :: n0, ngen, ngenout, fitNumber(1 : nSources * 2)
    integer(int8) :: dx
   
    real(double) :: funcValue   
    real(double) :: tolerance
    real(single) :: dc
    integer :: i, j, nop

    !---------------
    real(single) :: maxRadius1 
    integer :: ii,jj, imin2, imax2, jmin2, jmax2
    integer :: imin0, imax0, jmin0, jmax0
    maxRadius1 = real(20 / binx)
    !--------------


    diFit = di
    numberToFit = nSources

    nx = ubound(myimage, 1)
    ny = ubound(myimage, 2)

    !write(*,*) "L1372 di=", di, " numberToFit=", numberToFit

    !******************************************************************
    !
    ! Find out roughly how many pixels to be used in the fit
    !
    !******************************************************************
    n = 0

    imin = lbound(myimage, 1)
    imax = ubound(myimage, 1)

    jmin = lbound(myimage, 2)
    jmax = ubound(myimage, 2)

    imin0=imin
    imax0=imax
    jmin0=jmin
    jmax0=jmax

    dc = real(4 / binx)
    dx = di + dc
    !write(*,*) "L1390 dc=", dc, " dx=", dx

    do i = 1, nSources
        j = (i - 1) * 2 + 1
        fitNumber(j) = j
        fitNumber(j + 1) = j + 1

        x(j) = source(i)%xpos 

        bl(j) = source(i)%xpos - dc
        if (bl(j) .lt. imin0) bl(j)=imin0

        !write(*,*) "L1400 j=", j, " x(j)=", x(j), " bl(j)=", bl(j) 
 
!        bl(j) = max( -1.0_double / dble(binX), bl(j) )
        !write(*,*) "L1403  bl(j)=", bl(j)
    
    
        bu(j) = source(i)%xpos + dc
        if (bu(j) .gt. imax0) bu(j)=imax0

        !write(*,*) "L1406  bu(j)=", bu(j)   
!        bu(j) = min( dble(nx) + 0.5_double  + dble(1.0_double / dble(binX)), bu(j) )
        !write(*,*) "L1408  bu(j)=", bu(j)     
    
        x(j + 1) =  source(i)%ypos 
        bl(j + 1) = source(i)%ypos - dc
        if (bl(j+1) .lt. jmin0) bl(j+1)=jmin0

!        bl(j + 1) = max( -1.0_double / dble(binX), bl(j + 1) )
    
        bu(j + 1) = source(i)%ypos + dc
        if (bu(j+1) .gt. jmax0) bu(j+1)=jmax0

!        bu(j + 1) = min( dble(ny) + 0.5_double  + dble(1.0_double / dble(binX)), bu(j + 1) )
  
        !write(*,*) "L1409  j+1=", j+1, " x(j+1)=", x(j+1), " bl(j+1)=", bl(j+1), &
        !     " bu(j+1)=", bu(j+1)      

        i0 = max(0, int(bl(j) - 0.5))
        j0 = max(0, int(bl(j + 1) - 0.5))

        !write(*,*) "L1423 i0=", i0, " j0=", j0

        imin= i0-dx
        if (imin .lt. 0) imin=0
        imax=i0+dx
        if (imax .gt. nx) imax=nx
        jmin=j0-dx
        if (jmin .lt. 0) jmin=0
        jmax=j0+dx
        if (jmax .gt. ny) jmax=ny

        !write(*,*) "L1431 imin=", imin, " imax=", imax, " jmin=", jmin, " jmax=", jmax

        !if (i .eq. 1) then          
        !   imin0=imin
        !   imax0=imax
        !   jmin0=jmin
        !   jmax0=jmax
        !endif

    end do

    !write(*,*) "L1437 imin=", imin, " imax=", imax, " jmin=", jmin, " jmax=", jmax
    imin2=imin
    imax2=imax
    jmin2=jmin
    jmax2=jmax
    !---------------------
    if ((imax-imin) .lt. int(2.0*maxRadius1)) then
       ii=(int(2.0*maxRadius1+0.5)-(imax-imin))/2
       imax2=imax2+ii
       imin2=imin2-ii
    endif
    if ((jmax-jmin) .lt. int(2.0*maxRadius1)) then
       jj=(int(2.0*maxRadius1+0.5)-(jmax-jmin))/2
       jmax2=jmax2+jj
       jmin2=jmin2-jj
    endif
    !--------------------

    !n = (imax - imin + 1) * (jmax - jmin + 1) ! The maximum number of pixels to be used
    n = (imax2 - imin2 + 1) * (jmax2 - jmin2 + 1) 
    !write(*,*) "L1443 imin=", imin, " imax=", imax, " jmin=", jmin, &
    !     " jmax=", jmax 
    !write(*,*) "L1445 imin2=", imin2, " imax2=", imax2, " jmin2=", jmin2, &
    !     " jmax2=", jmax2 
    !write(*,*) "L1442 allocating xdata, n=", n, "<---------2"
    xdata => allocate2dDouble(1, n, 1, int(nSources, int32), "xdata")
    ydata => allocateDouble(1, n, "ydata")
    weight => allocateDouble(1, n, "weight")

    xdata2 => allocate2dDouble(1, n, 1, int(nSources, int32), "xdata2")

    nop = nSources * 2
 
    tolerance = 1e-6




!write(*, *) "Line 534 before calling optimize x(1)=", x(1), " x(2)=", x(2)
!write(*, *) "bl(1)=", bl(1), " bu(1)=", bu(1)
!write(*, *) "bl(2)=", bl(2), " bu(2)=", bu(2)


    do i = 1, 1

        n0 = 100
        ngen = 200  
        call optimize(nop, fitNumber, x, bl, bu, &
                   n0, ngen, tolerance, psfFunction100, funcValue, ngenout)
        !write(*, *) "Iteration ", i, ") funcvalue=", funcValue, " ngens=", ngenout
    end do  

  
 
!write(*, *) "Line 548 after calling optimize x(1)=", x(1), " x(2)=", x(2)

 

    deallocate(xdata, xdata2, ydata, weight)
    nullify(xdata, xdata2, ydata, weight)


    !imin=imin0
    !imax=imax0
    !jmin=jmin0
    !jmax=jmax0
   

end subroutine PSFPhotometry2


!***********************************************************
! 
! Fitting subroutine used by optimize
! 
!**********************************************************
function PSFFunction100(x) result(functionValue)

    real(double) :: x(:)
    real(double) :: functionValue
  
    real(double), dimension(1 : numberToFit) :: gradient, gradientError
    real(double), dimension(1 : numberToFit) :: gradient1

    real(double) :: constant, constantError
    real(double) :: constant1

    integer :: i, k, ndata
    real(double) :: fit, residual


    gradientError = 0.0

    constantError = 0.0
  
    maxRadius = real(diFit)

    i=0
    !***************************************************************
    !
    ! Do the least-squares fit
    !
    !**************************************************************

    call doLeastSquaresFit(numberToFit, x,                 &
                             constant, constantError,      &
                             gradient, gradientError, i, &
                             constant1, gradient1, ndata)   




    !****************************************************************
    ! Compute the weighted residuals
    !***************************************************************
    
    functionValue = 0.0_double

    do i = 1, ndata

        fit = dble(constant)

        do k = 1, numberToFit ! Sum for each source
            fit = fit + gradient(k) * xdata(i, k)
        end do
  
        residual = fit - ydata(i)

        functionValue = functionValue + residual * residual * weight(i)

!         functionValue = functionValue + residual * dble(weight(i))



    end do 

!    functionValue = -abs(functionvalue)


    functionValue = -functionvalue / dble(ndata) 

   

!write(*, *) "PSFFUNCTION100 x=", x, " functionvalue=", functionvalue

   
end function PSFFunction100

!*************************************************************
!
!
!
!************************************************************
real(single) function PsfInterpolation(xpos, ypos)
    real(single), intent(in) :: xPos, yPos

    psfInterpolation = max(0.0, gridInterpolate(xpos, ypos, psfInterpolant))

end function psfInterpolation


!*********************************************************
!
!
!
!*********************************************************
real(single) function weightingFunction( data )
    real(double), intent(in) :: data

!    real(single), intent(in) :: radius

    if ( data .gt. 0.0_double .and. weightingMethod .eq. 1) then
        

        weightingFunction = 1.0_double / max(1.0_double, data)
        
!        weightingFunction = (1.0_double - radius / maxRadius)  / max(1.0_double, data)
!        weightingFunction = max(0.0, weightingFunction)



    else
        weightIngFunction  = 1.0_double
    end if



end function weightingFunction


!*************************************************************************
!
! The main routine for PSF photometry 
!
!
!************************************************************************
subroutine PSFPhotometry(image, nSources, neSources, source, esource, di, type, minNumNeighbours)
  real(single), pointer :: image(:, :)
    integer(int32), intent(in) :: nSources
    type(SourceParameters), pointer :: source(:)

    integer(int32), intent(in) :: neSources
    type(SourceParameters), pointer :: esource(:)

    integer(int8), intent(in) :: di, type, minNumNeighbours
    integer :: i
    integer(int8) :: j, k, l
    real(single) :: tolerance, xOld, yOld
    real(double) :: x(1 : nSources * 2)
 
  
  
    type(SourceParameters), pointer :: tempSource(:) => null()

    real(double) :: deblendedCounts, deblendedCounts3, deblendedX, deblendedY

    ! vny 2010-10-09 reducing tolerance to 14
    !tolerance = real(26 / binx)

    !tolerance = real(14 / binx)
    tolerance = real(22 / binx)

    allocate(tempSource(1 : nSources))

    !write(*, *) "L718, nsources=", nSources, " summyimage=", sum(MyImage)

      
    !************************************************************
    !
    ! Loop through all the sources and perform aperture photometry
    !
    !************************************************************
    time = -10000.0
    displayResults = .true.

    !write(*, *) "L1616 tolerance=", tolerance, " minNumNeighbours=", minNumNeighbours

    do i = 1, nSources
    !do i = 76, nSources
        !write(*, *) i, ") line 725 maxCountRate=",maxRawCountRate, &
        !     " rawCountRate+",source(i)%rawCountRate 
        if(source(i)%extensionFlag == 1) then
            cycle
        end if

        if(source(i)%rawCountRate > maxRawCountRate) then
            cycle
        end if


        !write(*, *) i, ") line 736 tolerance=", tolerance, " minNumNeighbours=", minNumNeighbours
   

        k = 1
        tempSource(1) = source(i) 
  
       
        !**********************************************************
        ! Find the sources with distance <= tolerance from source i
        !
        !*********************************************************
        call  findNeighbours(i, source, esource, nSources, neSources, tolerance, tempSource, k)

       
        if( k .lt. minNumNeighbours) then
            cycle
        end if      

        call cpu_time(currentTime)

        !write(*,*) "L756 time=", time, " currentTime=", currentTime, &
        !     " diff=",currentTime - time
 
        !if( (currentTime - time) >= maxTime) then
            displayResults = .true.

            ! Write out the source list to the oswDataSet
!            write(errstr,'(a, i0, a, i0, a, f0.2, a, f0.2)') &
!                "L1615 ------ Source No.", i, ",    neighbours:  ", k-1, &
!                " x=",source(i)%xpos, " y=",source(i)%ypos 
!            call message(AppMsg, Verbose)
        !end if

!            write(*,'(a, i0, a, i0, a, f0.2, a, f0.2, a, i0)') "L1620 i=", i, &
!                 " neighbours=", k-1,  " x=",source(i)%xpos, " y=",source(i)%ypos, &
!                 " type=", type


        !write(*, *) "L776 Neighbours: ", k - 1      
        !write(*, *) "The sources in the fit"
        !do j = 1, k
        !    write(*, *) j, ") x=", tempSource(j)%xpos, " y=", tempSource(j)%ypos
        !end do
        !write(*,*) "--- "

        select case(type)

        case (1)

            !************************************************
            ! Do not fit the coordinates
            !
            !************************************************           
            call PSFPhotometry1(image, k, tempSource, di, i, &
                 deblendedCounts, deblendedCounts3, deblendedX, deblendedY)
           
        case (2)

            !************************************************
            !
            ! Fit the coordinates
            !
            !************************************************
            xOld = tempSource(1)%xpos
            yOld = tempSource(1)%ypos

           
            call PSFPhotometry2(k, di, tempSource, x, i)

         
        
            do j = 1, k
                l = (j - 1) * 2 + 1     
                tempSource(j)%xpos = x(l)
                tempSource(j)%ypos = x(l + 1)

            end do
            numberToFit = k 
            call PSFPhotometry1(image, numberToFit, tempSource, di, i, &
                 deblendedCounts, deblendedCounts3, deblendedX, deblendedY)


            if( displayResults ) then
    
                !write(errstr,'(a, f0.2, a, f0.2, a, a, f0.2, a)')                  &
                !    "(old-new)dx=", (xOld - tempSource(1)%xpos),                &
                !    " pixels, (old-new)dy=", (yOld - tempSource(1)%ypos), " pixels", & 
                !    ", computed background=", computedBackground, " counts per pixel"!
!
!                call message(AppMsg, Verbose)

!                write(*,'(a, f0.2, a, f0.2, a, f0.2, a, f0.2,a, f0.2)') "l1675 oldX=", xOld, &
!                     " newX=",tempSource(1)%xpos, &
!                     " oldY=", yOld, " newY=",tempSource(1)%ypos, " compBkg=",computedBackground  

!                write(*,'(a, i0,a, i0, 7(a, f0.2))') "l1680 i=", i, &
!                     " neighbours=", k-1, " x=", xOld, " y=", yOld,  &
!                     " compBkg=",computedBackground, " rate=", tempSource(1)%rawCountRate, &
!                     " dCounts=", deblendedCounts, " dX=", deblendedX, " dY=", deblendedY

!               write(*,'(a, i0,a, i0, 7(a, f0.2))') "l1729 i=", i, &
!                     " neighbours=", k-1, " x=", xOld, " y=", yOld,  &
!                     " compBkg=",computedBackground, " rate=", tempSource(1)%rawCountRate!, &
!                     !" rate3=", deblendedCounts3/1000.0
                
            end if
       
        end select

        source(i) = tempSource(1)
        time = currentTime
        displayResults = .false.
 
    end do


    ! continue with the extended source list
    displayResults = .true.

    !write(*,*) "L845 neSources=", neSources
    if (neSources .gt. 0) then
       !write(*,*) " Continuing with the extended source list"
       do i = 1, neSources

          if(.not. esource(i)%extensionFlag == 1) then
             cycle
          end if

          if(esource(i)%rawCountRate > maxRawCountRate) then
             cycle
          end if

          k = 1
   
          tempSource(1) = esource(i) 
  

          !**********************************************************
          ! Find the sources with distance <= tolerance from source i
          !
          !*********************************************************
          call  findNeighboursExtended(i, esource, source, tolerance, tempSource, k)


          if( k .le. minNumNeighbours) then
             cycle
          end if

          call cpu_time( currentTime)

          !if( (currentTime - time) >= maxTime) then
          displayResults = .true.

          ! Write out the source list to the oswDataSet
          write(errstr,'(a, i0, a, i0)') &
               "------ Extended source No.", i, ",    neighbours:  ", k-1
          call message(AppMsg, Verbose)
          !end if

          !do j = 1, k
          !    write(*, *) j, ") x=", tempSource(j)%xpos, " y=", tempSource(j)%ypos, &
          !         " raw=", tempSource(j)%rawCountRate, " err=",tempSource(j)%rawCountRateErr 
          !end do

          select case(type)

          case (1)

             !************************************************
             ! Do not fit the coordinates
             !
             !************************************************           
             call PSFPhotometry1(image, k, tempSource, di,i, &
                  deblendedCounts, deblendedCounts3, deblendedX, deblendedY)
             
          case (2)

             !************************************************
             !
             ! Fit the coordinates
             !
             !************************************************
             xOld = tempSource(1)%xpos
             yOld = tempSource(1)%ypos

           
             call PSFPhotometry2(k, di, tempSource, x, i)
         
        
             do j = 1, k
                l = (j - 1) * 2 + 1     
                tempSource(j)%xpos = x(l)
                tempSource(j)%ypos = x(l + 1)

             end do
             numberToFit = k 
             call PSFPhotometry1(image, numberToFit, tempSource, di,i, &
                  deblendedCounts, deblendedCounts3, deblendedX, deblendedY)


             if( displayResults ) then
    
                write(errstr,'(a, f0.2, a, f0.2, a, a, f0.2, a)')                  &
                     "(X old - new)=", (xOld - tempSource(1)%xpos),                &
                     " pixels, (Y old - new)=", (yOld - tempSource(1)%ypos), " pixels", & 
                     ", computed background=", computedBackground, " counts per pixel"
                
                call message(AppMsg, Verbose)
   
             end if
       


          end select

          esource(i) = tempSource(1)
          time = currentTime
          displayResults = .false.
 

       end do
    endif ! neSources>0


    deallocate(tempSource)
    nullify(tempSource)



    nx = size(myImage, 1)
    ny = size(myImage, 2)

  

    if(associated(myImage)) then
       nullify(myImage)
    end if


!write(*, *) "Line 1656, nsources=", nSources, " summyimage=", sum(MyImage)


!call saveImage(myImage, int(size(myImage, 1), int16), int(size(myImage, 2), int16), &
!    "checkimage.fits", "checkImage.fits")
!call fatal("stopped at line 1596")


end subroutine PSFPhotometry

!************************************************
!
! Locates all sources within a distance tolerance
! of source i and adds them to the list to be
! included in the PSF fitting process
!
!***********************************************
recursive subroutine findNeighbours( i, source, esource, nSources, neSources, &
     tolerance, tempSource, k)
    integer, intent(in) :: i
    integer(int32), intent(in) :: nSources
    integer(int32), intent(in) :: neSources
    type(SourceParameters), pointer :: source(:)
    type(SourceParameters), pointer :: esource(:)
    real(single), intent(in) :: tolerance

    integer(int8), intent(inout) :: k
    type(SourceParameters), pointer :: tempSource(:)
    integer :: j, je, l, minIndex, minIndexe    
    real(single) :: dx, dy, radius, minRadius = 1e6
    real(single) :: minRadiuse = 1e6
    logical, pointer :: found(:) => null()
    logical, pointer :: founde(:) => null()


    if( maxNumNeighbours == 0) then
        return
    end if

    allocate(found(1 : size(source)))
    if (neSources .gt. 0) then 
       allocate(founde(1 : size(esource)))
    endif

    found = .false.
    found(i) = .true.
    if (neSources .gt. 0) then 
       founde = .false.
    endif

    !if (i .eq. 2) then
    !   write(*,*) "L898 i=", i, " size(source)=", size(source), &
    !        " maxNumNeighbours=", maxNumNeighbours
    !   write(*,*) " coordinates  x0=", source(i)%xpos, " y0=", source(i)%ypos
    !endif

    do l = 1, maxNumNeighbours
        minIndex = -1
        minRadius = 1e6
        minIndexe = -1
        minRadiuse = 1e6

        do j = 1, size(source)
           !if (i .eq. 2) then
           !   write(*,*) "L907 j=", j, " raw=", source(j)%rawCountRate, &
           !        " x=", source(j)%xpos, " y=", source(j)%ypos, &
           !        " dx=", source(i)%xpos - source(j)%xpos, &
           !        " dy=", source(i)%ypos - source(j)%ypos
           !endif

            if( found(j)) cycle
            if (source(j)%extensionFlag == 1) cycle
            !if(source(j)%rawCountRate > maxRawCountRate) cycle
            
            dx = source(i)%xpos - source(j)%xpos
            if(dx > tolerance) cycle

            dy = source(i)%ypos - source(j)%ypos
            if(dy > tolerance) cycle

            radius = sqrt(dx * dx + dy * dy)

            if(radius <= tolerance) then
                if( radius .le. minRadius) then
                    minRadius = radius
                    minIndex = j
                
                end if
            end if    
        end do

        ! repeat the search among the extended sources
        if (neSources .gt. 0) then 
           do je = 1, size(esource)
              !if (i .eq. 2) then
              !   write(*,*) "L943 je=", je, " eraw=", esource(je)%rawCountRate, &
              !        " ex=", esource(je)%xpos, " ey=", esource(je)%ypos, &
              !        " edx=", source(i)%xpos - esource(je)%xpos, &
              !        " edy=", source(i)%ypos - esource(je)%ypos
              !endif

              if( founde(je)) cycle
              if (.not. esource(je)%extensionFlag == 1) cycle
              !if(source(j)%rawCountRate > maxRawCountRate) cycle
            
              dx = source(i)%xpos - esource(je)%xpos
              if(dx > tolerance) cycle

              dy = source(i)%ypos - esource(je)%ypos
              if(dy > tolerance) cycle

              radius = sqrt(dx * dx + dy * dy)

              if(radius <= tolerance) then
                 if( radius .le. minRadiuse) then
                    minRadiuse = radius
                    minIndexe = je                
                 end if
              end if
           end do
        endif

        if (minIndex > 0 .or. minIndexe > 0) then
           if(minIndex > 0) then
              k = k + 1
              found(minIndex) = .true.
              tempSource(k) = source(minIndex)
              if (source(minIndex)%rawCountRate<0) &
                   tempSource(k)%rawCountRate=-tempSource(k)%rawCountRate         
              if(k > maxNumNeighbours) then   
                 deallocate(found)
                 deallocate(founde)
                 return
              end if
           endif
           
           if (neSources .gt. 0) then
              if(minIndexe > 0) then
                 k = k + 1
                 founde(minIndexe) = .true.
                 tempSource(k) = esource(minIndexe)  
                 if (esource(minIndexe)%rawCountRate<0) &
                      tempSource(k)%rawCountRate=-tempSource(k)%rawCountRate       
                 if(k > maxNumNeighbours) then
                    deallocate(found)
                    deallocate(founde)
                    return
                 end if
              endif
           endif
        else
           exit
        end if

    end do

    deallocate(found)
    nullify(found)
   
    if (neSources .gt. 0) then
       deallocate(founde)
       nullify(founde)
    endif
    
end subroutine findNeighbours


recursive subroutine findNeighboursExtended( i, esource, source, tolerance, tempSource, k)
    integer, intent(in) :: i
    type(SourceParameters), pointer :: source(:)
    type(SourceParameters), pointer :: esource(:)
    real(single), intent(in) :: tolerance

    integer(int8), intent(inout) :: k
    type(SourceParameters), pointer :: tempSource(:)
    integer :: j, je, l, minIndex, minIndexe    
    real(single) :: dx, dy, radius, minRadius = 1e6
    real(single) :: minRadiuse = 1e6
    logical, pointer :: found(:) => null()
    logical, pointer :: founde(:) => null()


    if( maxNumNeighbours == 0) then
        return
    end if

    allocate(found(1 : size(source)))
    allocate(founde(1 : size(esource)))

    founde = .false.
    founde(i) = .true.
    found = .false.
 

    do l = 1, maxNumNeighbours
        minIndex = -1
        minRadius = 1e6
        minIndexe = -1
        minRadiuse = 1e6

        do j = 1, size(esource)
           !if (i .eq. 2) then
           !   write(*,*) "L907 j=", j, " raw=", source(j)%rawCountRate, &
           !        " x=", source(j)%xpos, " y=", source(j)%ypos, &
           !        " dx=", source(i)%xpos - source(j)%xpos, &
           !        " dy=", source(i)%ypos - source(j)%ypos
           !endif

           !write(*,*) "L1149 j=", j, " raw=", esource(j)%rawCountRate, &
           !     " err=",esource(j)%rawCountRateErr 

            if( founde(j)) cycle
            if (.not. esource(j)%extensionFlag == 1) cycle
            
            dx = esource(i)%xpos - esource(j)%xpos
            if(dx > tolerance) cycle

            dy = esource(i)%ypos - esource(j)%ypos
            if(dy > tolerance) cycle

            radius = sqrt(dx * dx + dy * dy)

            if(radius <= tolerance) then
                if( radius .le. minRadius) then
                    minRadiuse = radius
                    minIndexe = j
                
                end if
            end if    
        end do

        ! repeat the search among the point sources
        do je = 1, size(source)

            if( found(je)) cycle
            if (source(je)%extensionFlag == 1) cycle
            !if(source(j)%rawCountRate > maxRawCountRate) cycle
            
            dx = esource(i)%xpos - source(je)%xpos
            if(dx > tolerance) cycle

            dy = esource(i)%ypos - source(je)%ypos
            if(dy > tolerance) cycle

            radius = sqrt(dx * dx + dy * dy)

            if(radius <= tolerance) then
                if( radius .le. minRadiuse) then
                    minRadius = radius
                    minIndex = je                
                end if
            end if    
        end do

        if (minIndexe > 0 .or. minIndex > 0) then
           if(minIndexe > 0) then
              k = k + 1
              founde(minIndexe) = .true.
              tempSource(k) = esource(minIndexe)
              if (esource(minIndexe)%rawCountRate<0) &
                   tempSource(k)%rawCountRate=-tempSource(k)%rawCountRate
              if(k > maxNumNeighbours) then
                 deallocate(found)
                 deallocate(founde)
                 return
              end if
           endif

           if(minIndex > 0) then
              k = k + 1
              found(minIndex) = .true.
              tempSource(k) = source(minIndex)
              if (source(minIndex)%rawCountRate<0) &
                   tempSource(k)%rawCountRate=-tempSource(k)%rawCountRate
              if(k > maxNumNeighbours) then
                 deallocate(found)
                 deallocate(founde)
                 return
              end if
           endif
        else
            exit
        end if

    end do

    deallocate(found)
    nullify(found)
    deallocate(founde)
    nullify(founde)
    
end subroutine findNeighboursExtended





!*************************************************************************
!
! Least-squares to the PSF to include the background and an arbitrary
! number of linear constants
! ie y=constant+a1*x+a2*x, etc
! NO FIT TO THE COORDINATES
! nSources is the number of sources 
!
!************************************************************************
subroutine doLeastSquaresFit(nSources, x,               &
                             constant, constantError,   &
                             gradient, gradientError, &
                             isource, constant1, gradient1, ndata) 

    integer(int8), intent(in) :: nSources
    real(double)  :: x(:)
    real(double), intent(out), dimension(1 : nSources) :: gradient, gradientError
    real(double), intent(out) :: constant, constantError
    integer, optional, intent(out) :: nData

    integer, intent(in) :: isource

    integer :: i, j, k, l, n, k1, l1
    real(single) :: x1, y1, dx, dy, dx1, dy1
    real(single) :: radius, radius1
  
    logical :: printFlag
    real(double), dimension(1:2000,1:10) :: xSorted
    real(double), dimension(1:2000) :: ySorted
    real(double), dimension(1:2000) :: weightSorted
    real(double) :: aux, aux1, aux2, aux2a
    real(double) :: slope1
    real(double) :: yAver1, yAver2, xAver1, xAver2
    integer :: nsubset
    
    real(double), intent(out), dimension(1 : nSources) :: gradient1
    real(double), intent(out) :: constant1
    logical flagNeighbour
    logical flagAccepted
    real(single) :: maxRadius15, maxRadius075

    real(double) :: auxB, yAverB, sigmaB
    real(double) :: yMax
    real(double) :: nxaver2
    real(double) :: valueslope, sigmaSlope

    printFlag=.false.
    !if (isource .eq. 1) printFlag=.true.
    !if (isource .eq. 4) printFlag=.true.

    maxradius15=maxRadius*1.5
    maxRadius075=maxRadius*0.75
    if (printFlag) then
       write(*,*) "L2031 nSources=", nSources, " imin=", imin, " imax=", imax, &
            " jmin=", jmin, " jmax=", jmax
       write(*,*) " maxRadius15=", maxRadius15
       do i=1, nSources*2, 2
          write(*,*) "i=", i, " x(i)=", x(i), " x(i+1)=", x(i+1) 
       enddo
       write(*,*) "---------------------"
    endif
    !******************************************************************
    !
    ! Loop through all the pixels that are within the source area
    !
    !******************************************************************
    n=0
    flagAccepted=.true.

    do i = imin, imax
        x1 = real(i + 1)
        do j = jmin, jmax
            y1 = real(j + 1)

            !**********************************************************************
            !
            ! Get the PSF counts for this pixel
            !
            !**********************************************************************
            if (flagaccepted) then
               n=n+1
            endif

            flagNeighbour=.false.
            flagAccepted=.false.

            do k = 1, nSources ! nSources is the number of stars nearby
               !write(*,*) "L2101 i=", i, " j=", j, " k=", k, " n=", n
                xData(n, k) = 0.0  
                      
                l = (k  - 1) * 2 + 1
                dx = x1 - x(l)
         
                if(abs(dx) .gt. maxRadius15) then
                    cycle
                end if
                dy = y1 - x(l + 1)
             
                if(abs(dy) .gt. maxRadius15) then
                    cycle
                end if
  

                radius = sqrt(dx * dx + dy * dy)
                if(radius .gt. maxRadius15) then
                    cycle
                end if


                ! Check whether the point is affected by the neighbouring source
                flagNeighbour=.false.
                do k1=2, nSources
                   l1 = (k1  - 1) * 2 + 1
                   dx1 = x1 - x(l1)
                   dy1 = y1 - x(l1 + 1)


                   radius1 = sqrt(dx1 * dx1 + dy1 * dy1)
                   !if (printFlag) then
                   !   write(*,*) "  k1=", k1, " x1=",x(l1), " y1=",x(l1 + 1), " radius1=", radius1, flagNeighbour  
                   !endif
                   if (radius1 .lt. maxRadius) then
                      flagNeighbour=.true.
                   endif
                   !if (printFlag) then
                   !   write(*,*) "  new flagNeighbour=", flagNeighbour  
                   !endif

                enddo

              
                if (printFlag) then
                   !write(*,*) n,") i=", i, " j=", j, " dx=", dx, " dy=", dy, " xData=", xData(n, k)
                   !write(*,*) n,") i=", i, " j=", j, " dx=", dx, " dy=", dy, &
                   !     " image=", myimage(i,j), " flagNeighbour=", flagNeighbour
                   write(*,*) n,") i=", i, " j=", j, " radius=", radius, " image=", myimage(i,j), &
                        " xData=", PsfInterpolation(dx, dy), flagNeighbour
                endif
                if (.not. flagNeighbour) then
                   flagAccepted=.true.
                   xData(n, k) = PsfInterpolation(dx, dy)
                endif
                
            end do


            if (flagaccepted .and. .not. flagNeighbour) then
               
               
               !****************************************************
               ! Data point okay
               !*************************************************** 
               yData(n) = dble(myimage(i, j))
                
               !**************************************************************************
               ! NB Set all weights to 1, because according ti Mighell 1989 MNRAS, 238, 807
               ! this is the best weighting procedure
               !**************************************************************************
               weight(n) = weightingFunction( ydata(n) )  
               
            endif

        end do !   do i = imin, imax
    end do  ! do j = jmin, jmax
    n=n-1
    if (printFlag) then
       write(*,*) "L1351 n=", n
       do i=1,n
          !write(*,*) "i=", i, " xData=", xData(i,1), " yData=", yData(i), " weight=", weight(n)
          write(*,'(a,f12.8,a,f8.2,a)') "{",xData(i,1), ", ", yData(i), "},"
          !write(*,'(a,f12.8,a,f8.2,a)') "{",xData(i,1), ", ", log(yData(i)), "},"
       enddo
       write(*,*) "--------------"
    endif

    !-----------------------------------------------
    ! 2010-11-11 vny: test sorting the x-and y-data
    !-----------------------------------------------
    !if (printFlag) then


       ! compute the averages
       !aux1=0.0
       !aux2=0.0
       !xAver1=0.0
       !xAver2=0.0
       !yAver1=0.0
       !yAver2=0.0

       !nsubset=0
       !do i=1,n
       !   if (xData(i,1) .lt. 0.004) then
       !      aux1=aux1+1.0
       !      xAver1=xAver1+xData(i,1)
       !      yAver1=yAver1+yData(i)
       !   endif

       !   if (xData(i,1) .gt. 0.03) then
       !      aux2=aux2+1.0
       !      xAver2=xAver2+xData(i,1)
       !      yAver2=yAver2+ydata(i)
       !   endif
          
       !enddo
       !if (aux1 .gt. 1.0) then
       !   xAver1=xAver1/aux1
       !   yAver1=yAver1/aux1
       !endif
       !if (aux2 .gt. 1.0) then
       !   xAver2=xAver2/aux2
       !   yAver2=yAver2/aux2
       !endif

       !slope1=(yAver2-yAver1)/(xAver2-xAver1)

    !if (printFlag) then
    !   write(*,*) "L2029 xAver1=", xAver1, " yAver1=", yAver1
    !   write(*,*) "L2030 xAver2=", xAver2, " yAver2=", yAver2
    !   write(*,*) "L2031 slope1=", slope1
    !   write(*,*) "---------------"
    !endif

       !write(*,*) "L1395 nsubset=",nsubset
       !write(*,*) "-------------"
       !do i=1,nsubset
       !   !write(*,*) "i=", i, " xSorted=", xSorted(i,1), " ySorted=", ySorted(i)
       !   !write(*,'(a,f12.8,a,f8.2,a)') "{",xData(i,1), ", ", yData(i), "},"
       !   !write(*,'(a,f12.8,a,f8.2,a)') "{",xData(i,1), ", ", log(yData(i)), "},"
       !   write(*,'(a,f12.8,a,f8.2,a)') "{",xSorted(i,1), ", ", ySorted(i), "},"
       !enddo
       !write(*,*) "--------------"

       !call LeastSquaresFit21test(nsubset, xdata=xSorted, ydata=ySorted, &
       !     weight=weightSorted, nterms=int(nSources, int8),         &
       !     maxiter=maxIter, nstd1=3.0,                              &
       !     gradient=gradient, gradienterror=gradientError,          &
       !     constant=constant, constanterror=constantError, std1=std)
       !write(*,*) "L1422 subset constant=", constant, " gradient=", gradient
    !endif


    ! compute the averages
    aux1=0.0
    aux2=0.0
    aux2a=0.0
    xAver1=0.0
    xAver2=0.0
    yAver1=0.0
    yAver2=0.0
    auxB=0.0
    yAverB=0.0
    sigmaB=0.0
    nxaver2=0.0

    yMax=-99999.9
    do i=1,n
       if (yData(i) .gt. yMax) yMax=yData(i)
       if (xData(i,1) .lt. 0.001) then
          auxB=auxB+1.0
          yAverB=yAverB+yData(i)
       endif
    enddo
    if (auxB .gt. 1.0) then
       yAverB=yAverB/auxB
    endif
    auxB=0.0
    do i=1,n
       if (xData(i,1) .lt. 0.001) then
          auxB=auxB+1.0
          sigmaB=sigmaB+(yAverB-yData(i))*(yAverB-yData(i))
       endif
    enddo

    if (auxB .gt. 1) then
       sigmaB=sqrt(sigmaB/(auxB-1.0))
    endif

    if (printFlag) then
       write(*,*) "L2291 auxB=", auxB, " yAverB=", yAverB, " sigmaB=", sigmaB
       write(*,*) "L2308 yMax=", yMax
    endif
    if (auxB .gt. 1) sigmaB=sigmaB/sqrt(auxB)
    if (printFlag) then
       write(*,*) "L2291  sigmaB/sqrt(n)=", sigmaB
    endif

    nsubset=0
    do i=1,n
       !if (printFlag) then
       !   write(*,*) "L2283 i=", i, " xData(i,1)=", xData(i,1)
       !endif

       !if (xData(i,1) .lt. 0.004 .and. xData(i,1) .gt. 0.0) then
       if (xData(i,1) .lt. 0.004) then
       !if (yData(i) .gt. yAverB .and. yData(i)  .lt. (yAverB+6.0*sigmaB)) then
       !if (yData(i)  .lt. (yAverB+3.0*sigmaB)) then
       !if (yData(i)  .lt. yMax*0.01) then
          aux1=aux1+1.0
          xAver1=xAver1+xData(i,1)
          yAver1=yAver1+yData(i)
       endif

       !if (xData(i,1) .lt. 0.001) then
       !   auxB=auxB+1.0
       !   yAverB=yAverB+yData(i)
       !endif


       !if (xData(i,1) .gt. 0.01) then
       if (xData(i,1) .gt. 0.005) then
       !if (yData(i) .gt. yAverB+300.0*sigmaB) then
       !if (yData(i) .gt. yMax*0.01) then
          aux2=aux2+1.0
          xAver2=xAver2+xData(i,1)
          yAver2=yAver2+ydata(i)         
       endif
          
    enddo
    if (aux1 .gt. 1.0) then
       xAver1=xAver1/aux1
       yAver1=yAver1/aux1
    endif

    if (aux2 .gt. 1.0) then
       xAver2=xAver2/aux2
       yAver2=yAver2/aux2
    endif


    !if (auxB .gt. 1.0) then
    !   yAverB=yAverB/auxB
    !endif

   
    !slope1=(yAver2-yAver1)/(xAver2-xAver1)
    slope1=(yAver2-yAver1)/(xAver2-xAver1)

    if (printFlag) then
       write(*,*) "L2274 xAver1=", xAver1, " yAver1=", yAver1
       write(*,*) "L2275 xAver2=", xAver2, " yAver2=", yAver2
       write(*,*) "L2276 yAverB=", yAverB, " auxB=", auxB
       write(*,*) "L2277 slope1=", slope1
    endif

    ! compute dispersion relative to the slope
    aux=0.0
    sigmaSlope=0.0
    do i=1,n
       if (xData(i,1) .gt. xAver1 .and. xData(i,1) .lt. xaver2) then
          valueSlope=slope1*(xData(i,1)-xAver1)+yAver1
          aux=aux+1.0
          sigmaSlope=sigmaSlope+(valueSlope-yData(i))*(valueSlope-yData(i))
       endif
    enddo
    if (aux .gt. 1) sigmaSlope=sigmaSlope/(aux-1)
    sigmaSlope=sqrt(sigmaSlope)

    if (printFlag) then
       write(*,*) "L2437 slope1=", slope1, " sigmaSlope=", sigmaSlope
    endif
    if (aux .gt. 1) sigmaSlope=sigmaSlope/sqrt(aux)
    if (printFlag) then
       write(*,*) "L2461  sigmaSlope/sqrt(n)=", sigmaSlope
    endif

    !**********************************************************************************
    !
    ! Least-squares fit with constant (background) term
    !
    !**********************************************************************************  
!
!    call LeastSquaresFit21(n, xdata=xdata, ydata=ydata,              &
!            weight=weight, nterms=int(nSources, int8),               &
!            maxiter=maxIter, nstd1=3.0,                              &
!            gradient=gradient, gradienterror=gradientError,          &
!            constant=constant, constanterror=constantError, std1=std)
!
!    gradient1=gradient
!    constant1=constant


    gradient1=slope1
    constant1=yAverB

    ! test 2010-11-12
    !constant=yAver1
    constant=yAverB
    gradient(1)=slope1

    constantError=sigmaB
    gradientError=sigmaSlope

    !write(*,*) "L2113 gradient1=", gradient1(1), " new gradient=", gradient(1)
    !write(*,*) "L2116 constant1=", constant1, " new constant=", constant

    if(present(nData)) then
        nData = n
    end if

end subroutine doLeastSquaresFit


!****************************************************
! Subroutine RealImage
!****************************************************
subroutine SaveImageTest2(image, i1,i2,j1,j2, fileName, heading)

    !integer, intent(in) :: nx, ny    
    integer, intent(in) :: i1, i2,j1,j2   

    !real (single), intent(in), dimension(0 : nx - 1, 0 : ny - 1) :: image
    real (single), intent(in), dimension(i1 : i2, j1 : j2) :: image
    character(*), intent(in) :: fileName
    character(*), intent(in) :: heading
    type(DataSetT)       :: imageSet 
    real (single), dimension(:, :), pointer :: imageP
    type(ArrayT)         :: imageArray         ! array for background
   
 !write(*, *) "omutility_mod saverealimage1 line 1300, saving to file ", trim(fileName)
    imageSet = DataSet(trim(fileName), Create)

    if (hasBlock(imageSet, "PRIMARY")) then
        call deleteBlock( imageSet, "PRIMARY" );
    end if
    !if (heading .eq. "test") then
!
!    end if  

    imageArray = addArray(imageSet, "PRIMARY", real32, shape(image))
    imageP => real32Array2Data(imageArray)
    imageP = image
    call release(imageArray)
    call release(imageSet)   
end subroutine SaveImageTest2

!**************************************************************
! Deblend the source from the neigbour sources
!***************************************************************
subroutine deblendSource(nSources, x, maxSrcRecalculated, extractionRadius, maxRadius1, &
                             deblendedCounts, deblendedCounts3, deblendedX, deblendedY, &
                             constant, constantError,   &
                             gradient, gradientError, &
                             isource, constant1, gradient1, ndata) 

    integer(int8), intent(in) :: nSources
    real(double)  :: x(:)
    real(double), intent(out), dimension(1 : nSources) :: gradient, gradientError
    !-----------------
    real(double), intent(out), dimension(1 : nSources) :: maxSrcRecalculated
    !-----------------
    real(double), intent(out) :: constant, constantError
    integer, optional, intent(out) :: nData

    real(double), intent(out) :: deblendedCounts, deblendedCounts3
    real(double), intent(out) :: deblendedX, deblendedY

    real(single):: totalCounts

    integer, intent(in) :: isource

    integer :: i, j, k, l, n, k1, l1, ii,jj
    real(single) :: x1, y1, dx, dy, dx1, dy1, dx2, dy2
    real(single) :: radius, radius1, radiusPsf
  
    real(single), intent(in) :: extractionRadius, maxRadius1
    real(single) :: extractionRadius10, maxRadius10

    logical :: printFlag
    real(double), dimension(1:2000,1:10) :: xSorted
    real(double), dimension(1:2000) :: ySorted
    real(double), dimension(1:2000) :: weightSorted
    real(double) :: aux, aux1, aux2, aux2a
    real(double) :: slope1
    real(double) :: yAver1, yAver2, xAver1, xAver2
    integer :: nsubset
    
    real(double), intent(out), dimension(1 : nSources) :: gradient1
    real(double), intent(out) :: constant1
    logical flagNeighbour
    logical flagAccepted
    real(single) :: maxRadius15, maxRadius075

    real(double) :: auxB, yAverB, sigmaB
    real(double) :: auxB2, yAverB2, sigmaB2

    real(double) :: yMax, yMin
    real(double) :: nxaver2
    real(double) :: valueslope, sigmaSlope

    real(double), pointer, dimension(:, :) :: ratioImage => null()
    real(double) :: sumSrcPsf0, sumSrcPsf, sumSrcPsf2


    !*****************************************
    ! 2011-11-14
    ! a higher-resolution image for each source
    real(single), pointer :: psfImageTest(:, :) => null()
    integer :: iin,jjn
    integer :: nxmin, nxmax, nymin, nymax
    integer :: imax2, imin2, jmax2, jmin2
    integer :: imax2lim, imin2lim, jmax2lim, jmin2lim
    integer :: imax2b, imin2b, jmax2b, jmin2b
    integer :: i1r10, j1r10, i2r10, j2r10
    integer :: i1r5, j1r5, i2r5, j2r5
    real(single) :: xc, yc, xc10, yc10, xc10r, yc10r
    real(single):: totalCounts10
    real(single):: totalCountsAper3,totalCountsAper6
    real(single) :: sumx, sumy,  nsumx, nsumy
    real(single) :: pminValue
    real(double) :: xDataMin, xDataMax
    real(double) :: bkgThresh1  
    real(double) :: sumPsf010, sumPsf005, auxPsf, ratioPsf
    !*****************************************

    printFlag=.false.
    !if (isource .eq. 1) printFlag=.true.
    !if (isource .eq. 12) printFlag=.true.
    !if (isource .eq. 2) printFlag=.true.
    !if (isource .eq. 4) printFlag=.true.
    !if (isource .eq. 45) printFlag=.true.
    !if (isource .eq. 6) printFlag=.true.
    !if (isource .eq. 9) printFlag=.true.
    !if (isource .eq. 12) printFlag=.true.
    !if (isource .eq. 28) printFlag=.true.
    !if (isource .eq. 51) printFlag=.true.
    !if (iSource .eq. 25) printFlag=.true.
    totalCounts=0.0

    deblendedCounts=0.0
    deblendedX=0.0
    deblendedY=0.0

    maxradius15=extractionRadius*1.5
    maxRadius075=maxRadius*0.75

    nxmin = lbound(myImage, 1)
    nxmax = ubound(myImage, 1)

    nymin = lbound(myImage, 2)
    nymax = ubound(myImage, 2)

    if (printFlag) then
       write(*,*) "L2596 nSources=", nSources, " imin=", imin, " imax=", imax, &
            " jmin=", jmin, " jmax=", jmax, " nxmin=", nxmin, " nxmax=", nxmax, &
            " nymin=", nymin, " nymax=", nymax

       write(*,*) " maxRadius15=", maxRadius15
    endif

    xc=x(1)
    yc=x(2)

   if (printFlag) then
       do i=1, nSources*2, 2
          ii=i/2+1
          write(*,*) "L2605 i=", i, " x(i)=", x(i), " x(i+1)=", x(i+1), " max=", maxSrcRecalculated(ii) 
       enddo
       write(*,*) "---------------------"
    endif



    !******************************************************************
    !
    ! Loop through all the pixels that are within the source area
    !
    !******************************************************************
    n=0
    flagAccepted=.true.

    

    !-------------------------------------------------------
    ! An auxiliary image to create a higher-resolution image
    ! in order to extract sub-pixels counts
    !--------------------------------------------------------
    imax2=imax
    imin2=imin
    jmax2=jmax
    jmin2=jmin

    imax2b=imax
    imin2b=imin
    jmax2b=jmax
    jmin2b=jmin

    iin=(imax-imin)*10
    jjn=(jmax-jmin)*10


    xc10=(xc-real(imin))
    yc10=(yc-real(jmin))

    if (printFlag) then
       write(*,*) "L2698 imin=", imin, " imax=", imax, " jmin=", jmin, " jmax=", jmax
       write(*,*) "L2699 xc=",xc, " yc=", yc, " xc10=", xc10, " yc10=", yc10  
    endif

    xc10=xc10*10.0
    yc10=yc10*10.0

    if (printFlag) then
       write(*,*) "L2706  xc10=", xc10, " yc10=", yc10  
    endif

    if ((imax-imin) .lt. int(2.0*maxRadius1)) then
       ii=(int(2.0*maxRadius1+0.5)-(imax-imin))/2
       imax2=imax2+ii
       imin2=imin2-ii
       xc10=(xc-real(imin2)+0.5)*10.0
       iin=(imax2-imin2)*10
    endif
    if ((jmax-jmin) .lt. int(2.0*maxRadius1)) then
       jj=(int(2.0*maxRadius1+0.5)-(jmax-jmin))/2
       jmax2=jmax2+jj
       jmin2=jmin2-jj
       yc10=(yc-real(jmin2)+0.5)*10.0
       jjn=(jmax2-jmin2)*10
    endif
    if (printFlag) then
       write(*,*) "L2657  xc10=", xc10, " yc10=", yc10  
    endif

    if ((imax-imin) .lt. int(2.0*maxRadius15)) then
       ii=(int(2.0*maxRadius15+0.5)-(imax-imin))/2
       imax2b=imax2b+ii
       imin2b=imin2b-ii
    endif
    if ((jmax-jmin) .lt. int(2.0*maxRadius15)) then
       jj=(int(2.0*maxRadius15+0.5)-(jmax-jmin))/2
       jmax2b=jmax2b+jj
       jmin2b=jmin2b-jj
    endif


    imin2lim=imin2
    if (imin2lim .lt. nxmin) imin2lim=nxmin
    jmin2lim=jmin2
    if (jmin2lim .lt. nymin) jmin2lim=nymin
    imax2lim=imax2
    if (imax2lim .gt. nxmax) imax2lim=nxmax
    jmax2lim=jmax2
    if (jmax2lim .gt. nymax) jmax2lim=nymax

    maxRadius10=maxRadius1*10.0
    extractionRadius10=extractionRadius*10.0

    ! The image containing the ratio between the PSFs of the main and molesting source
    ratioImage => allocate2dDouble(imin2, imax2, jmin2, jmax2, "ratioimage")

    ratioImage=-1.0

    if (printFlag) then
       write(*,*) "L2781 imin2=", imin2, " imax2=", imax2, " jmin2=", jmin2, &
            " jmax2=", jmax2, " xc10=", xc10, " yc10=", yc10
       write(*,*) "L2783 imin2b=", imin2b, " imax2b=", imax2b, " jmin2b=", jmin2b, &
            " jmax2b=", jmax2b
       
    endif

    psfImageTest => allocate2dSingle(0, iin, 0, jjn, "psfimagetest")
    psfImageTest=-1.0
    !--------------------------------------------------------


    !*******************************************
    ! Expand the image aroud the source to increase the accuracy
    ! of the cumulative count curve extraction
    !******************************************************
    do ii=0,iin
       do jj=0,jjn
          i=ii/10+imin2
          j=jj/10+jmin2

          if (i .ge. nxmin .and. i .le. nxmax .and. j .ge. nymin .and. j .le. nymax) then
             psfImageTest(ii,jj)= myimage(i, j)
          endif

       enddo
    enddo

    if (printFlag) then
       call SaveImageTest2(psfImageTest, 0, iin, 0, jjn, "psfprofiletest0.fit", "test")
    endif

    if (printFlag) then
       write(*,*) "L2815 maxRadius1=", maxRadius1, " imin2=", imin2, " imax2=", imax2, &
            " jmin2=", jmin2, " jmax2=", jmax2
    endif

    !*******************************
    ! Calculate the ration of the PSF withing the 
    ! half-to-full extraction radius
    ! 
    sumPsf005=0.0
    sumPsf010=0.0
    do i = imin2, imax2
       x1 = real(i)
       do j = jmin2, jmax2
          y1 = real(j)
          dx = x1 - x(1)
          dy = y1 - x(2)
          radiusPsf = sqrt(dx * dx + dy * dy)
          auxPsf=PsfInterpolation(dx, dy) 
          if (radiusPsf .lt. 0.4*extractionRadius) then
             sumPsf005=sumPsf005+auxPsf
          endif
          if (radiusPsf .lt. extractionRadius) then
             sumPsf010=sumPsf010+auxPsf
          endif
       enddo
    enddo
    ratioPsf=sumPsf010/sumPsf005
    if (printFlag) then
       write(*,*) "L3089 sumPsf010=", sumPsf010, " sumPsf005=", sumPsf005, &
            " ratioPsf=", ratioPsf
    endif
    do i = imin2lim, imax2lim
    !do i = imin2b, imax2b
        !x1 = real(i + 1)
        x1 = real(i)
        do j = jmin2lim, jmax2lim
        !do j = jmin2b, jmax2b
            !y1 = real(j + 1)
            y1 = real(j)

            !ratioImage(i,j)=1.0
            ratioImage(i,j)=0.0


            !**********************************************************************
            !
            ! Get the PSF counts for this pixel
            !
            !**********************************************************************
            if (flagaccepted) then
               n=n+1
            endif

            !if (printFlag) then
            !   write(*,*) "L2784 imin=", imin, " imax=", imax, " imin2=", imin2, &
            !        " imax2=", imax2
            !   write(*,*) "L2786 jmin=", jmin, " jmax=", jmax, " jmin2=", jmin2, &
            !        " jmax2=", jmax2
            !endif

            flagNeighbour=.false.
            flagAccepted=.false.

            sumSrcPsf0=0.0
            sumSrcPsf=0.0
            sumSrcPsf2=0.0

            do k = 1, nSources ! nSources is the number of stars nearby
               !write(*,*) "L2101 i=", i, " j=", j, " k=", k, " n=", n, " nSources=", nSources
                xData(n, k) = 0.0       
                xData2(n, k) = 0.0 
                      
                l = (k  - 1) * 2 + 1
                dx = x1 - x(l)
         
                !if(abs(dx) .gt. maxRadius1) then
                !    cycle
                !end if

                dy = y1 - x(l + 1)
             
                !if (printFlag .and. n .lt. 35) then
                !   write(*,*) "L2622 (source)k=", k, " x=",x(l), " y=", x(l + 1), &
                !        " dx=", dx, " dy=", dy, " maxRadius15=",maxRadius15, &
                !        " maxRadius1=",maxRadius1, " extractionRadius=", extractionRadius 
                !endif


                !if(abs(dy) .gt. maxRadius1) then
                !    cycle
                !end if
  

                radius = sqrt(dx * dx + dy * dy)

                !if(radius .gt. maxRadius1) then
                !    cycle
                !end if
                if (printFlag .and. i .eq. 77 .and. j .eq. 85) then
                   write(*,*) "L2901 i=", i, " j=", j, " radius=", radius, &
                        " dx=", dx, " dy=", dy, " psf=",PsfInterpolation(dx, dy) 
                endif

                !if (i .lt. imin .or. i .gt. imax .or. &
                !     j .lt. jmin .or. j .gt. jmax) then
                !   cycle
                !endif

                if (printFlag .and. i .eq. 77 .and. j .eq. 85) then
                   write(*,*) "L2911 i=",i, " j=", j
                endif
                !---------------------------------------------------
                ! psf of the central source for the pixel number n
                ! Calculate the contribution of this source to the pixel
                ! counts
                !---------------------------------------------------
                xData2(n,k)=PsfInterpolation(dx, dy)
                xData(n,k)=PsfInterpolation(dx, dy)

                !******************************************
                !if (radius .lt. extractionRadius*0.3) xData(n,k)=1.0
                !******************************************

                if (radius .lt. extractionRadius) then

                   ! the sum of PSFs from all sources in this pixel
                   sumSrcPsf=sumSrcPsf+xData2(n,k)*maxSrcRecalculated(k)

                   ! sumSrcPsf0 contains the counts from the main source only
                   sumSrcPsf0=sumSrcPsf0+xData2(n,k)*maxSrcRecalculated(k)

                   ! the value of the main source in the current pixel
                   !ratioImage(i,j)=xData2(n,k)*maxSrcRecalculated(k) 
                else
                   ratioImage(i,j)=0.0
                endif

                if (printFlag .and. i .eq. 77 .and. j .eq. 85) then
                   write(*,*) "L2931 i=", i, " j=", j, " sumSrcPsf=", sumSrcPsf, &
                        " sumSrcPsf0=", sumSrcPsf0
                endif

                ! Check whether the point is affected by the neighbouring source
                flagNeighbour=.false.
                do k1=2, nSources
                   l1 = (k1  - 1) * 2 + 1
                   dx1 = x1 - x(l1)
                   dy1 = y1 - x(l1 + 1)

                   radius1 = sqrt(dx1 * dx1 + dy1 * dy1)


                   dx2=dx1
                   dy2=dy1

                   !dx2=dx1-0.01*extractionRadius/6.0
                   !if (dx2 .lt. 0.) dx2=0.0
                   !dy2=dy1-0.01*extractionRadius/6.0
                   !if (dy2 .lt. 0.) dy2=0.0

                   ! psf of the molesting source for the current pixel
                   xData2(n,k1)=PsfInterpolation(dx2, dy2)

                   !*******************************************************
                   ! 2011-11-17 Adjust the counts of molesting sources
                   if (radius1 .le. extractionRadius*0.5 &
                        .and. .not. radius .lt. extractionRadius*0.5) xData2(n,k1)= 0.3
                   if (radius1 .le. extractionRadius*0.3) xData2(n,k1)= 1.0
                   if (radius1 .le. extractionRadius*0.5 .and. &
                        radius .gt. extractionRadius*0.8) xData2(n,k1)= 1.0
                   !**********************************************************

                   !--------------------------------------
                   ! Exclude the influence of the neighbour if this neighbour
                   ! is far from the main source
                   !---------------------------------------
                   if (radius1 .lt. extractionRadius*0.5) then
                      ! sumSrcPsf contains the contribution from all of the sources,
                      ! including the central one
                      sumSrcPsf=sumSrcPsf+xData2(n,k1)*maxSrcRecalculated(k1)

                      ! sumSrcPsf2 contains only the contribution from neighbours
                      sumSrcPsf2=sumSrcPsf2+xData2(n,k1)*maxSrcRecalculated(k1)
                   endif

                   if (printFlag .and. i .eq. 108 .and. j .eq. 156) then
                      write(*,*) "L2960 i=", i, " j=", j, " sumSrcPsf=", sumSrcPsf, &
                           " sumSrcPsf2=", sumSrcPsf2, " sumSrcPsf0=", sumSrcPsf0   
                   endif

                   !if (printFlag .and. n .lt. 35) then
                   !   write(*,*) "  k1=", k1, " x1=",x(l1), " y1=",x(l1 + 1), " radius1=", radius1, flagNeighbour  
                   !endif

                   !if (radius1 .lt. maxRadius15 .and. xData2(n,k1) .gt. 0.0005) then
                   if (radius1 .lt. extractionRadius) then
                      flagNeighbour=.true.
                   endif
                   !if (printFlag .and. n .lt. 35) then
                   !   write(*,*) "  new flagNeighbour=", flagNeighbour  
                   !endif

                enddo
                 

                if (printFlag .and. i .eq. 108 .and. j .eq. 156) then
                   write(*,*) "L2977 i=", i, " j=", j, " sumSrcPsf=", sumSrcPsf, &
                        " sumSrcPsf2=", sumSrcPsf2
                   !ratioImage(i,j)=sumSrcPsf-sumsrcPsf2
                   ratioImage(i,j)=sumSrcPsf0
                   write(*,*) "L2980 ratiImage=",ratioImage(i,j)
                   ratioImage(i,j)=ratioImage(i,j)/sumSrcPsf
                   write(*,*) "L2964 after division ratiImage=",ratioImage(i,j)
                endif

                !-----------------------------------
                ! When the contribution from all of the molesting sources
                ! is calculated, compute the ratio of the main source to all
                !----------------------------------
                !if (i .ge. imin .and. i .le. imax .and. &
                !     j .ge. jmin .and. j .le. jmax) then
                if (i .ge. imin2lim .and. i .le. imax2lim .and. &
                     j .ge. jmin2lim .and. j .le. jmax2lim) then

                   if (sumSrcPsf .gt. 0.) then
                      !ratioImage(i,j)=sumSrcPsf-sumsrcPsf2
                      !ratioImage(i,j)=ratioImage(i,j)/sumSrcPsf  
                      !!ratioImage(i,j)=ratioImage(i,j)/(sumSrcPsf+sumSrcPsf2)
                      ratioImage(i,j)=sumSrcPsf0/sumSrcPsf
                   else
                      ratioImage(i,j)=0.0!1.0
                   endif

                   !if (printFlag) then
                   !   if (i .gt. 137 .and. i .lt. 141 .and. j .gt. 126 &
                   !        .and. j .lt. 130) then
                   !      write(*,*) "L3262 i=", i, " j=", j, " ratioImage=", &
                   !           ratioImage(i,j), " sumSrcPsf0=", sumSrcPsf0, &
                   !           " sumSrcPsf=", sumSrcPsf 
                   !   endif
                   !endif
                   if (ratioImage(i,j) .gt. 1.0) ratioImage(i,j)=1.0
                   if (ratioImage(i,j) .lt. 0.0) ratioImage(i,j)=0.0
                   !******************************************************
                   ! 2011-11-19 adjustment of the flux
                   !*******************************************************
                   !if (radius .lt. extractionRadius*0.18) ratioImage(i,j)=1.0
                   if (radius .lt. extractionRadius*0.4) ratioImage(i,j)=1.0
                   if (radius1 .lt. extractionRadius*0.3) ratioImage(i,j)=0.0
                   !*******************************************************

                endif

                !if (printFlag .and. n .lt. 35) then
                !   !write(*,*) n,") i=", i, " j=", j, " dx=", dx, " dy=", dy, " xData=", xData(n, k)
                !   !write(*,*) n,") i=", i, " j=", j, " dx=", dx, " dy=", dy, &
                !   !     " image=", myimage(i,j), " flagNeighbour=", flagNeighbour
                !   write(*,*) "n=", n,") i=", i, " j=", j, " radius=", radius, " image=", myimage(i,j), &
                !        " xData=", PsfInterpolation(dx, dy), flagNeighbour
                !endif
                if (radius .le. maxRadius1) then
                   if (.not. flagNeighbour) then
                      flagAccepted=.true.
                      xData(n, k) = PsfInterpolation(dx, dy)
                   endif
                endif
            end do

            !if (printFlag .and. n .eq. 35) then
            !   write(*,*) "---"
            !   do k = 1, nSources
            !      write(*,*) "L2673 n=", n, " k=", k, " xData2(n,k)=", xData2(n, k), &
            !           " max=", maxSrcRecalculated(k) 
            !   enddo
            !   write(*,*) "---"
            !endif

            if (flagaccepted .and. .not. flagNeighbour) then
               
               
               !****************************************************
               ! Data point okay
               !*************************************************** 
               !if (i .ge. imin .and. i .le. imax .and. &
               !     j .ge. jmin .and. j .le. jmax) then
               if (i .ge. nxmin .and. i .le. nxmax .and. &
                    j .ge. nymin .and. j .le. nymax) then
                  yData(n) = dble(myimage(i, j))
               
                
                  !**************************************************************************
                  ! NB Set all weights to 1, because according ti Mighell 1989 MNRAS, 238, 807
                  ! this is the best weighting procedure
                  !**************************************************************************
                  weight(n) = weightingFunction( ydata(n) )
               endif
               
            endif

        end do !   do i = imin2lim, imax2lim

    end do  ! do j = jmin2lim, jmax2lim

    n=n-1


    ! compute the averages
    aux1=0.0
    aux2=0.0
    aux2a=0.0
    xAver1=0.0
    xAver2=0.0
    yAver1=0.0
    yAver2=0.0
    auxB=0.0
    yAverB=0.0
    sigmaB=0.0
    nxaver2=0.0

    if (printFlag) then
       write(*,*) "-------"
       do i=1,n
          write(*,*) "L3381 i=", i, " xData=", xData(i,1), " yData=", yData(i)
       enddo
       write(*,*) "-------"
    endif

    ! Check  the min and max values of xData
    xDataMin=99999.9
    xDataMax=-99999.9
    yMax=-99999.9
    yMin=+9.9E18
    auxB=0.0
    yAverB=0.0
    do i=1,n
       if (xData(i,1) .lt. 0.0015) then
          auxB=auxB+1.0
          yAverB=yAverB+yData(i)
       endif

       if (xData(i,1) .lt. xDataMin) then
          xDataMin= xData(i,1)
       endif
       if (xData(i,1) .gt. xDataMax) then
          xDataMax= xData(i,1)
       endif
       if (yData(i) .gt. yMax) then
          yMax= yData(i)
       endif
       if (yData(i) .lt. yMin) then
          yMin= yData(i)
       endif
    enddo
    if (auxB .gt. 1.0) then
       yAverB=yAverB/auxB
    endif
    if (printFlag) then
       write(*,*) "L3420 initial yAverB=", yAverB
    endif

    bkgThresh1=yAverB+0.1*yAverB

    if (printFlag) then
       write(*,*) "L3408 xDataMin=", xDataMin, " xDataMax=", xDataMax, " n=", n
       write(*,*) "L3409 yMin=", yMin, " yMax=", yMax, " bkgThresh1=", bkgThresh1
    endif

    auxB=0.0
    yAverB=0.0
    do i=1,n 
       if (xData(i,1) .lt. 0.0015 .and. yData(i) .lt. bkgThresh1) then
          auxB=auxB+1.0
          yAverB=yAverB+yData(i)
       endif
    enddo

    if (auxB .gt. 1.0) then
       yAverB=yAverB/auxB
    endif

    auxB=0.0
    do i=1,n
       if (xData(i,1) .lt. 0.0015 .and.  yData(i) .lt. bkgThresh1) then
          auxB=auxB+1.0
          sigmaB=sigmaB+(yAverB-yData(i))*(yAverB-yData(i))
       endif
    enddo

    if (auxB .gt. 1) then
       sigmaB=sqrt(sigmaB/(auxB-1.0))
    endif

    if (printFlag) then
       write(*,*) "L3438 auxB=", auxB, " yAverB=", yAverB, " sigmaB=", sigmaB
       write(*,*) "L3439 yMax=", yMax
    endif
    if (auxB .gt. 1) sigmaB=sigmaB/sqrt(auxB)
    if (printFlag) then
       write(*,*) "L3443  sigmaB/sqrt(n)=", sigmaB
    endif

    !------------------------------------------------------
    ! NB
    ! repeat the background calculation with rejected values
    !------------------------------------------------------
    yAverB2=0.0
    auxB2=0.0
    do i=1,n
       if (xData(i,1) .lt. 0.0015 .and.  yData(i) .lt. bkgThresh1) then
          if (abs(yData(i)-yAverB) .lt. sigmaB) then
             auxB2=auxB2+1.0
             yAverB2=yAverB2+yData(i)
          endif
       endif
    enddo
    if (printFlag) then
       write(*,*) "L3452 yAverB2=", yAverB2, " auxB2=", auxB2
    endif

    if (auxB2 .gt. 1.0) then
       yAverB2=yAverB2/auxB2
    endif

    sigmaB2=0.0
    auxB2=0.0
    do i=1,n
       if (xData(i,1) .lt. 0.0015 .and.  yData(i) .lt. bkgThresh1) then
          if (abs(yData(i)-yAverB) .lt. sigmaB) then
             auxB2=auxB2+1.0
             sigmaB2=sigmaB+(yAverB2-yData(i))*(yAverB2-yData(i))
          endif
       endif
    enddo
    if (auxB2 .gt. 1) then
       sigmaB2=sqrt(sigmaB2/(auxB2-1.0))
    endif

    if (printFlag) then
       write(*,*) "L3472 yAverB2=", yAverB2, " sigmaB2=", sigmaB2
    endif

    if (auxB2 .ge. 1.0) then
       yAverB=yAverB2
    endif
    if (printFlag) then
       write(*,*) "L3480 yAverB=", yAverB
    endif

    if (printFlag) then
       write(*,*) "L2806 n=", n
       !do i=1,n
       !   write(*,'(a,f12.8,a,f8.2,a)') "{",xData(i,1), ", ", yData(i), "},"
       !enddo
       !write(*,*) "--------------"
       !do i = imin, imax
       !   do j = jmin, jmax
       !      write(*,*) "i=", i, " j=",j, " ratioImage=", ratioImage(i,j)
       !   enddo
       !enddo
       !write(*,*) "L2816 --------------"
    endif

    !***********************************************************
    ! Compute the total counts above the background (yAverConst)
    !***********************************************************
    totalCounts=0.0
    totalCountsAper3=0.0
    totalCountsAper6=0.0
    do i = imin2lim, imax2lim
       !x1 = real(i + 1)
       x1 = real(i)
       do j = jmin2lim, jmax2lim
          !y1 = real(j + 1)
          y1 = real(j)
           dx = x1 - x(1)
           !if(abs(dx) .gt. extractionRadius) then
           !   cycle
           !end if
           dy = y1 - x(2)
           !if(abs(dy) .gt. extractionRadius) then
           !   cycle
           !end if
           radius = sqrt(dx * dx + dy * dy)
           if(radius .gt. extractionRadius) then
              cycle
           end if
           totalCounts=totalCounts+(myimage(i,j)-yAverB)*ratioImage(i,j)
    
           !*****************************************
           ! Very strange: when changing from the whole
           ! extraction radius to 0.5, the totalCountsAper3 is increasing
           ! instead of decreasing
           !-----------------------------------------
           if (radius .le. extractionRadius*0.4) then
                 totalCountsAper3=totalCountsAper3+(myimage(i,j)-yAverB)
           endif
           if (radius .le. extractionRadius) then
                 totalCountsAper6=totalCountsAper6+(myimage(i,j)-yAverB)
           endif
       enddo
    enddo
    if (printFlag) then
    write(*,*) "L3112 totalCounts=", totalCounts, " totalCountsAper3=", &
         totalCountsAper3, " totalCountsAper6=",totalCountsAper6," yAverB=", yAverB
    endif

    !***********************************************************
    ! Compute the total counts above the background (yAverConst)
    ! using the HiRes image, but first re-calculate the x- and y-coordinates
    !***********************************************************
    i1r10=int(xc10-extractionRadius10-0.5)
    i2r10=int(xc10+extractionRadius10+0.5)
    j1r10=int(yc10-extractionRadius10-0.5)
    j2r10=int(yc10+extractionRadius10+0.5)

    i1r5=int(xc10-extractionRadius10/2.0-0.5)
    i2r5=int(xc10+extractionRadius10/2.0+0.5)
    j1r5=int(yc10-extractionRadius10/2.0-0.5)
    j2r5=int(yc10+extractionRadius10/2.0+0.5)

    if (printFlag) then
       write(*,*) "L2926 i1r10=", i1r10, " i2r10=", i2r10, " j1r10=", j1r10, &
            " j2r10=", j2r10
       write(*,*) "L2928 xc10=", xc10, " yc10=", yc10, " extractionradius10=", &
            extractionRadius10
    endif


    !********************************
    ! re-calculate x- and y-coordinates
    !********************************
    xc10r=-1.0
    yc10r=-1.0
    sumx=0.0
    sumy=0.0
    nsumx=0.0
    nsumy=0.0
    pminValue=9.9E13
!    do ii = i1r5, i2r5
!       if (ii .ge. 0 .and. ii .le. iin) then
!          do jj = j1r5, j2r5
!             if (jj .ge. 0 .and. jj .le. jjn) then
!    
!             endif
!          enddo
!       endif
!    enddo

    totalCountsAper3=0.0
    totalCountsAper6=0.0

    do ii = i1r10, i2r10
       if (ii .ge. 0 .and. ii .le. iin) then
          x1 = real(ii)
          do jj = j1r10, j2r10
             if (jj .ge. 0 .and. jj .le. jjn) then    
               y1 = real(jj)
                dx = x1 - xc10  
                if(abs(dx) .gt. extractionRadius10) then
                   cycle
                end if
                dy = y1 - yc10
                if(abs(dy) .gt. extractionRadius10) then
                   cycle
                end if
                radius = sqrt(dx * dx + dy * dy)
                if(radius .gt. extractionRadius10) then
                   cycle
                end if

                if (psfImageTest(ii, jj) .ge. 0.0) then
                   i=ii/10+imin2
                   j=jj/10+jmin2
                   
                   if (i .ge. nxmin .and. i .le. nxmax .and. &
                        j .ge. nymin .and. j .le. nymax) then
                      sumx=sumx+x1*((psfImageTest(ii, jj)-yAverB)*ratioImage(i,j))
                      sumy=sumy+y1*((psfImageTest(ii, jj)-yAverB)*ratioImage(i,j))
                      nsumx=nsumx+(psfImageTest(ii, jj)-yAverB)*ratioImage(i,j)
                   endif
                   if(radius .le. extractionRadius10) then
                      totalCountsAper6=totalCountsAper6+(psfImageTest(ii, jj)-yAverB)
                   endif
                   if(radius .le. 0.4*extractionRadius10) then
                      totalCountsAper3=totalCountsAper3+(psfImageTest(ii, jj)-yAverB)
                   endif

                endif

             endif
          enddo
       endif
    enddo
    if (nsumx .gt. 1.0) then
       xc10r=sumx/nsumx
       yc10r=sumy/nsumx
    endif
    deblendedX=xc10r/10.0+real(imin2)-0.5+1.0
    deblendedY=yc10r/10.0+real(jmin2)-0.5+1.0
    if (printFlag) then
       write(*,*) "L3589 xc10r=", xc10r, " yc10r=", yc10r, &
            " deblendedX=", deblendedX, " deblendedY=", deblendedY
    endif
    totalCountsAper3=totalCountsAper3/100.0
    totalCountsAper6=totalCountsAper6/100.0
    if (printFlag) then
       write(*,*) "L3595 totalCountsAper3=", totalCountsAper3, &
            " totalCountsAper6=", totalCountsAper6, " ratioPsf=", ratioPsf
    endif

    totalCountsAper3=totalCountsAper3*ratioPsf
    if (printFlag) then
       write(*,*) "L3675 totalCountsAper3=", totalCountsAper3
    endif

    totalCounts10=0.0
    do ii = i1r10, i2r10
       if (ii .ge. 0 .and. ii .le. iin) then
          x1 = real(ii)
          do jj = j1r10, j2r10
             if (jj .ge. 0 .and. jj .le. jjn) then
                y1 = real(jj)
                !dx = x1 - xc10 
                dx = x1 - xc10r               
 
               if (printFlag .and. ii .eq. 115 .and. jj .eq. 56) then
                   write(*,*) "L3223 ii=", ii, " jj=", jj, " i=", i, " j=", j
                   write(*,*) "L3224 psfImageTest=", psfImageTest(ii, jj), &
                        " ratioImage=",ratioImage(i,j) 
                endif

               if(abs(dx) .gt. extractionRadius10) then
                   cycle
                end if
                !dy = y1 - yc10
                dy = y1 - yc10r
                if(abs(dy) .gt. extractionRadius10) then
                   cycle
                end if
                radius = sqrt(dx * dx + dy * dy)
                if(radius .gt. extractionRadius10) then
                   cycle
                end if
 
                if (psfImageTest(ii, jj) .ge. 0.0) then
                   i=ii/10+imin2
                   j=jj/10+jmin2

                   !if (printFlag .and. ii .eq. 105 .and. jj .eq. 95) then
                   !if (printFlag) then
                   !   write(*,*) "L3246 ii=", ii, " jj=", jj, " i=", i, " j=", j
                   !   write(*,*) "L3247 psfImageTest=", psfImageTest(ii, jj), &
                   !        " ratioImage=",ratioImage(i,j) 
                   !endif
                   if (i .ge. nxmin .and. i .le. nxmax .and. &
                        j .ge. nymin .and. j .le. nymax) then
                      totalCounts10=totalCounts10+(psfImageTest(ii, jj)-yAverB)*ratioImage(i,j)
                   endif
                endif
             endif
          enddo
       endif
    enddo

    if (printFlag) then
       write(*,*) "L3730 HiRes totalCounts10/100=", totalCounts10/100.0, " totalCounts=", &
            totalCounts
    endif

    !if (abs(totalCounts10/100-totalCountsAper3) .gt. totalCountsAper3*0.05) then
    !   totalCounts10=totalCountsAper3*100.0
    !endif
 
    if (printFlag) then
       write(*,*) "L3739 New HiRes totalCounts10/100=", totalCounts10/100.0
    endif

    deblendedCounts=totalCounts10/100.0
    deblendedCounts3=totalCountsAper3

    !**********************************************

    nsubset=0
    do i=1,n
       !if (printFlag) then
       !   write(*,*) "L2283 i=", i, " xData(i,1)=", xData(i,1)
       !endif

       !if (xData(i,1) .lt. 0.004 .and. xData(i,1) .gt. 0.0) then
       if (xData(i,1) .lt. 0.004) then
       !if (yData(i) .gt. yAverB .and. yData(i)  .lt. (yAverB+6.0*sigmaB)) then
       !if (yData(i)  .lt. (yAverB+3.0*sigmaB)) then
       !if (yData(i)  .lt. yMax*0.01) then
          aux1=aux1+1.0
          xAver1=xAver1+xData(i,1)
          yAver1=yAver1+yData(i)
       endif

       !if (xData(i,1) .lt. 0.001) then
       !   auxB=auxB+1.0
       !   yAverB=yAverB+yData(i)
       !endif


       !if (xData(i,1) .gt. 0.01) then
       if (xData(i,1) .gt. 0.005) then
       !if (yData(i) .gt. yAverB+300.0*sigmaB) then
       !if (yData(i) .gt. yMax*0.01) then
          aux2=aux2+1.0
          xAver2=xAver2+xData(i,1)
          yAver2=yAver2+ydata(i)         
       endif
          
    enddo
    if (aux1 .gt. 1.0) then
       xAver1=xAver1/aux1
       yAver1=yAver1/aux1
    endif

    if (aux2 .gt. 1.0) then
       xAver2=xAver2/aux2
       yAver2=yAver2/aux2
    endif


    !if (auxB .gt. 1.0) then
    !   yAverB=yAverB/auxB
    !endif

   
    !slope1=(yAver2-yAver1)/(xAver2-xAver1)
    slope1=(yAver2-yAver1)/(xAver2-xAver1)

    if (printFlag) then
       write(*,*) "L2274 xAver1=", xAver1, " yAver1=", yAver1
       write(*,*) "L2275 xAver2=", xAver2, " yAver2=", yAver2
       write(*,*) "L2276 yAverB=", yAverB, " auxB=", auxB
       write(*,*) "L2277 slope1=", slope1
    endif

    ! compute dispersion relative to the slope
    aux=0.0
    sigmaSlope=0.0
    do i=1,n
       if (xData(i,1) .gt. xAver1 .and. xData(i,1) .lt. xaver2) then
          valueSlope=slope1*(xData(i,1)-xAver1)+yAver1
          aux=aux+1.0
          sigmaSlope=sigmaSlope+(valueSlope-yData(i))*(valueSlope-yData(i))
       endif
    enddo
    if (aux .gt. 1) sigmaSlope=sigmaSlope/(aux-1)
    sigmaSlope=sqrt(sigmaSlope)

    if (printFlag) then
       write(*,*) "L2437 slope1=", slope1, " sigmaSlope=", sigmaSlope
    endif
    if (aux .gt. 1) sigmaSlope=sigmaSlope/sqrt(aux)
    if (printFlag) then
       write(*,*) "L2461  sigmaSlope/sqrt(n)=", sigmaSlope
    endif

    !**********************************************************************************
    !
    ! Least-squares fit with constant (background) term
    !
    !**********************************************************************************  
!
!    call LeastSquaresFit21(n, xdata=xdata, ydata=ydata,              &
!            weight=weight, nterms=int(nSources, int8),               &
!            maxiter=maxIter, nstd1=3.0,                              &
!            gradient=gradient, gradienterror=gradientError,          &
!            constant=constant, constanterror=constantError, std1=std)
!
!    gradient1=gradient
!    constant1=constant


    gradient1=slope1
    constant1=yAverB

    ! test 2010-11-12
    !constant=yAver1
    constant=yAverB
    gradient(1)=slope1

    constantError=sigmaB
    gradientError=sigmaSlope

    !write(*,*) "L2113 gradient1=", gradient1(1), " new gradient=", gradient(1)
    !write(*,*) "L2116 constant1=", constant1, " new constant=", constant

    if(present(nData)) then
        nData = n
    end if

   deallocate(ratioImage)
   nullify(ratioImage)

   deallocate(psfImageTest)
   nullify(psfImageTest)

end subroutine deblendSource

end module ompsfphotometry_mod










