! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!*************************************************************
! Module psffunctions_mod
! Created 15th April 2002- C.Brindle
! Contains various subroutine that make use of the point-spread function 
! 
!****************************************************************
module psffunctions_mod
    use cal
    use dal     
    use OmAllocation_mod     ! For allocating memory for arrays
    use OmSourceParam_mod    ! Defines structures for storing source parameters
    use parameters_mod       ! For computing source moments
    use OmConstants_mod      ! For binx, biny, etc
    use omsmoothimage_mod           ! For polynomial smoothing
    use omnumerics_mod 
    use leastsquares_mod
    use ominterpolate_mod
    use brent_mod
    use cubicshepard_mod
    use splinefit_mod
    use gridinterpolation_mod   
    use ompsf_mod
    use omphotometry_mod



!*************************************************************
!
!************************************************************
type PsfParameters
    real(single) :: psfRatio(3)
    real(single), pointer :: psfImage(:, :) => null()
    real(single) :: psfSemiA = -1.0, psfsemiB = -1.0
    real(single) :: sigmax = -1.0, sigmay = -1.0
end type PsfParameters

type(PsfParameters), save :: psfParams 



contains

!*****************************************************
!
! Computes the semia and semib  of the psf using
! moments 
!
!*****************************************************
subroutine GetPSFWidths(semiA, semiB)

    real(single), intent(out) :: semiA, semiB

    integer(int32) :: psfNx, psfNy
    type(sourceParameters) :: source
   
    real(single), pointer :: psfImage1(:, :) => null()


    if (associated(psfParams%psfImage)) return

    psfnx = 12 ! 25 pixel width
    psfny = 12 ! 25 pixel height

    !*************************************************
    ! Get the point-spread function image
    !*************************************************
    call CreatePSFImage21(psfNx, psfNy, psfParams%psfImage, &
          1024.5_double, 1024.5_double,                     &
          ratio=0.5_double, bin=binX, res=1) 

    
    psfImage1 => allocate2dSingle(0, psfnx - 1, 0, psfny - 1, "psfimage1")


    source%xPos = real(psfNx) / 2.0
    source%yPos = real(psfNy) / 2.0
       
    psfImage1 = psfParams%psfImage * 100.0


    !************************************************************
    !
    ! Compute the semi-major and minor axes of this psf source
    !
    !************************************************************

    call ComputePointSourceParameters(int(-1, int16), int(psfNx - 1, int16), &
                int(psfNy - 1, int16), psfImage1, int(psfNx / 2 - 1, int16), &
                int(psfNy / 2 - 1, int16), source, int(4 / binx, int8))


!    write(*, '( a, f8.2, a, f6.2, a, f8.2, a, f6.2)') "Computed PSF semia=", source%semia, "+/-", source%semiaerr,  &
!       " b=", source%semib, " +/-", source%semibErr

    semiA = source%semiA
    semiB = source%semiB


!write(*, *) "psffunctions_mod.f90 line 102 binx=", binx
!write(*, *) "psf semia=", semia, " semib=", semib

    deallocate(psfimage1)    
    nullify(psfimage1)

      

end subroutine GetPSFWidths








!******************************************************************
!
!
!*****************************************************************
subroutine ClassifySource(source, image, backgroundImage)
    
    type(sourceParameters), intent(inout) :: source
    real(single), pointer, dimension(:, :) :: image, backgroundImage
    integer(int16) :: i0, j0 
    real(single), parameter :: nstd = 3.0 
    integer(int16) :: il, iu, jl, ju
    real(single) :: deltaA, diff, radius, counts1, counts2, bCounts
    real(single) :: xpos, ypos

    source%semiaErr = max(0.1,  source%semiaErr) ! Was 0.02
    source%semibErr = max(0.1,  source%semibErr)

         
    ! Check classification

    il = lbound(image, 1)
    jl = lbound(image, 2)
    iu = ubound(image, 1)
    ju = ubound(image, 2)


    i0 = int(source%xpos - 0.5)
    j0 = int(source%ypos - 0.5)
    i0 = max(il, i0)
    i0 = min(i0, iu)
    j0 = max(jl, j0)
    j0 = min(j0, ju)  


!write(*, *) "imin=", source%imin, " imax=", source%imax
!write(*, *) "jmin=", source%jmin, " jmax=", source%jmax


    deltaA = source%semia - psfparams%psfsemia
    diff =  nstd * source%semiaErr

    if ((image(i0, j0) - backgroundImage(i0, j0)) .lt. 5000.0 .and. & 
            deltaA .gt. diff) then

        xpos = source%xpos - 0.5
        ypos = source%ypos - 0.5
        source%extensionFlag = 1
        !*******************************************************************
        ! Do a further check on the counts
        !*******************************************************************
        radius = real(6 / binx)
        call getCounts2(image, xpos, ypos, radius, counts1)

        !**********************************************************************
        ! Get the background counts
        !*********************************************************************
        call getCounts2(backgroundImage, xpos, ypos, radius, bCounts)
        counts1 = counts1 - bCounts  ! Background subtracted


        !***************************************************************************
        !
        ! Use a bigger aperture
        !**************************************************************************
        radius = real(10 / binx)
        call getCounts2(image, xpos, ypos, radius, counts2)

        !**********************************************************************
        ! Get the background counts
        !*********************************************************************
        call getCounts2(backgroundImage, xpos, ypos, radius, bCounts)
        counts2 = counts2 - bCounts  ! Background subtracted

        if(counts2 > 0.0) then
            if(counts1 / counts2 .ge. 0.8) then
                source%extensionFlag = 0
            end if

        end if 

    else
        source%extensionFlag = 0


     end if

    
!if(source%number .eq. 154) then
!write(*, *) "Classifysource sourcenumber=", source%number, " il=", il, " iu=", iu, " jl=", jl, " ju=", ju
!write(*, *) "x=", source%xpos, " y=", source%ypos, " i0=", i0, " j0=", j0
!write(*, *) "Classify source x=", source%xpos, " y=", source%ypos, " semia=", source%semia, " error=", source%semiaerr, &
!" psfsemia=", psfparams%psfsemia    
!write(*, *) "count1=", counts1, " counts2=", counts2
!write(*, *) "extension=", source%extensionFlag
!end if

end subroutine ClassifySource




!*********************************************************************
! 
!
!********************************************************************
subroutine IsOddSource( source )

    type(sourceParameters), intent(inout) :: source

    real(single) :: diff
    real(single), parameter :: nstd = 3.0 

    diff =  nstd * source%semiaErr

    if ( (source%semiA - diff) < psfparams%psfsemia) then ! Very peculiar source
 
        source%confusionFlag = ibset(source%confusionFlag, 4) 



    end if
 

end subroutine IsOddSource




end module psffunctions_mod


