! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!*******************************************************************************************************
!
! omdetect_mod
!
! Authors:  Tim Harries - MSSL (orig) 
!           almost completely rewritten by J. Mittaz MSSL (Version 3)
!           Then almost completely rewritten by Chris Brindle - MSSL (Versions 4 and 5)
!           Modified by Vladimir Yershov 
! Description: Uccchecksses a simple algorithm to detect sources in an OM OSW imagessm
! 
! Comments: create
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacnti Exrcheckep$ 
!
! History!
!
!    Created: Jan 1998 (TJH)
! 
! 2010-04-20 (vny) The use of the parameter "psfphotometryenabled" is restricted as the
! performance of the psf-photometry algorithm is not yet satisfactory
!
! 2010-03-19 (vny)  Introduced checking of the source proximity to each other in order to avoid 
!	duplicated detections of the same source in the Fast mode (SSC-SPR-6601). The current 
!       proximity threshold is set to 3 unbinned pixels for the Fast mode
! 2010-10-17 (vny) A new parameter (psfphotometryset) is introduced.
! 2010-11-14 (vny) improved psf-photmetry algorithm (a few bugs are fixed, new method
! of calculating the slope of the psf-profile is introduced, the least-squares solution
! is removed from one of the subroutines to accelerate the calculation)
! 2011-02-19 (vny) The use of the parameter psfphotometryenabled is 
!restricted to test cases only (to enable this mode the parameter 
! maxRawCountRates should be set to zero)
! 2011-11-20 (vny) Further improvements of PSF-photometry: the calculation of the
! source profile slope is removed; instead, the PSF phtotometry of close neighbouring
! sources is made by using the calibrated PSF from CAL 
! 2013-11-07 The source x- and y-coordinates are rounded to one-tenth of the pixel size
! to make the photometry output the same on different computers
! 2013-11-12 (vny) channged the name of the local variable apertureRadius to
! apertureRadiusLocal as the variable with the name apertureRadius was already
! declared as a global variable for the whole task in mssllib (module 
! omphotometrydefs_mod.f90).
! 2013-12-04 (vny): setting the maximal rate for corrections to be 0.97 counts per frame
!******************************************************************** 
module omdetect_mod

  use dal
  use cal
  use param  
  use sasdal
  use OmAllocation_mod         ! Functions for allocating memory to arrays.
  use OmSourceParam_mod        ! Source definitions, etc
  use OmNumerics_mod           ! Various numerical subroutines and functions
  use Parameters_mod           ! Computes source parameterscall 
  use OmConstants_mod
  use backgroundimage_mod      ! Computes background image
  use OmdetectDetection_mod    ! Does source detection      
  use OmUtility_mod            ! For storing columns in fits files
  use OmSkyBackground_mod      ! Computes sky mask1ackground
  use OmPhotometry_mod         ! Does aperture photometry
  use OmSmoothImage_mod        ! For smoothing the image
  use getFilterID_mod
  use errorHandling
  use omSrcTable_mod           ! For creating the source-list table
  use frametime_mod, only : getFrameTime ! mssllib
  use sourceanalysis_mod
  use omsourceflags_mod        ! For setting source flags
  use ominterpolate_mod
  use ompsfphotometry_mod
  use ommode_mod 
  implicit none
  private
  public  omdetect


   
  real(double) :: sn, bck
  logical(boolean), pointer   :: pmaskImage(:, :) => Null() ! The image-pixel quality array
  logical(boolean), pointer   :: emaskImage(:, :) => Null() ! The image-pixel quality array
  character(8), private       :: omModeType     ! FAST or IMAGING mode 
  real(single), private :: exposureConstant = 1000.0 ! 3188.0
contains 

  !**************************************************************
  ! The main program
  !**************************************************************
  subroutine omdetect
  
    ! DAL variables
    type(DataSetT)       :: oswDataSet ! their associated handles 
  
    ! binning variables
    logical                 :: binBPE
    
    character(6)            :: omFilter   ! the OM filter
    
    ! input image variables  
    real(single)                       :: exposure = 0.0
    real(single), pointer   :: image(:, :) => Null()
    real(single), pointer   :: tempImage(:, :) => Null() 
    real(single), pointer   :: backgroundImage(:, :) => null()
    
    ! source searching variables
    real(single)                  :: nSigma
    integer(int16), pointer  :: levelImage(:, :) => Null()   
    integer(int32)              :: allocErr
    integer(int16)              :: nSources=0
    integer(int16)              :: nSourcesPsf=0
    real(double) :: lambda
    real(double) :: plateScale
    real(single) :: apertureRadiusLocal     ! Radius (pixels) of circular aperture to be used for photometry      
    real(single)   :: bkgSigma, modeBackground      
    integer(int32) :: istat
    integer(int32) :: windowX0 = 0     ! WindowX0
    integer(int32) :: windowY0 = 0     ! WindowY0
    integer(int32) :: windowDX = 0     ! WindowDX
    integer(int32) :: windowDY = 0     ! WindowDY

    logical  :: fastMode = .false., grismMode = .false.
    integer(int32) :: filterIdInt
    type(sourceParameters), pointer :: sourceData(:)=> null()
    type(sourceParameters), pointer :: sourceDataPsf(:)=> null()

    logical :: psfPhotometryEnabled
    real(single) :: maxRawCountRate
    integer(int8) :: neighboursForPsfPhotometry;

    character(len=1024) :: psfPhotometrySourceListName
    

    integer :: i

    integer(int32) :: xAxisLength = 0           ! Length of image x-axis.
    integer(int32) :: yAxisLength = 0           ! Length of image y-axis.
    integer(int32) :: derivedBinX=0
    integer(int32) :: derivedBinY=0

    ! get the task parameters
    write(errstr,'(a)') "Loading task parameters"
    call message(AppMsg, Verbose)
    
    nSigma = realParameter("nsigma")
    
    ! get handle on osw image
    !oswDataSet = DataSet( trim(stringParameter("set")), read)
    oswDataSet = DataSet( trim(stringParameter("set")), modify)

    ! Get the exposure time in secs
    exposure = real32Attribute(oswDataSet, "EXPOSURE")
    exposureConstant = exposure
    omFilter = trim(stringAttribute(oswDataSet, "FILTER"))
    
    psfPhotometryEnabled=booleanparameter("psfphotometryenabled")
    neighboursForPsfPhotometry = intParameter("neighboursforpsfphotometry") 

    maxRawCountRate=realParameter("maxrawcountrate")
    
    !**************************************************************
    ! The following 8 lines are to enable using psf-photometry 
    ! to sas-developers. For the rest of the users the psf photometry
    ! is disabled 2010-04-20
    !**************************************************************
    !if (maxRawCountRate .eq. 0.0) then
    !   maxRawCountRate=50.0
    !   write(errstr,'(a)') "Setting maximal raw count rate for PSF-photometry to 50 c/s"
    !   call message(AppMsg, Verbose)
    !else
    !   psfPhotometryEnabled=.false.
    !   call warning("Source PSF-photometry is currently disabled in this version")
    !endif



    !******************************************************
    ! Set the radius at which the source significances
    ! will be computed for
    !*****************************************************
    call setSignificanceRadius( omFilter )



    ! Get the OM mode (IMAGING or FAST)
    omModeType = trim(stringAttribute(oswDataSet, "DATAMODE"))

    if (trim(omModeType) .eq. "FAST") then
       fastMode = .true.
    else
       if (omFilter .eq. "Grism1" .or. omFilter .eq. "Grism2" &
            .or. omFilter .eq. "GRISM1" .or. &
            omFilter .eq. "GRISM2") then
          
       end if
       fastMode = .false.
        
    end if
 
    !********************************************************
    ! Find out if there is an exposure image
    !********************************************************
    if( hasBlock(oswDataSet, "EXPOSURE" ) ) then
       ommodeType = "MOSAICED"
        
    end if

    !**********************************************
    ! Load the image from the PRIMARY FITS block
    write(errstr,'(a)') "Loading image"
    call message(AppMsg, Verbose)
   
    call loadTheImage(oswDataSet, nx, ny, image)

    exposure = exposureConstant
   

    !**************************************************************************
    !
    ! A check on the image size - remove this check - SPR-7514
    !
    !**************************************************************************
    !if (nx .gt. 3000 .or. ny .gt. 3000) then  ! Check that image not too big 
       !nx = min(3000_int16, nx)
       !ny = min(3000_int16, ny)
       
       !tempImage => allocate2dSingle(0_int16, nx - 1_int16, 0_int16, ny - 1_int16, "tempimage")
       !tempImage = image(0_int16 : nx - 1_int16, 0_int16 : ny - 1_int16)
       !deallocate(image)
       !image => tempImage
   
       !write(errstr,'("Image too big! ny or ny > 3000 - reset to 3000")')  
       !call message(AppMsg, Verbose)      ! Verbose was 2
       
    !end if


    ! compute binning
    binx = int32Attribute(oswDataSet, "BINAX1")
    biny = int32Attribute(oswDataSet, "BINAX2")
    binBPE = booleanAttribute(oswDataSet, "BINBPE")

    binx = 2 ** binx
    biny = 2 ** biny

    if (binBPE) then
       binx = binx * 2
       biny = biny * 2
    end if

    !write(*,*) "L239 binx=", binx, " biny=", biny
    if (binx .ne. biny) then
       call fatal("diffBin","This version of omdetect cannot deal with different binning factors in x and y")
    end if

    ! Need the frame-time for the photomtery coincidence-loss corrections
    ! NB value is in milli-seconds
    call GetCCDConstants(oswDataSet, binBPE)

    ! Get the window parameters
    call GetImageSize(oswDataSet, windowX0,  windowY0,  windowDX, windowDY)
   
    !write(*,*) "L251 windowX0=", windowX0, " windowY0=", windowY0, &
    !     " windowDX=", windowDX, " windowDY=", windowDY
    !write(*,*) "L253 nx=", nx, " ny=", ny
    !write(*,*) "L254 omModeType=", omModeType

    ! Check the binning factor
    ! (skipping checking for mosaiced images)

    derivedBinX=0
    if (nx .gt. 0) then 
       derivedBinX=windowDX/nx
    endif
    derivedBinY=0
    if (ny .gt. 0) then
       derivedBinY=windowDY/ny
    endif
    !write(*,*) "L266 derivedBinX=", derivedBinX, " derivedBinY=", derivedBinY

    if(.not. omModeType .eq. "MOSAICED") then
       if (derivedBinX .gt. 0 .and. derivedBinY .gt. 0) then
          if (derivedBinX .ne. binX .or. derivedBinY .ne. binY) then
             write (errstr, *) "The header-based binning factor ", binX, " seems to be incorrect (fixing)"    
             call message(AppMsg, Verbose)
             binX=derivedBinX 
             binY=derivedBinY
             write (errstr, *) "Derived binning factor ", binX    
             call message(AppMsg, Verbose)
          endif
       endif
    else
       !write(*,*) "L281 skipping checking the binning factor for a mosaiced image"
    endif

    !  Get the filter Id integer value
    filterIdInt = filterIdToInt(omFilter)

    call CAL_setState(OM, filterId = filterIdInt)
    
    !*****************************************
    ! Get the Plate scale (arcsecs/pixel)
    !*****************************************
    call CAL_omGetPlateScale(lambda, plateScale)
    
 


    !******************************************************************
    ! Detect sources 
    !
    !******************************************************************
   

    if (fastMode) then
       !*************************************************************
       ! Fast mode detection
       !************************************************************
       levelImage => Allocate2Dint16(0, nx - 1, 0, ny - 1, "levelImage")
       levelImage = 0
        
  
       call FastDetection(oswDataSet, image, backgroundimage, levelImage, &
            nSigma, exposure, nSources, sourceData, psfPhotometryEnabled, &
            maxRawCountRate,neighboursForPsfPhotometry)

    else
       if (omFilter .ne. "Grism1" .and. omFilter .ne. "Grism2" &
            .and. omFilter .ne. "GRISM1" .and. omFilter .ne. "GRISM2") then
          !*************************************************************
          ! Imaging mode detection
          !*************************************************************
          call ImagingDetection(oswDataSet, image, backgroundimage, levelImage, &
               nSigma, nSources, sourceData, nSourcesPsf, sourceDataPsf, windowX0, windowY0, exposure, &
               psfPhotometryEnabled, maxRawCountRate, neighboursForPsfPhotometry, &
               modeBackground, bkgSigma)

       else
          grismMode = .true.
          !**************************************************************
          ! Grism detection
          !**************************************************************
          call GrismDetection(image, backgroundimage, levelImage, nsigma, nSources, sourceData)
          
       end if
    end if
 
    write(errstr,'(a, i0)') &
         "Finished running source detection algorithm - total number of sources=", nSources
    
    call message(AppMsg, Verbose)
  
    !********************************************************************************
    ! In order to convert the source counts to magnitudes have to use a 12 pixel
    ! aperture radius to sum the counts
    ! subroutine AperturePhotometry conducts aperture photometry on each source
    !
    !
    ! Need the quality array
    !********************************************************************************
    ! from mssllib:
    ! ComputeBackgroundMode(image, levelImage, nx, ny, nSigma, sigma)
    !
    ! Computes the background mode (counts/per pixel)
    ! using the image (background + source) and 
    ! the source-map image (levelImage). Ignores
    ! any image pixels that correspond to a levelImage
    ! one that has a non-zero value (ie a source pixel)
    ! Also returns the variance of the sky-background 
    !-----------------
    modeBackground = ComputeBackgroundMode(image, levelImage, nx, ny, 2.0, bkgSigma)

    write(errstr,'(a, f0.1, a, f0.1)') "Image background mode=", modeBackground, &
                "+/-", bkgSigma
    !write(*,*) "L321 modeBackground=", modeBackground, " bkgSigma=", bkgSigma

    call message(AppMsg, Verbose)
    
    apertureRadiusLocal = real(12 / binx)
    
    nx = size(levelImage, 1)
    ny = size(levelImage, 2)
 
    if(stringParameter("levelimage") .ne. "")  then
        
       call SaveImage( levelimage, nx, ny,          &
            trim(stringParameter("levelimage")), "level image")
    end if

    ! Save it if required
    if(trim(stringParameter("backgroundimage")) .ne. "")  then
       if (associated(backgroundImage)) then

          call saveBackgroundImage( backgroundImage)
 
       end if
    end if
    
    deallocate(levelImage, stat = istat)
    nullify(levelimage)
    call CheckDeallocation(istat, "levelImage")

    !*************************************
    ! Create the region file if required
    !***********************************

    if (stringParameter("regionfile") .ne. "") then
       if (fastMode .or. grismMode) then
          
          call CreateRegionFile21(int(nSources), sourceData, trim(stringParameter("regionfile")), "green" )
       else
          call CreateRegionFile1(int(nSources), sourceData, trim(stringParameter("regionfile")) )
 
       end if
    end if

    !***********************************************************************
    !
    ! Create the source list
    ! nsigma = 5.0 for computing the detection limiting-magnitude (5 sigma) 
    !
    !************************************************************************
    nSigma = 5.0

    if (associated(backgroundImage)) deallocate(backgroundImage)    
    !write(*,*) "L390 size(sourceData)=", size(sourceData)
    !do i=1,size(sourceData)
    !   write(*,*) " i=", i, " x=", sourceData(i)%xpos, " y=", sourceData(i)%ypos, &
    !        " raw=", sourceData(i)%rawCountRate, &
    !        " err=",sourceData(i)%rawCountRateErr, " rawBkg=", sourceData(i)%rawBackGroundCountRate, &
    !        " corrSrc=",sourceData(i)%corrCountRate, " corrBkg=", sourceData(i)%corrBackGroundCountRate 
    !enddo
    !write(*,*) "------------"
    !write(*,*) "L428 calling CreateSourceListFile binx=", binx, " plateScale=", plateScale 
    call CreateSourceListFile(trim( stringParameter("outset") ),  &
         oswDataSet, int(nSources),              &
         sourceData, plateScale, apertureRadiusLocal, &
         modeBackground, nSigma, exposure,       &
         windowX0, windowY0, windowDX, windowDY) 


    if (psfPhotometryEnabled) then
       !write(*,*) "L437 psf photometry is enabled"
       if (trim( stringParameter("psfphotometryset") )=="") then
          
          write(errstr,'(a)') "Parameter psfphotometryset not defined (using default)"
          call message(AppMsg, Verbose)
          psfPhotometrySourceListName = "psf_photometry_src_list.fit"
          
       else
          psfPhotometrySourceListName = trim(stringParameter("psfphotometryset"))
          !write(*,*) "L358 Name for the psf-photometry source list file:",psfPhotometrySourceListName 
       endif
       !write(*,*) "L448 Creating an extra source list (for PSF-photometry)"
       call CreateSourceListFile(psfPhotometrySourceListName,  &
            oswDataSet, int(nSourcesPsf),              &
            sourceDataPsf, plateScale, apertureRadiusLocal, &
            modeBackground, nSigma, exposure,       &
            windowX0, windowY0, windowDX, windowDY) 
    endif

    if(omModeType .eq. "MOSAICED" .and. nSources .gt. 0) then

       !*******************************************************************************
       ! 
       ! Set RA and DEC columns
       !
       !******************************************************************************* 
       call setRADECColumns(oswdataSet, nSources, sourceData) 
       
    end if

    if (associated(sourceData)) then
       deallocate (sourceData, stat = allocErr)
       call CheckDeallocation(allocErr, "sourceData")
    end if


    call release(oswDataSet)
    write(errstr,'(a)') "Exiting..."
    call message(AppMsg, Verbose)
    
  end subroutine omdetect

  !*********************************************
  !
  ! Subroutine GetImageSize
  !
  !********************************************
  subroutine GetImageSize(oswDataSet, windowX0,  windowY0,  windowDX, windowDY)
    type(DataSetT), intent(in) :: oswDataSet 
    integer(int32), intent(out) ::  windowX0,  windowY0,  windowDX, windowDY
    
    type(TableT) :: inTableH
    integer, parameter :: noColumns = 4
    type(ColumnT) :: inColumnH(noColumns)
    integer(int32) :: i
    integer(int32), pointer ::  windowX0P(:),  windowY0P(:),  windowDXP(:), windowDYP(:)
   
    
    if (hasAttribute(oswDataSet, "WINDOWX0") .and. &
         hasAttribute(oswDataSet, "WINDOWY0") .and. &
         hasAttribute(oswDataSet, "WINDOWDX") .and. &
         hasAttribute(oswDataSet, "WINDOWDY")) then

       windowx0 = int32Attribute(oswdataset, "WINDOWX0")
       windowy0 = int32Attribute(oswdataset, "WINDOWY0")
       windowdx = int32Attribute(oswdataset, "WINDOWDX")
       windowdY = int32Attribute(oswdataset, "WINDOWDY")
       return
    end if

    !**************************************************
    ! Get a handle on the modes binary table extension.
    !**************************************************
    inTableH = table(oswDataSet, "MODES")
    
    !**************************************************
    ! Get handles and pointers on the modes binary table 
    ! extension columns and extract the window data.
    !**************************************************
    inColumnH(1) = column(inTableH, "WINDOWX0", read)
    windowX0P => int32Data(inColumnH(1))
    windowX0 = windowX0P(0)

    inColumnH(2) = column(inTableH, "WINDOWY0", read)
    windowY0P => int32Data(inColumnH(2))
    windowY0 = windowY0P(0)
    
    inColumnH(3) = column(inTableH, "WINDOWDX", read)
    windowDXP => int32Data(inColumnH(3))
    windowDX = windowDXP(0)
    
    inColumnH(4) = column(inTableH, "WINDOWDY", read)
    windowDYP => int32Data(inColumnH(4))
    windowDY = windowDYP(0)


    !************************************
    ! Release the handles
    !***********************************
    do i = 1, noColumns
       call release(inColumnH(i))
    end do
    call release(inTableH)
    
  end subroutine GetImageSize

  !**************************************************
  !
  ! CheckSourceSignificances
  !
  !**************************************************
  subroutine CheckSourceSignificances(noSources, source, nx, ny, levelImage)
    integer(int16), intent(inout) :: noSources
    type(sourceParameters), pointer :: source(:)
    
    integer(int16), intent(in) :: nx, ny
    integer(int16), pointer :: levelImage(:, :)
    
    real(single) :: minSignificance
    integer(int16) :: i, num
    
    integer(int16) :: noRejects1, noRejects2
    integer(int16), pointer :: tempLevelImage(:, :) => null()


    ! Check source signal-to-noise ratios
    minSignificance = realParameter("minsignificance")
    if (minSignificance .lt. 0.001) minsignificance = -100.0
 
    num = 0 ! New source number
    
    noRejects1 = 0
    noRejects2 = 0

    write(errstr,'(a)') "Checking source significances"
    call message(AppMsg, Verbose)

    tempLevelImage => Allocate2Dint16(0, nx - 1, 0, ny - 1, "tempLevelImage")
    tempLevelImage = 0

    do i = 1, nosources
       !write(*,*) "L442 i=", i, " x=", source(i)%xPos, " y=", source(i)%yPos, &
       !     " signif=", source(i)%significance, " rate=", source(i)%rawCountRate
       if (source(i)%significance .ge. minSignificance) then !  .and. &
          
          num = num + 1           
          call ReNumberLevelImagePixels(source(i), levelImage, tempLevelImage, num)
          source(num) = source(i)
          source(num)%number = num
         
       else
              

          !            write(errstr,'(a, i0, a, f7.2, a, f7.2, a, f7.1, a)') "Rejecting source ", i, &
          !                          " (x=", source(i)%xpos, " y=", source(i)%ypos, &
          !                  " significance=", source(i)%significance, ")"
          !            call message(AppMsg, Verbose)
          
  
          if (source(i)%number .ge. int(0, int16)) then     
             noRejects1 = noRejects1 + 1                              
          else                                           
             noRejects2 = noRejects2 + 1
          end if
          
       end if
    end do

    deallocate(levelImage)
    levelImage => tempLevelImage
    noSources = num


    if (noRejects1 .gt. 0) then
       write(errstr,'(a, i0)') "Number of point sources rejected=", noRejects1
       call message(AppMsg, Verbose)
    end if
    if (noRejects2 .gt. 0) then
       write(errstr,'(a, i0)') "Number of extended sources rejected=", noRejects2
       call message(AppMsg, Verbose)
    end if
   
  end subroutine CheckSourceSignificances

  !**************************************************
  !
  ! CheckSourceProximity
  !
  !**************************************************
  subroutine CheckSourceProximity(noSources, source, nx, ny, levelImage, allowedDistance)
    integer(int16), intent(inout) :: noSources
    type(sourceParameters), pointer :: source(:)
    
    integer(int16), intent(in) :: nx, ny
    integer(int16), pointer :: levelImage(:, :)
    integer(int16), intent(in) :: allowedDistance
    
    
    integer(int16) :: i, j, num
    
    integer(int16) :: noRejects1, noRejects2
    integer(int16), pointer :: tempLevelImage(:, :) => null()
    
    logical, pointer :: proximityFlag(:) => null()
    real :: distance, x1, y1, x2,y2
    
    num = 0 ! New source number
    
    noRejects1 = 0
    noRejects2 = 0

    write(errstr,'(a)') "Checking source proximity"
    call message(AppMsg, Verbose)

    tempLevelImage => Allocate2Dint16(0, nx - 1, 0, ny - 1, "tempLevelImage")
    tempLevelImage = 0

    allocate(proximityFlag(1:noSources))
    proximityFlag=.false.

    do i=1, noSources
       if ( .not. proximityFlag(i)) then ! check whethere there are some nearby sources 
          x1=source(i)%xPos
          y1=source(i)%yPos
          do j=i+1,noSources
             x2=source(j)%xPos
             y2=source(j)%yPos
             distance=(x2-x1)*(x2-x1)+(y2-y1)*(y2-y1)
             distance=sqrt(distance)
             if (distance .lt. allowedDistance) then
                ! A nearby source found
                ! Check the significance of both sources
                if (source(i)%significance .ge. source(j)%significance) then
                   ! Flag the source j for its removal
                   proximityFlag(j)=.true.
                else
                   proximityFlag(i)=.true.
                endif
             endif
          enddo
       endif
    enddo
    
    !do i=1,noSources
    !   write(*,*) "L545 i=", i, " x=", source(i)%xPos, " y=", source(i)%yPos, &
    !        " signif=", source(i)%significance, " rate=", source(i)%rawCountRate, &
    !        " proxFlag=", proximityFlag(i)
    !enddo

    ! Remove the source that have been flagged as proximate to the brighter ones 
    do i = 1, nosources
       if (.not. proximityFlag(i)) then !  .and. &
          num = num + 1           
          call ReNumberLevelImagePixels(source(i), levelImage, tempLevelImage, num)
          source(num) = source(i)
          source(num)%number = num
       else
          if (source(i)%number .ge. int(0, int16)) then     
             noRejects1 = noRejects1 + 1                              
          else                                           
             noRejects2 = noRejects2 + 1
          end if              
       end if
    end do

    deallocate(levelImage)
    levelImage => tempLevelImage
    noSources = num

    deallocate(proximityFlag)

    if (noRejects1 .gt. 0) then
       write(errstr,'(a, i0)') "Source proximity: number of point sources rejected as duplicates =", noRejects1
       call message(AppMsg, Verbose)
    end if
    if (noRejects2 .gt. 0) then
       write(errstr,'(a, i0)') "Source proximity: number of extended sources rejected as duplicates =", noRejects2
       call message(AppMsg, Verbose)
    end if
   
  end subroutine CheckSourceProximity

  !********************************************************************
  !  
  !
  !
  !
  !*******************************************************************
  subroutine reNumberLevelImagePixels(source, levelImage, tempLevelImage, number)

    type(sourceParameters), intent(inout) :: source
    
    integer(int16), pointer :: levelImage(:, :)
    integer(int16), pointer :: tempLevelImage(:, :)
    
    integer(int16), intent(in) :: number
    integer(int16) :: i, j


    do j = source%jmin, source%jmax
       do i = source%imin, source%imax
          
          if (levelImage(i, j) .eq. source%number) then 
             tempLevelImage(i, j) = number
             
          end if
       end do
    end do
     
  end subroutine reNumberLevelImagePixels


  !***********************************************************
  ! GetCCDConstants
  ! CCD frame-time and dead-fraction
  !***********************************************************
  subroutine GetCCDConstants(oswDataSet, binBpe)
    type(DataSetT), intent(in) :: oswDataSet 
    logical, intent(in) :: binBpe    
    type(DataSetT) :: wdxDataSet

    !********************************************************************
    ! Need the frame-time for the photomtery coincidence-loss corrections
    ! NB value is in milli-seconds
    !********************************************************************
    if (HasAttribute(oswDataSet, "FRAMTIME") .and. HasAttribute(oswDataSet, "DEADFRAC")) then
       frameTime = Real32Attribute(oswDataSet, "FRAMTIME") / 1000.0   ! Converted to seconds
    
       deadFrac = Real32Attribute(oswDataSet, "DEADFRAC")   
       
    else
       wdxDataSet = dataSet(stringParameter("wdxset"), read)


       call getFrameTime(wdxDataSet, frameTime, deadFrac, binBpe)
       !************************************************
       ! Add CCD frame-time and dead-fraction to header
       !************************************************
!        call setAttribute(oswDataSet, "FRAMTIME", frameTime, "millisecs", "CCD frame time")
!        call setAttribute(oswDataSet, "DEADFRAC", deadfrac, "", "CCD dead fraction")


       frameTime = frameTime / 1000.0
       call release(wdxDataSet)
    end if

  end subroutine GetCCDConstants
  !*****************************************************
  !
  ! This subroutine is for FAST mode data
  !
  !****************************************************

  subroutine FastDetection(oswDataSet, image, backgroundimage, &
       levelImage, nSigma, exposure, nSources, sourceData, &
       psfPhotometryEnabled, maxRawCountRate, neighboursForPsfPhotometry)
    type(DataSetT), intent(in) :: oswDataSet 
    
    real(single), pointer :: image(:, :), backgroundImage(:, :)
    integer(int16), pointer  :: levelImage(:, :)
    real(single), intent(in) :: exposure
    real(single), intent(in) :: nSigma
    integer(int16), intent(out) :: nSources
    type(sourceParameters), pointer :: sourceData(:)

    type(SourceLinkedList) :: sourceList 

    logical, intent(in) :: psfPhotometryEnabled
    real(single), intent(in) :: maxRawCountRate
    integer(int8), intent(in) :: neighboursForPsfPhotometry

    integer(int16) :: nx, ny
    integer(int32) :: minNoPixels    

    real(single), pointer   :: backgroundSigmaImage(:, :) => Null()
   
    integer(int8) :: minCounts, extensionFlag
    integer(int16) :: nPSources = 0
    integer(int16), pointer :: pLevelImage(:, :) => null()
    real(single) :: height
    logical(boolean), pointer   :: qualArray(:, :) => Null() ! The image-pixel quality array

    integer(int16) :: allowedDistance

    integer :: i
    logical  :: fastMode = .true.

    real(single) :: apertureRadiusLocal

    nSources = 0    
    npSources = 0    
    nx = size(image, 1)
    ny = size(image, 2)

    backgroundImage => Allocate2DSingle(0, nx - 1, 0, ny - 1, "backgroundImage") 
    backgroundSigmaImage => Allocate2DSingle(0, nx - 1, 0, ny - 1, "backgroundImage") 
    backgroundSigmaImage = sqrt(backgroundImage)

    qualArray => GetQualityArray(oswDataSet, nx, ny)
  
    call modifyQualityImage(image, qualArray) 
 

    !************************************************************************
    !
    ! Check for hot pixels
    !
    !************************************************************************
    call checkForHotPixels(image, qualArray)


    call createBackgroundImage(image, qualArray, int(24 / binx, int16), backgroundimage, &
             backgroundsigmaimage) ! ws 10000000

    write(errstr,'(a)') "Running FAST mode source detection algorithm"
    call message(AppMsg, Verbose)
  

    ! Main point-source detection call
    call DetectPointSources(nx, ny, image,  image,     &
                            backgroundImage,           &
                            backgroundSigmaImage,      &
                            levelimage=levelImage,     &
                            qualityImage=qualArray,    &
                            nsigma=nsigma,             &
                            exposure=exposure,         & 
                            sourcelist=sourceList,     &
                            nosources=nSources,        &
                            fastMode=fastMode)  


    !**********************************************************
    ! 
    !*********************************************************
    call InitialiseDetection(nx, ny, 1.0, 2, 2.0, 2.0, 1.5) 


    call SearchForPointSources(image,                &
                               image,                &
                               backgroundImage,      &
                               backgroundSigmaImage, &                          
                               levelImage,           &
                               sourceList,           &
                               nSources,             &
                               nSigma,               & 
                               IsPointSource4,       &
                               HeightFunction1,      &
                               fastMode)



    !***************************************************
    !
    ! Check to see if at least one source has been detected
    !
    !***************************************************
    if (nSources .eq.  0) then


        call InitialiseDetection(nx, ny, 0.75, 2, 2.5, 2.5, 1.5) ! was 2.0   
        call SearchForPointSources(image, image,     & 
                               backgroundImage,      &  
                               backgroundSigmaImage, &                         
                               LevelImage,           &
                               sourceList,           & 
                               nSources,             &
                               3.0,                  & ! The height above the background
                               IsPointSource,        & 
                               heightfunction2,      &
                               fastMode) 
                         
    end if


 
    ! Check that at least one source has been detected
    if (nSources .eq. 0) then          
       minNoPixels  = 8 / binX      
       height = nSigma
       minCounts = 1    
       call settitle( "point-sources" )
       extensionFlag = 0
       pLevelImage => Allocate2DInt16(0, nx - 1, 0, ny - 1, "pLevelImage") 
       
       call DetectExtendedSources(nx, ny, image, image, &
            backgroundImage, maxval(image), plevelImage, levelImage,    &  
            sourceList, npSources, nSources, minNoPixels,               &
            5000 / binX,                                                &      
            minCounts, height, heightFunction1, NoChecks,               &
            extensionFlag, .false.)   


       deallocate(plevelImage)
       nullify(plevelImage)
       
       where(levelImage .gt. 0) 
          levelImage = 0
       endwhere

       where(levelImage .lt. 0) 
          levelImage = abs(levelImage)    
       endwhere
    end if

    if (nSources .gt. 0) then
         

       deallocate(qualArray)

       qualArray => GetQualityArray(oswDataSet, nx, ny)


       !************************************************************************
       !
       ! Check for hot pixels
       !
       !************************************************************************
       call checkForHotPixels(image, qualArray)
        
       write(errstr,'(a, i0)') "Number of sources detected=", nsources
       call message(AppMsg, Verbose)
  
       write(errstr,'(a)') "Conducting aperture photometry on the point sources"
       call message(AppMsg, Verbose)
       
       sourceData => SourceDataArray(sourceList, nSources)
       call DeleteSourceList(sourceList)
       apertureRadiusLocal = real(12 / binx)
       sourceData%exposure = exposure
        
       sourceData%extensionFlag = 0 
       !write(*,*) "L960 >>>> calling AperturePhotometry" 
       call AperturePhotometry(nx, ny, image, backgroundImage, levelImage, qualArray,  & 
            apertureRadiusLocal, nSources, sourceData)


       !write(*,*) "L922 nSources=", nSources, " apertureRadius=", apertureRadius
       !do i=1, nSources 
       !   write(*,*) "L924 i=", i, " raw=", sourceData(i)%rawcountrate, &
       !        " rawbkg=", sourceData(i)%rawbackgroundcountrate  
       !enddo
       !***********************************************************************
       !
       ! Recompute photometry for sources with a small signal-to-noise (10.0)
       !
       !**********************************************************************
       !write(*,*) "L975 >>>> calling AperturePhotometry for s/n=10"
       call AperturePhotometry(sourceData, int(nSources), image, levelImage, &
            qualArray, real(6 / binx), 10.0) ! was 10
   
       !************************************************************************
       ! If command-line parameter psfphotometryenabled equals true then
       ! recompute the photometry using point-spread function fitting.
       !
       !*********************************************************************** 
       if (psfPhotometryEnabled) then

          write(errstr,'(a)') "Conducting PSF photometry on point sources"
          call message(AppMsg, Verbose)
             
          call initialisePSFPhotometry( image, maxRawCountRate )
          !***********************************************************************
          ! 
          !
          !**********************************************************************
          !call PSFPhotometry(size(sourceData), size(sourceData), sourceData, sourceData, &
          !     int(8 / binx, int8), 2_int8, 0_int8)
          call PSFPhotometry(image, size(sourceData), size(sourceData), sourceData, sourceData, &
               int(8 / binx, int8), 2_int8, neighboursForPsfPhotometry)
       end if


       call CheckSourceSignificances(nSources, sourceData, nx, ny, levelImage)
       
       allowedDistance=3/binx
       call CheckSourceProximity(nSources, sourceData, nx, ny, levelImage, allowedDistance)

    end if


    call delete ! Delete some arrays
 
    deallocate(qualArray)
    nullify(qualArray)

    deallocate(backgroundsigmaimage)
    nullify(backgroundsigmaimage)

  end subroutine FastDetection

  !*****************************************************
  !
  ! Subroutine for imaging mode detection
  !
  !
  !****************************************************
  subroutine ImagingDetection(oswDataSet, image, backgroundimage, levelImage, &
       nSigma, nSources, sourceData, nSourcesPsf, sourceDataPsf, x0, y0, exposure, &
       psfPhotometryEnabled, maxRawCountRate, neighboursForPsfPhotometry, &
       modeBackground, bkgSigma)

    type(DataSetT), intent(in) :: oswDataSet
 
    real(single), pointer :: image(:, :), backgroundimage(:, :)
  
    integer(int16), pointer  :: levelImage(:, :)
    real(single), intent(in) :: nSigma
    integer(int16), intent(out) :: nSources
    integer(int16), intent(out) :: nSourcesPsf
    real(single), intent(in)   :: bkgSigma, modeBackground  

    type(sourceParameters), pointer :: sourceData(:)  
    type(sourceParameters), pointer :: esourceData(:)  
    type(sourceParameters), pointer :: sourceDataPsf(:)  
    type(sourceParameters), pointer :: esourceDataPsf(:)  
    integer, intent(in) :: x0, y0
    real(single), intent(inout) :: exposure 
    real(single), pointer :: exposureImage(:, :) =>  null()    

    logical, intent(in) :: psfPhotometryEnabled
    real(single), intent(in) :: maxRawCountRate
    integer(int8), intent(in) :: neighboursForPsfPhotometry

    !*************************************************************************
    !
    ! Local variables
    !
    !************************************************************************
    integer(int16) :: nx, ny
    !AI casting to integer kind9 to avoid overflow 
    !multiplication
    integer(int32) :: ax,ay
    integer(int16)              :: nPSources = 0, nESources = 0
    integer(int8) :: minCounts, extensionFlag, kMax
    integer(int16) :: dx
    real :: t1, t2
    logical :: detectExtended, detectLargeextended
    
    real(single), pointer   :: backgroundsigmaImage(:, :) => null() 

    integer(int16), pointer  :: pLevelImage(:, :) => Null()  ! For point sources
    integer(int16), pointer  :: eLevelImage(:, :) => Null()  ! For extended sources
    integer(int16), pointer  :: lLevelImage(:, :) => Null()  ! For large extended sources
    integer(int16), pointer :: qualArrayInt16(:, :) => null()
    integer(int16) :: i, k, j    
    real(single) :: height
    real(single)   :: neSigma        
    ! Quality array in the boolean form
    logical(boolean), pointer   :: qualArray(:, :) => Null() ! The image-pixel quality array

    ! Quality array in the integer form
    type(ArrayT) :: qualMapH
    integer(int16), dimension(:, :),  pointer :: qualMapP   
    integer(int16), pointer :: qualMap(:, :) => Null()
    integer(int8), dimension(:, :),  pointer :: qualMapInt8ImageP     ! Integer8 array pointer.

    !type(ArrayT) :: qualMap2H
    !integer(int16), dimension(:, :),  pointer :: qualMap2P  => Null() 
    !integer(int16), pointer :: qualMap2(:, :) => Null()
    !integer(int32) :: qualMapDim(2) 

    type(SourceLinkedList)  :: sourceList 
    type(SourceLinkedList)  :: sourceListPsf
 
    real(single) :: maxCounts, minPointSourceDistance = 0.0 
     
    type(SourceLinkedList) :: psourceList ! For point sources 
    type(SourceLinkedList) :: esourceList ! For extended sources 
    type(SourceLinkedList) :: lsourceList ! For large extended sources 
    integer(int32) :: minNoPixels
    logical :: filterImage
    integer(int8) :: backgroundMethod
    logical :: fastMode=.false.


    ! for checking the magnitude calculation
    real(kind=double) :: corrRate  
    real(kind=double) :: corrRateErrPlus                              ! Corrected rate error + (asymmetric).
    real(kind=double) :: corrRateErrMinus                             ! Corrected rate error - (asymmetric).
    real(kind=double) :: corrMagErrPlus                               ! Corrected magnitude error + (asymmetric).
    real(kind=double) :: corrMagErrMinus                              ! Corrected magnitude error - (asymmetric)
    real(kind=double) :: corrMag

    real(single) :: apertureRadiusLocal

    backgroundMethod = 6 ! The background method for aperture photometry
    backgroundMethod = intParameter("backgroundmethod")    

    !write(*,*) "L1113 backgroundMethod=", backgroundMethod

    neSigma = nSigma + 1.0 
    nPSources = 0 ! The number of point-sources
    nESources = 0 ! The number of extended-sources

    detectExtended = booleanparameter("detectextended") ! if FALSE don't look for extended sources 
    

    nx = size(image, 1) ! The number of image columns  
    ny = size(image, 2) ! The number of image rows

    call CPU_TIME(t1)             


    !Check
    !qualMapDim(1)=nx
    !qualMapDim(2)=ny
    !qualMap2H = addArray(oswDataSet, "QUALITY2", Integer16, qualMapDim)
    !qualMap2P => int16Array2Data(qualMap2H)
    !qualMap2P=1
    !call release(qualMap2H)



    ! Need the quality array from the image fits file
    qualArray => GetQualityArray(oswDataSet, nx, ny)
    call modifyQualityImage(image, qualArray) 
 
    !write(*,*) "L1104 checking whether the quality map is int8"
    
    ! Need also the same array in the int8-form from the image fits file
    allocate(qualMap(0 : nx - 1, 0 : ny - 1))
    qualMapH = array(oswDataSet, "QUALITY", modify)
    !write(*,*) "L1110 arrayDataType=", arrayDataType(qualMapH)
    ! Convert the quality image to int16, if necessary
    select case(arrayDataType(qualMapH))
    case(integer8)
       !write(*,*) "L1114 array is of integer8 type"
       qualMapInt8ImageP => int8Array2Data(qualMapH) 
       qualMap = qualMapInt8ImageP      
       qualMapP => qualMap
    case(integer16)
       !write(*,*) "L1119 array is of integer16 type"
       qualMapP => int16Array2Data(qualMapH)
       qualMap=qualMapP
    end select
    
    !where (qualMap .eq. 0) qualMapP=1    
    !write(*,*) "L1087 saving qualMap"
    !call saveImage(qualMap, nx, ny, "qualmap.fits", "quality map image")

    !write(*,*) "L1165 checking for hot pixels"
    !************************************************************************
    !
    ! Check for hot pixels
    !
    !************************************************************************
    call checkForHotPixels(image, qualArray)


    !*************************************************************************************
    !
    ! Create the background image
    !
    !************************************************************************************

    dx = 50 / binx  ! The actual width

    if (dx .gt. nx) dx = max(nx, 20_int16)
    if (dx .gt. ny) dx = max(ny, 20_int16)

    !write(*,*) "L1185 creating background image"
    call createBackgroundImage(image, qualArray, dx,        &
         backgroundimage, backgroundsigmaimage,  &
         maxheight = 10000000.0)

    !write(*,*) "L1098 saving the background image"
    !call saveImage(backgroundimage, nx, ny, "pbkgimage.fits", "background image")
        
    write(errstr,'(a)') "Running the Imaging Mode source detection algorithm"
    call message(AppMsg, Verbose)

     
    !*****************************************************************************************
    !
    ! Source-number image for the point-sources
    !
    !*****************************************************************************************
    pLevelImage => Allocate2DInt16(0, nx - 1, 0, ny - 1, "pLevelImage") ! Point-source level image

    !write(*,*) "L1204 calling DetectPointSources"
    !*************************************************
    ! Main point-source detection call
    !************************************************
    call DetectPointSources(nx, ny, image,  image,      &
                            backgroundImage,            & 
                            backgroundsigmaimage,       &
                            levelimage=plevelImage,     &
                            qualityImage= qualArray,    &
                            nsigma=nsigma,              &
                            exposure=exposure,          &
                            sourcelist=pSourceList,     &
                            nosources=nPSources,        &
                            fastMode=fastMode)  

    !****************************************************************
    !
    ! The following parameter determines the minimum allowed radius between 
    ! 2 point-sources 
    ! NB was 6 / binx
    !
    !***************************************************************/
    minPointSourceDistance = real(6 / binX)
  
    call MergePointSources(npsources, pSourceList, image, &
         backgroundimage, plevelImage, minPointSourceDistance)


    call CPU_TIME(t2)  ! Check how much cpu time it has taken           

    detectLargeExtended = .true.
    
    !AI casting to a integer kind 9 to avoid multiplication
    !overflow
    ax = int(nx,int32)
    ay = int(ny,int32)   
    if (real(npSources) / (ax * ay) .gt. 0.0005 .or. (t2 - t1) .gt. 20000.0 ) then
       
       detectExtended = .false.      
       detectLargeExtended = .false.      
       
       write(errstr,'(a)') 'Crowded field- "large" extended-source detection switched off'
       call message(AppMsg, Verbose)
 
    end if

    pmaskImage => getMaskImage()   ! Need the pixel mask images
    emaskImage => getEMaskImage()
                
    !*****************************************************
    ! Large extended-source detection
    !*****************************************************
    !write(*,*) "L1164 detectLargeExtended=", detectLargeExtended
    write(errstr,'(a)') "Detection of large extended sources is specified"
    call message(AppMsg, Verbose)
    lLevelImage => Allocate2DInt16(0, nx - 1, 0, ny - 1, "lLevelImage")
    
    minCounts = 2   
    minNoPixels = 1000 / binx
    extensionFlag = 1
    if  (detectLargeExtended) then
       !write(*,*) "L1261 running detection of large extended sources"
       write(errstr,'(a)') "Running the detection of large extended sources"
       call message(AppMsg, Verbose)
       call DetectLargeExtendedSources(nx, ny, backgroundImage, backgroundImage, qualMapP,  &
            maxval(image), modeBackground, bkgSigma, &
            lLevelImage, lsourceList, 1_int16, nSources, &
            minNoPixels, 80000 / binX,                                           &
            minCounts, height, extendedheightFunction, IsValidLargeExtendedSource,     &
            extensionFlag, .false.) 
            !extensionFlag, .true.) ! true for erosion of the extended sources 
       !write(*,*) "L1175 saving llevelImage"
       !call saveImage(llevelimage, nx, ny, "llevelimage.fits", "llevel image")

    endif

    !*****************************************************
    ! Extended-source detection
    !*****************************************************
    eLevelImage => Allocate2DInt16(0, nx - 1, 0, ny - 1, "eLevelImage")
     
    !*********************************************************
    ! If any point sources have been picked up as extended 
    ! in the point-source list, transfer
    ! them to the extended source-list
    !*********************************************************
    call transfereSources(nPsources, pSourceList, plevelImage, &
         nesources, eSourceList, elevelImage)

  
    where(emaskImage .and. elevelImage == 0) elevelImage = 1
    
    if(omModeType .eq. "MOSAICED") then
       filterImage = .true.
    else
       filterImage = .false.
    end if

    !*************************************************
    !
    ! Look for extended sources
    !*************************************************
    if (detectExtended) then

       !*************************************************************
       ! Try to locate smoke-rings near bright sources from their
       ! predicted positions
       !
       !*************************************************************
       if(omModeType .ne. "MOSAICED" .and. npSources .gt. 0) then
          call FindSmokeRings(  pSourceList, npsources, x0, y0,      &
               image, eSourceList, neSources,           &
               eLevelImage, plevelimage, qualArray)
       end if


       !******************************************************************
       !
       ! Detect extended sources - number them negatively from the 
       ! number of the last source found 
       !******************************************************************
       minCounts = 2    
             
       call settitle( "large extended-sources" )      
       extensionFlag = 1
       
       write(errstr,'(a)') "Looking for extended sources"
       call message(AppMsg, Verbose)
  
       kMax = 2  
       
       maxCounts = maxval(image)

       do k = 0, kmax              

          height = max(1.0, 1.5 + real(k) + nsigma - 2.0) 
          
          ! Extended-source detection
          call DetectExtendedSources(nx, ny, image, image, backgroundImage,      &
               maxcounts, pLevelImage, eLevelImage, eSourceList, 1_int16,    &
               nESources, 500 / binX, 12000 / binX,                          &                          
               minCounts, height, heightFunction11, IsValidExtendedSource,   &
               extensionFlag, filterImage) 


          where(eLevelImage .gt. 1) eLevelImage = 0

       end do

       write(errstr,'(a, i0)') "Number detected so far=", nesources
       call message(AppMsg, Verbose)
  

       if (nESources .gt. 0) then 
          
          where(eLevelImage .lt. 0) 
             pmaskImage = .true.
             emaskImage = .true.
          end where
                  
          where(emaskImage .and. elevelimage .eq. 0) elevelimage = 1
 
       end if

       !*************************************************************************
       !
       ! Mask out large scattered-light regions
       !
       !************************************************************************
       !call MaskScatteredLightRegions(image,                   &
       !     eLevelImage, plevelImage,    &
       !     real(x0), real(y0))
       !write(*,*) "L1277 calling maskScatteredLightRegions2"
       call MaskScatteredLightRegions2(image,                   &
            eLevelImage, plevelImage, lLevelImage,   &
            real(x0), real(y0))
 
    end if ! (detectextended)


    call CPU_TIME(t2)             

    if (npSources .gt. 0) then

       where(plevelimage .ne. 0) 
          pmaskimage = .true.
          emaskimage = .true.

       end where
       
    end if

    !*********************************************************
    ! If any point sources have been picked up as extended 
    ! in the point-source list, transfer
    ! them to the extended source-list, etc
    !*********************************************************
    call transfereSources(nPsources, pSourceList, plevelImage, &
         nesources, eSourceList, elevelImage)
    
   
    where(elevelImage < 0) 
       
       pmaskImage = .true.
       
    end where

   
    where (emaskImage .and. elevelimage .eq. 0) elevelImage = 1


    !**********************************************
    !
    ! Search for "small" extended sources
    !
    !**********************************************
    if (detectExtended .and. (t2 - t1) .lt. 72000.0) then
     
   
       minCounts = 2  

       call settitle( "small extended-sources" )
       extensionFlag = 1            
       minNoPixels = 50 / binx ! was 20 / binx
       kmax = 2 
       maxcounts = maxval(image)
       do k = 0, kmax              
 
          height = max(1.0, real(k) + nsigma - 2.0) 

         
          call DetectExtendedSources(nx, ny, image, image,                                   &
               backgroundImage, maxCounts, pLevelImage, eLevelImage,                   &  
               eSourceList, 1_int16, nESources, minNoPixels, (2000 / (binX * binx)),   &     
               minCounts,  height, heightFunction1, IsValidSmallExtendedSource,        &
               extensionFlag, filterImage)           

          where(eLevelImage .gt. 1 .and. .not. emaskImage) eLevelImage = 0
       end do


       write(errstr,'(a, i0)') "Number detected so far=", nesources
       call message(AppMsg, Verbose)
  

       where(eLevelImage .gt. 1) eLevelImage = 0
       

       where(eLevelImage .lt. 0) 
          pmaskImage = .true.
          emaskimage = .true.
       end where
       
     
    end if  ! (detectExtended)

    where (emaskImage .and. elevelimage .eq. 0) eLevelImage = 1



!call displaySourceList(psourcelist, npsources)
!call saveImage(plevelimage, nx, ny, "plevelimage.fits", "level image")
!call fatal("stopped line 1420 displaying plevelimage.fits")


    nSources = nPSources + neSources
   
    call settitle( 'faint sources' )
    extensionFlag = 0


 
    call fillInHoles(neSources, esourceList, elevelImage)

    if (nSigma <= 2.0000001) then
           
       where  ( (pmaskImage .or. emaskImage) .and. elevelimage .eq. 0) 
          eLevelImage = 1
       endwhere
       
       where(elevelImage < 0) 
          pmaskImage = .true.
       endwhere
       
       !***********************************************************************
       ! Slightly smooth the image
       !
       !***********************************************************************/.
       call lpdfilter( image, qualArray )

       call InitialiseDetection(nx, ny, 1.0, 3, 2.0, 2.0, 1.5) 

       !write(*,*) "L1492 calling SearchForPointSources"
       call SearchForPointSources(image,                &
            image,                &
            backgroundImage,      &
            backgroundSigmaImage, &                          
            plevelImage,          &
            psourceList,          &
            npSources,            &
            nSigma,               & 
            IsPointSource4,       &
            HeightFunction1,      &
            fastMode)

       where(plevelImage .ne. 0) elevelImage = 1
      
       write(errstr,'(a)') "Looking for faint sources"
       call message(AppMsg, Verbose)
  
       extensionFlag = 1
       minCounts = 1

       maxCounts = 5000.0
       
       do i = 2, 5
          
          height = nsigma + real(i - 2) 
          if(height < 1.0) cycle

          if(omModeType .eq. "MOSAICED") then

             minNoPixels = 6 / binx ! was 12 
          else
             minNoPixels = 6 / binx        
             call setMinimumHeight(3.0)
          end if

          call DetectExtendedSources(nx, ny, image, image,                          &
               backgroundImage, maxcounts, pLevelImage, eLevelImage,          &  
               eSourceList, 1_int16, nESources,                               &
               minNoPixels,                                                   &
               10000 / (binX * binx),                                         &                             
               minCounts, height, extendedheightFunction, IsValidFaintSource, &  
               extensionFlag, filterlevelimage = .true.)                    


          where(eLevelImage .gt. 1) eLevelImage = 0
  
       end do

       where(elevelImage < 0) pmaskImage = .true.

    end if




    !******************************************************
    !
    ! Need to reload the image
    !
    !******************************************************        
    call loadTheImage(oswDataSet, nx, ny, image)

   
    !*********************************************************
    !
    ! Have now finished trying to detect all the sources
    !
    !*********************************************************
    
    where(eLevelImage .gt. 0) elevelImage = 0
    
    write(errstr,'(a, i0)') "Number detected=", nesources
    call message(AppMsg, Verbose)
    
    !*********************************************************
    ! If any sources have been picked up as "point" in 
    ! the extended-source list, transfer
    ! them to the point source-list
    !*********************************************************    
    call transferpSources(nPsources, pSourceList, plevelImage, &
         nesources, eSourceList, elevelImage)


    !*********************************************************************
    ! Check for any 2 point-sources closer than minPointSourceDistance
    ! Merge them if necessary
    !*********************************************************************/
    minPointSourceDistance = real(4 / binX)
    call MergePointSources(npsources, pSourceList, image, &
         backgroundimage, plevelImage, minPointSourceDistance) 
    
    
    !*********************************************************
    ! If any point sources have been picked up as extended 
    ! in the point-source list, transfer
    ! them to the extended source-list
    !*********************************************************
    call transfereSources(nPsources, pSourceList, plevelImage, &
         nesources, eSourceList, elevelImage)
    
    if (nESources .gt. 0) then
       
       !*******************************************************
       ! See if any source can be split into 2 or more sources
       !******************************************************* 
       call decomposeExtendedSources(image,             &
            plevelimage,       &
            elevelImage,       &
            eSourceList,       &
            nESources,         &
            isValidSplitSource)


       !**********************************************
       ! 
       !*********************************************
       call MergeExtendedSources(nesources, eSourceList, image,   &
            backgroundimage, elevelImage, 3.0) ! was 3.0 
       
       if(nESources .gt. 0) then
          esourceData => SourceDataArray(esourceList, neSources)
          
              
          !***********************************************************************************
          ! Create a new background image, using the old one
          !***********************************************************************************
          
          if (maxval(esourceData%noSourcePixels) .gt. 200 / binx) then 
             deallocate(esourceData)
             nullify(esourceData)  
             dx = 60 / binx
             if (dx .gt. nx) dx = max(nx, 20_int16)
             if (dx .gt. ny) dx = max(ny, 20_int16)
             where(plevelImage .gt. 0 .and. elevelImage .eq. 0) elevelImage = plevelImage
             
             call createBackgroundImage(backgroundimage, elevelimage, dx, maxheight = 1.0) 
             where(elevelimage .gt. 0) elevelimage = 0
          else
             deallocate(esourceData)
             nullify(esourceData)
          end if
          
       end if

       !********************************************************************************
       !
       ! Check the classification of the extended sources
       !
       !******************************************************************************* 
       call ReclassifyExtendedSources(neSources, eSourceList, image, backgroundImage, elevelImage)
       
       !*********************************************************
       ! If any sources have been picked up as "point" in 
       ! the extended-source list, transfer
       ! them to the point source-list
       !*********************************************************    
       call transferpSources(nPsources, pSourceList, plevelImage, &
            nesources, eSourceList, elevelImage)
       
    end if

    nSources = nPSources + neSources ! The total number of sources
    
    apertureRadiusLocal = real(12 / binx) 

    !*****************************************************************
    !
    ! Construct the final source map
    !
    !****************************************************************
    levelImage => allocate2dint16(0, nx -1 , 0, ny - 1, "levelImage")

    !*********************************************************
    ! If running on an mosaiced sky image need to re-load
    ! the image and get the exposure image
    !
    !*********************************************************
    if (ommodetype .eq. "MOSAICED") then
         
       call GetExposureImage( oswdataSet, exposureImage) 
       call setExposureImage( exposureImage, exposureConstant )
       
       !**************************************************************
       !
       ! Need to set the exposure to 1 since an exposure-image
       ! is being used
       !
       !*************************************************************   
       exposure = 1.0
       
      
       !************************************************************
       ! Compute the aperture radius
       ! cdelt1 is the plate-scale (degrees per pixels and negative
       !************************************************************
       !        cdelt1 = -real64Attribute( oswDataSet, "CDELT1") * 3600.0  ! RA plate-scale (degrees and negative)
       !        apertureRadius = apertureRadius / cDelt1      
           
       write(errstr,'(a)') "Using the exposure image for photometry"
       call message(AppMsg, Verbose)

    end if





    !*********************************************************
    !
    !
    !*********************************************************

    deallocate(qualArray)
    qualArray => GetQualityArray(oswDataSet, nx, ny)

    !************************************************************************
    !
    ! Check for hot pixels
    !
    !************************************************************************
    call checkForHotPixels(image, qualArray)


    !write(*,*) "L1588 nPSources=", nPSources

    if (nPSources .gt. 0) then ! Aperture photometry on the point sources

       !write(*,*) "L1592 implementing aperture photometry of point sources"
       !****************************************************
       ! Load the point sources from the linked list
       !****************************************************    
       sourceData => SourceDataArray(pSourceList, npSources)
       !write(*,*) "L1639 -------------- npSources=", npSources
       
       !do i=1,npSources
       !   write(*,*) " i=", i, " x=", sourceData(i)%xpos, " y=", sourceData(i)%ypos, &
       !        " raw=", sourceData(i)%rawCountRate, &
       !        " err=",sourceData(i)%rawCountRateErr, &
       !        " rawbkg=", sourceData(i)%rawBackgroundCountRate
       !enddo
       !write(*,*) "------------"
       
       !*********************************************************************
       ! If psf-photometry is requested, then create a copy of the source data
       ! to keep the aperture photometry intact
       !*********************************************************************


       esourceData => SourceDataArray(esourceList, neSources)

       !write(*,*) "L1669 -------------- neSources=", neSources

       !if (neSources .gt. 0) then
       !   do i=1,neSources
       !      write(*,*) " i=", i, " x=", esourceData(i)%xpos, " y=", esourceData(i)%ypos, &
       !           " raw=", esourceData(i)%rawCountRate, &
       !           " err=",esourceData(i)%rawCountRateErr 
       !   enddo
       !   write(*,*) "------------"
       !else
       !   write(*,*) "L1581 no extended sources to display"
       !endif



       !**************************************************************************
       !
       ! Remove bumps due to point-sources from background image
       !
       !**************************************************************************
       call removeSources(backgroundimage, plevelimage, elevelimage, sourceData, npSources)

       !write(*,*) "L1720 saving the background image - 2"
       !call saveImage(backgroundimage, nx, ny, "pbkgimage2.fits", "background image 2")

       !****************************************************
       !
       ! Perform aperture photometry on the point-sources
       !
       !****************************************************
       sourceData%exposure = exposure
       if (neSources .gt.0) then
          esourceData%exposure = exposure
       endif
       
       write(errstr,'(a)') "Conducting aperture photometry of point sources"
       call message(AppMsg, Verbose)
       !write(*,*) "L1780 backgroundMethod=", backgroundMethod

       if( backgroundMethod < 5_int8) then 
          !write(*,*) "L1783"
          call setBackgroundMethod( backgroundMethod ) ! Set the background method
          !write(*,*) "-----------"
          !write(*,*) "L1786 npSources=", npSources, " sourceData:"
          !do i=1, npSources
          !   write(*,*) "L1788 before  i=", i, " x=", sourceData(i)%xPos, " y=", sourceData(i)%yPos !, &
          !enddo
          !write(*,*) "L1798 >>>> calling AperturePhotometry: apertureRadiusLocal=", apertureRadiusLocal
          call AperturePhotometry(nx, ny, image, levelImage, qualArray,  & 
               apertureRadiusLocal, npSources, sourceData)

          !write(*,*) "-----------"
          do i=1, npSources
             !write(*,*) "L1803 after1 i=", i, " x=", sourceData(i)%xPos, " y=", sourceData(i)%yPos, &
             !     " mag=", sourceData(i)%magnitude, " rawbkg=", sourceData(i)%rawBackgroundCountRate, &
             !     " apRadius=", sourceData(i)%apertureRadius, " innerBkg=", sourceData(i)%innerBckRadius, &
             !     " corrCountRateErr=",sourceData(i)%corrCountRateErr 


             ! Check calculating the magnitude
             corrRate=sourceData(i)%corrCountRate
             corrRateErrMinus=sourceData(i)%corrCountRateErr
             corrRateErrPlus=sourceData(i)%corrCountRateErr

             !write(*,*) "L1693 i=", i, "  coorRate=", corrRate, " corrMag=", corrMag
             !call CAL_omPhotoMagnitude(corrRate, corrMag, -corrRateErrMinus, &
             !             corrRateErrPlus, corrMagErrMinus, corrMagErrPlus)
             
          enddo
          !***********************************************************************
          !
          ! Recompute photometry for sources with a small signal-to-noise (10.0)
          !
          !**********************************************************************
          !write(*,*) "-----------"
          !write(*,*) "L1826 >>>> calling AperturePhotometry for s/n=10"
          call AperturePhotometry(sourceData, int(npSources), image, levelImage, &
               qualArray, real(6 / binx), 10.0) ! was 10

          do i=1, npSources
             !write(*,*) "L1688 after2 i=", i, " x=", sourceData(i)%xPos, " y=", sourceData(i)%yPos, &
             !     " corrCountRateErr=",sourceData(i)%corrCountRateErr 
             !     " exposure=", sourceData(i)%exposure, " signif=", sourceData(i)%significance, &
             !     " raw=", sourceData(i)%rawCountRate, " corr=", sourceData(i)%corrCountRate, &
             !     " mag=", sourceData(i)%magnitude, " rawbkg=", sourceData(i)%rawBackgroundCountRate, &
             !     " CLcorrBkg=", sourceData(i)%corrBackgroundCountRate, " err=", sourceData(i)%corrBackgroundCountRateErr, &
             !     " apRadius=", sourceData(i)%apertureRadius, " innerBkg=", sourceData(i)%innerBckRadius, &
             !     " outerBkg=", sourceData(i)%outerBckRadius, " iMin=", sourceData(i)%iMin, &
             !     " iMax=", sourceData(i)%iMax, " jMin=", sourceData(i)%jMin, " jMax=", sourceData(i)%jMax
             !write(*,*) "L1843 i=", i, " x=", sourceData(i)%xPos, " y=", sourceData(i)%yPos, &
             !     " signif=", sourceData(i)%significance, &
             !     " raw=", sourceData(i)%rawCountRate, " corr=", sourceData(i)%corrCountRate, &
             !     " rawbkg=", sourceData(i)%rawBackgroundCountRate, &
             !     " CLcorrBkg=", sourceData(i)%corrBackgroundCountRate  

             ! Check calculating the magnitude
             corrRate=sourceData(i)%corrCountRate
             corrRateErrMinus=sourceData(i)%corrCountRateErr
             corrRateErrPlus=sourceData(i)%corrCountRateErr
             !write(*,*) "L1723 coorRate=", corrRate, " corrMag=", corrMag


             !call CAL_omPhotoMagnitude(corrRate, corrMag, -corrRateErrMinus, &
             !             corrRateErrPlus, corrMagErrMinus, corrMagErrPlus)
             
          enddo


       else
          !write(*,*) "L1863 >>>> calling AperturePhotometry "
          call AperturePhotometry(nx, ny, image, backgroundImage, levelImage,   &
               qualArray,  apertureRadiusLocal, npSources, sourceData)

          !***********************************************************************
          !
          ! Recompute photometry for sources with a small signal-to-noise (10.0)
          !
          !**********************************************************************
          !write(*,*) "L1872 >>>> calling AperturePhotometry for s/n=10"
          call AperturePhotometry(sourceData, int(npSources), image, backgroundImage, &
               levelImage, qualArray, real(6 / binx), 10.0) ! was 10
       end if

       !write(*,*) "L1874 after aperture photometry -------------- npSources=", npSources
       !do i=1,npSources
       !   write(*,*) " i=", i, " x=", sourceData(i)%xpos, " y=", sourceData(i)%ypos, &
       !        " raw=", sourceData(i)%rawCountRate, &
       !        " err=",sourceData(i)%rawCountRateErr, " rawBkg=",sourceData(i)%rawBackgroundCountRate  
       !enddo
       !write(*,*) "------------"

       if (neSources .gt. 0) then
          write(errstr,'(a)') "Aperture photometry of extended sources"
          call message(AppMsg, Verbose)
       
          if( backgroundMethod < 5_int8) then 
             call setBackgroundMethod( backgroundMethod ) ! Set the background method
             call AperturePhotometry(nx, ny, image, levelImage, qualArray,  & 
                  apertureRadiusLocal, neSources, esourceData)
          endif
          !write(*,*) "L1727 -------------- neSources=", neSources
       
          !do i=1,neSources
          !   write(*,*) " i=", i, " x=", esourceData(i)%xpos, " y=", esourceData(i)%ypos, &
          !        " raw=", esourceData(i)%rawCountRate, &
          !        " err=",esourceData(i)%rawCountRateErr 
          !enddo
          !write(*,*) "------------"
       
       endif

       !************************************************************************
       ! If command-line parameter psfphotometryenabled euals true then
       ! recompute the photometry using point-spread function fitting.
       !
       !*********************************************************************** 

       if (psfPhotometryEnabled) then
          write(errstr,'(a)') "Conducting PSF photometry of point sources"
          call message(AppMsg, Verbose)
          !write(*,*) "L1924 maxRawCountrate=",maxRawCountRate
          call initialisePSFPhotometry(image, maxRawCountRate)
   
          !**********************************************************************
          ! Create new source arrays to keep the aperture-photometry data intact
          !**********************************************************************
          allocate(sourceDataPsf( 1 : npSources))

          if (neSources .gt. 0) then
             allocate(esourceDataPsf( 1 : neSources))
          endif
          
          do i=1,npSources
             sourceDataPsf(i)=sourceData(i)
          enddo

          if(neSources .gt. 0) then
             do i=1,neSources
                esourceDataPsf(i)=esourceData(i)
             enddo
          endif

          !***********************************************************************
          ! 
          !
          !**********************************************************************           
          ! vny 2010-10-09 checking psf-photometry for all sources (the min number of neighbours=0)
          !call PSFPhotometry(int(npSources), int(neSources), sourceData, esourceData, &
          !     int(8 / binx, int8), 2_int8, 0_int8)
          !
          !call PSFPhotometry(int(npSources), int(neSources), sourceDataPsf, esourceDataPsf, &
          !     int(8 / binx, int8), 2_int8, 1_int8)
          call PSFPhotometry(image, int(npSources), int(neSources), sourceDataPsf, esourceDataPsf, &
               int(8 / binx, int8), 2_int8, neighboursForPsfPhotometry)
           
       end if
       

       levelImage = 0
       !*******************************************************************
       !
       ! Check very bright sources for faint neighbours
       !
       !******************************************************************
       !call checkPointSources( nPSources, sourceData, image, plevelimage)
       call checkPointSources( nPSources, sourceData, image, plevelimage, llevelImage)
       !*********************************************************
       !
       ! Add the point sources to the final source linked-list
       !
       !*********************************************************

       !write(*,*) "-------------------------------------"
       !write(*,*) "L1965 frameTime=", frameTime
       !write(*,*) "-------------------------------------"
       do i = 1, nPsources  ! Add the sources to the final list
          if (sourceData(i)%rawCountRate <= 0.0) then
             
             sourceData(i)%rawCountRate = nullValue     
             sourceData(i)%corrCountRate = nullValue     
             sourceData(i)%rawCountRateErr = nullValue     
             sourceData(i)%corrCountRateErr = nullValue     
             
          end if
          
          if ((sourceData(i)%rawCountRate+sourceData(i)%rawBackGroundCountRate)*frameTime .gt. 0.97) then
             !write(*,*) "L1977 i=", i, " rawCountRate=", sourceData(i)%rawCountRate, &
             !     " raw bkg=", sourceData(i)%rawBackGroundCountRate, " exceeds 0.97 c/f"
             sourceData(i)%corrCountRate=nullValue
             sourceData(i)%corrCountRateErr=nullValue
          endif
          
          call addSourceToList(sourceList, sourceData(i))
          !**************************************************************
          ! Creating the second source list to store psf-photometry data
          !**************************************************************
          if (psfPhotometryEnabled) then
             call addSourceToList(sourceListPsf, sourceDataPsf(i))
          endif
       end do


       !*********************************************************
       !
       ! Add the extended psf-photometry sources to the final psf-source linked-list
       !
       !*********************************************************
       if (psfPhotometryEnabled) then
          if (neSources .gt. 0) then
             do i = 1, neSources  ! Add the extended sources to the final list
                if (esourceData(i)%rawCountRate <= 0.0) then
                   
                   esourceData(i)%rawCountRate = nullValue     
                   esourceData(i)%corrCountRate = nullValue     
                   esourceData(i)%rawCountRateErr = nullValue     
                   esourceData(i)%corrCountRateErr = nullValue     
                    
                end if
                 
                 
                !**************************************************************
                ! Creating the second source list to store psf-photometry data
                !**************************************************************
                call addSourceToList(sourceListPsf, esourceDataPsf(i))
                
             end do

          endif

          nSourcesPsf=npSources+neSources

          deallocate(sourceDataPsf)
          nullify(sourceDataPsf)
       
          if (neSources .gt. 0) then
             deallocate(esourceDataPsf)
             nullify(esourceDataPsf)
          endif
       endif

       deallocate(sourceData)
       nullify(sourceData)

       call DeleteSourceList(pSourceList)  
       
    end if ! npSources >0(aperturephotometry on the point-sources)

    where(pLevelImage > 0) levelImage = pLevelImage
     
    if (neSources .gt. 0) then  ! Photometry on extended sources
 
       ! Load the extended sources
       sourceData => SourceDataArray(eSourceList, neSources)
       
       sourceData%exposure = exposure ! The exposure length (seconds)
  
       !****************************************************
       ! Perform aperture photometry on the extended-sources
       !****************************************************
       write(errstr,'(a)') "Conducting photometry on the extended sources"
       call message(AppMsg, Verbose)
       call AperturePhotometry(nx, ny, image, backgroundImage, elevelImage, qualArray,  & 
            apertureRadiusLocal, neSources, sourceData)
 
       !*******************************************************************************
       !
       ! Check the extended-sources for proximity to very bright point-sources
       !
       !*******************************************************************************
       write(errstr,'(a)') "Checking quality of the extended sources"
       call message(AppMsg, Verbose)
       !call checkExtendedSources( npSources, sourcelist, neSources, sourceData, elevelimage)
       call checkExtendedSources( npSources, sourcelist, neSources, sourceData, &
            elevelimage, lLevelImage)

       do i = 1, nEsources ! Add the extended sources to the master source-list
          
          if (sourceData(i)%rawCountRate <= 0.0) then

             sourceData(i)%rawCountRate = nullValue     
             sourceData(i)%corrCountRate = nullValue     
             sourceData(i)%rawCountRateErr = nullValue     
             sourceData(i)%corrCountRateErr = nullValue     

          end if

          call addSourceToList(sourceList, sourceData(i))   

       end do
        
       deallocate(sourceData)
       nullify(sourceData)
       call DeleteSourceList(eSourceList)  ! Finished with extended source-list
 
       where(elevelImage .lt. 0) levelImage = elevelImage
    
    end if ! (nesources > 0)

    nSources = nPSources + nESources ! The total number of sources

    !**************************************************************
    !
    ! Do some more checks then set source flags
    !
    !**************************************************************

    if (nSources .gt. 0) then 
        
       ! Load the combined source-list 
       sourceData => SourceDataArray(SourceList, nSources)
       
       if (psfPhotometryEnabled) then
          sourceDataPsf => SourceDataArray(SourceListPsf, nSourcesPsf)
       endif

       !do i = 1, nSources
       !   write(*,*) "L1662 i=", i, "num=", sourceData(i)%number, " x=", sourceData(i)%xPos, &
       !        " y=", sourceData(i)%yPos, &
       !        " src=", sourceData(i)%rawCountRate, " bkg=", sourceData(i)%rawBackGroundCountRate
        !********************************************************
        ! Set source flag if source lies on the Jupiter patch
        ! which has bit 11 set
        !********************************************************
       qualArrayInt16 => GetQualityArrayInt16(oswDataSet, nx, ny)

       do i = 1, nSources
          !write(*,*) "L1662 i=", i, "num=", sourceData(i)%number, " x=", &
                 !sourceData(i)%xPos, " y=", sourceData(i)%yPos, &
               !" src=", sourceData(i)%rawCountRate, " bkg=", sourceData(i)%rawBackGroundCountRate
          if (sourceData(i)%xPos .gt. 0 .and. & 
                              sourceData(i)%yPos .gt. 0) then
            !write(*,*) "    Pixel flag: ",qualArrayInt16(int(sourceData(i)%xPos), int(sourceData(i)%yPos))

             if (qualArrayInt16(int(sourceData(i)%xPos), &
                               int(sourceData(i)%yPos)) .ge. 2**11) then
                sourceData(i)%qualityFlag = & 
                              ibset(sourceData(i)%qualityFlag, 11)
             endif 
          endif 
       enddo

       !if(associated(qualArrayInt16) ) then
        !  deallocate(qualArrayInt16)
         ! nullify(qualArrayInt16)
       !endif

       !*********************************************
       !
       !
       !********************************************

       if( ommodeType .ne. "MOSAICED") then
          
          call LookForReadOutStreaks1( sourceData, image)
       end if

       !************************************************************************
       ! Check for highly elongated sources along read-out streaks
       !************************************************************************
       call removeSpuriousStreakSources(nSources, sourceData, levelImage)
       
       call CheckSourceSignificances(nSources, sourceData, nx, ny, levelImage)
       
       call resetConfusionFlags(nSources, sourceData)

       !*********************************************
       !
       ! Set the quality flags
       !
       !*********************************************
       if( ommodeType .ne. "MOSAICED" ) then
          

          if (associated(plevelimage) .and. associated(elevelimage)) then
             call SetSourceFlags(nSources, sourceData, x0, y0, plevelimage, elevelimage)
          else 
             call SetSourceFlags(nSources, sourceData, x0, y0, levelimage)
          end if
       else
       
          !*********************************************************
          !
          ! Can only set bright-source proximity flags
          !
          !*********************************************************
          call SetSourceFlags2(nSources, sourceData)
              
       end if
        
    end if

    if(associated(exposureImage ) ) then
       deallocate(exposureImage)
       nullify(exposureImage)
    end if


    if( ommodeType .eq. "MOSAICED" ) then
       exposure = exposureConstant
    end if

    call delete ! Delete the detection arrays
    
    if (associated(qualArray)) then
       deallocate(qualArray)
       nullify(qualArray)
    end if
          
    if (associated(pLevelImage)) then
       
       deallocate(pLevelImage)
       nullify(pLevelImage)
    end if
 
    if (associated(eLevelImage)) then
       deallocate(eLevelImage)
       nullify(eLevelImage)
    end if
 
    if (associated(lLevelImage)) then
       deallocate(lLevelImage)
       nullify(lLevelImage)
    end if

    if (associated(qualMap)) then
       call release(qualMapH)
       deallocate(qualMap)
       nullify(qualMap)
       
    end if

    if (associated(backgroundSigmaImage)) then
       deallocate(backgroundSigmaImage)
       nullify(backgroundSigmaImage)
    end if

    nullify(pmaskImage, eMaskImage)

  end subroutine ImagingDetection

!************************************************
!
!
!**********************************************
subroutine GrismDetection(image,  backgroundimage, &
                          levelImage, nsigma,  &
                          nSources, sourceData)
 
    real(single), pointer :: image(:, :), backgroundimage(:, :)
    integer(int16), pointer  :: levelImage(:, :)
    real(single), intent(in) :: nSigma
    integer(int16), intent(out) :: nSources
    type(sourceParameters), pointer :: sourceData(:)
 
    type(SourceLinkedList) :: sourceList 
   
    integer(int16) :: i, nx, ny

    real(single), pointer   :: tempImage(:, :) => Null()

    real(single), pointer   :: backgroundSigmaImage(:, :) => Null()
 
   
    integer(int8) :: minCounts, extensionFlag
    integer(int16) :: nSources1
    integer(int8) :: iter
    integer(int16), pointer :: pLevelImage(:, :) => null()  
 
    integer(int32) :: minNoPixels

    real(single) :: height ! , modeBackground, sigma 
       
    nx = size(image, 1)
    ny = size(image, 2)

    minCounts = 3    
    tempImage => Allocate2DSingle(0, nx - 1, 0, ny - 1, "tempImage")
    tempImage = image


    BackgroundImage => Allocate2DSingle(0, nx - 1, 0, ny - 1, "backgroundImage")
    
    pLevelImage => Allocate2DInt16(0, nx - 1, 0, ny - 1, "plevelImage")
    pLevelImage = 0

    levelImage => Allocate2DInt16(0, nx - 1, 0, ny - 1, "levelImage")
    levelImage = 0
    backgroundSigmaImage => Allocate2DSingle(0, nx - 1, 0, ny - 1, "backgroundImage") 
  

    call createGrismBackgroundImage(image, int(40 / binX, int16), int(10 / binY, int16) , &
               backgroundImage, backgroundsigmaImage,         &
               levelimage, 10000.0)


    !***********************************************************
    !
    !
    !***********************************************************
    call setBackgroundSigmaImagePtr(backgroundsigmaImage)
    

    call SmoothGrismImage(tempImage, int(4 / binX, int16), int(12 / binY, int16))

    where(image .le. 0.0) 
        levelimage = 1_int16
    end where

                       
    extensionFlag = 1
  
    iter = 6
 

    write(errstr,'(a)') "Running GRISM mode source detection algorithm"
    call message(AppMsg, Verbose)

  

    call settitle( "first-order grism-sources" )
   
    write(errstr,'(a)') "Looking for first-order spectra"
    call message(AppMsg, Verbose)

    do i = 1, iter
 

!        write(errstr,'(a, i2)') "Iteration ", i
!        call message(AppMsg, Verbose)
   
     
        height = nsigma + real(i - 1) * 0.5

        call DetectExtendedSources(nx, ny, tempImage, image, backgroundImage, &
                     maxval(image), pLevelImage, levelImage, sourceList, 1_int16, &
                     nSources, 200 / binX, &  ! was 400
                     100000 / binX, &  !            was 50 / binx then 20 / ninx, then 6                                  
                     minCounts, height, extendedheightFunction, IsValidFirstOrderSource, &
                     extensionFlag, .false.)  

        where(levelImage .gt. 1_int16) levelImage = 0    
     
    end do

    nSources1 = nSources
    write(errstr,'(a, i0)') "Number detected =", nSources
    call message(AppMsg, Verbose)
  

    call settitle( "zeroth-order grism-sources" )

    call SmoothLevelImage3(levelImage, 2 / binx, 20 / binX)
    tempImage = image

    call SmoothGrismImage(tempImage, int(6 / binX, int16), int(8 / binY, int16))

    where(image .le. 0.0) 
        levelimage = 1_int16
    end where

 
    iter = 6
    write(errstr,'(a)') "Looking for zeroth-order spectra"
    call message(AppMsg, Verbose)



    minNoPixels = 20 / binx
    do i = 1, iter

!       write(errstr,'(a, i2)') "Iteration ", i
!       call message(AppMsg, Verbose)
   
        height = nsigma + real(i - 1) * 0.5


        call DetectExtendedSources(nx, ny, tempImage, image, backgroundImage,               &
                     maxval(image), pLevelImage, levelImage, sourceList, 1_int16, nSources, &
                     minNoPixels,                                                           &
                     5000 / binX,                                                           &                                   
                     minCounts, height, extendedheightFunction, IsValidZeroOrderSource,     &
                     extensionFlag, .false.) 

   
        where(levelImage .gt. 1_int16) levelImage = 0    


    end do

    where(levelimage .gt. 0) levelimage= 0 
    write(errstr,'(a, i0)') "Number detected =", (nsources - nSources1)
    call message(AppMsg, Verbose)


    deallocate(backgroundsigmaimage, pLevelImage, tempImage)
    nullify(backgroundsigmaimage, pLevelImage, tempImage)

    if (nsources .gt. 0) then
        sourceData => SourceDataArray(sourceList, nSources)
        call DeleteSourceList(sourceList)

    end if

  
end subroutine GrismDetection


!**********************************************
!
!**********************************************
subroutine SmoothGrismImage(image, dx, dy)
    real(single), pointer :: image(:, :)
    integer(int16), intent(in) :: dx, dy
    real(single), pointer :: tempImage(:, :) => null()
    real(single) :: fMax
    integer(int16) :: i, j, k, l, nx, ny
    
    
    nx = ubound(image, 1)
    ny = ubound(image, 2)

    tempImage => Allocate2DSingle(0_int16, nx, 0_int16, ny, "tempImage")

    tempImage = image


    do j = 0, ny
        do i = 0, nx
            if (image(i, j) .lt. 0.0) cycle
            fMax = 0.0
            do k = max(0_int16, i - dx), min(nx, i + dx)
                do l = max(0_int16, j - dy), min(ny, j + dy)
                    
                    fMax = max(fMax, tempImage(k, l))
                end do
            end do
            image(i, j) = fMax


        end do
    end do


    deallocate(tempImage)
    nullify(tempImage)


end subroutine SmoothGrismImage

!*******************************************
!  Given the total-counts, x, the 
! background-counts, bck, the signal-to-noise,
! sn, the function returns the difference between 
! the required signa-to-noise and the computed 
! one from the counts
!*******************************************
real(double) Function SZero(x)
    real(double), intent(in) :: x

    SZero = sn - (x - bck) / sqrt(x)

end function SZero

!************************************************************
!
!
!************************************************************
subroutine MergeSourceLists(inputList, nSourcesIn, outputList)
    type(SourceLinkedList), intent(inout) :: inputList 
    integer(int16), intent(inout) :: nSourcesIn
    type(SourceLinkedList), intent(inout) :: outputList 
  
    type(sourceParameters), pointer :: sourceData(:) => null()
    integer(int16) :: i

    ! Load the combined source-list 
    sourceData => SourceDataArray(inputList, nSourcesIn)
    do i = 1, nSourcesIn
        call addSourceToList(outputList, sourceData(i))   
    end do

    deallocate(sourceData)
    nullify(sourceData)
  
end subroutine MergeSourceLists

!**********************************************************************
!
!
!**********************************************************************
subroutine ReclassifyExtendedSources(nESources, eSourceList, image, backgroundImage, levelImage)
    integer(int16), intent(inout) :: neSources
    type(SourceLinkedList), intent(inout) :: eSourceList 

  
    real(single), pointer, dimension(:, :) :: image, backgroundImage
    integer(int16), pointer, dimension(:, :) :: levelImage
   

    integer(int16) :: i, n, npSources
    integer(int16) :: nx, ny
    real(single), parameter :: nstd = 3.0 ! Was 3

    type(sourceParameters), pointer :: source(:) => null()

    type(sourceParameters) :: tempSource

    if (.not. associated(source)) return
    write(errstr,'(a)') "Checking classification of extended sources"
    call message(AppMsg, Verbose)
 

    n = size(source)
    nx = ubound(image, 1)
    ny = ubound(image, 2)

    npSources = maxval(levelImage) ! The number of point sources
  
    
    ! Load the combined source-list 
    source => SourceDataArray( eSourceList, neSources)

    ! Delete the list
    call DeleteSourceList( eSourceList )  

  
    do i = 1, n
        if (source(i)%noSourcePixels < 6 .or. btest(source(i)%qualityflag, 2) ) cycle


         tempSource = source(i)

         source(i)%rawCountRate = -maxval( image(source(i)%imin : source(i)%imax, source(i)%jmin :: source(i)%jmax), &
                     
                        levelImage(source(i)%imin : source(i)%imax, source(i)%jmin :: source(i)%jmax) &
                                 .eq. source(i)%number )
                                         
         call ComputeExtendedSrcParameters2(int(source(i)%number, int16), nx, ny, image, backgroundImage, &
                levelImage, source(i), .true.)  ! True for Monte-carlo errors

         if( abs(source(i)%xpos - tempSource%xpos) > real(binx / 4) .or. &
             abs(source(i)%ypos - tempSource%ypos) > real(binx / 4)) then



             source(i)%xpos = tempSource%xpos
             
             source(i)%ypos = tempSource%ypos
            

         end if
 

        if ( source(i)%semia .le. 0.0) cycle

        ! Check if originally classified as an extended source
        if (source(i)%extensionFlag .eq. 1) then

 
            if ((source(i)%semia - psfparams%psfsemia) .le. (nstd * source(i)%semiaErr)) then


                source(i)%extensionFlag = 0
                source(i)%noSourcePixels = 0

                
                npSources = nPSources + 1
                
                write(errstr,'(a, i0, a, f0.2, a, f0.2, a)') "Reclassified source", &
                source(i)%number, " (", source(i)%xpos, ",", source(i)%ypos, ")"
                call message(AppMsg, Verbose)


            end if
        end if 
        !**********************************************************
        ! Add to the temporary list
        !**********************************************************
        call addSourceToList(eSourceList, source(i))

    end do ! Finished looping through the sources
    
    deallocate(source)
    nullify(source)
   
end subroutine reclassifyExtendedSources

!**********************************************************************
! ReclassifyPointSources
!
!
!**********************************************************************
subroutine ReclassifyPointSources(source, image, backgroundImage, &
         plevelImage, elevelImage, npSources, eSourceList, neSources)
    type(sourceParameters), pointer :: source(:)
    real(single), pointer, dimension(:, :) :: image, backgroundImage
    integer(int16), pointer, dimension(:, :) :: plevelImage, elevelImage
    type(SourceLinkedList), intent(inout) :: esourceList ! For extended sources 
 
    integer, intent(inout) :: npSources, neSources
    real(single) :: meanSemia
    real(single), pointer :: data(:) => null()
    integer(int16) :: i, j
    integer(int16) :: n, nx, ny
    integer :: i0, j0
    integer(int16) :: k, dx
   
    integer(int16) :: oldSourceNumber
    real(single), parameter :: nstd = 3.0
  
    if (.not. associated(source) .or. .not. associated(elevelimage)) return
    write(errstr,'(a)') "Checking classification of point sources"
    call message(AppMsg, Verbose)
   

    nx = ubound(image, 1)
    ny = ubound(image, 2)
    dx = 20 / binx
    
    !**************************************************************
    ! Compute the median semi-a values
    !*************************************************************
    n = npsources

  
    if (n .gt. 10) then
        data => allocateSingle(1_int16, n, "data")
        data = source%semia
        
        meanSemia = MedianValue1(data, int(n))
        write(errstr,'(a, f0.3)') "Initial median source semi-a=", meanSemia
        call message(AppMsg, Verbose)


        write(errstr,'(a, f0.3, a, f0.3)') "Minimum semi-a=", minval(data), " max=", maxval(data)
        call message(AppMsg, Verbose)
 

        deallocate(data)
        nullify(data)
    else
        meanSemia = psfParams%psfsemia
    end if

  

    npsources = 0

    do i = 1, n
   

        ! Check if originally classified as a point source but now probably 
        if (source(i)%extensionFlag .eq. 1 .or. source(i)%confusionFlag .gt. 1) cycle

           
        !***********************************************************************************
        ! Recompute the source parameters     
        !***********************************************************************************
        call ComputeExtendedSrcParameters2(int(source(i)%number, int16), nx, ny, image, backgroundImage, &
                plevelImage, source(i), .true.)  ! True for Monte-carlo errors


        source(i)%semiaErr = max(0.1,  source(i)%semiaErr) ! Was 0.02
        source(i)%semibErr = max(0.1,  source(i)%semibErr)

         
        ! Check classification


        i0 = int(source(i)%xpos - 0.5)
        j0 = int(source(i)%ypos - 0.5)

       
        if ((image(i0, j0) - backgroundImage(i0, j0)) .lt. 5000.0 .and. & 
            (source(i)%semia - psfparams%psfsemia) .gt. (nstd * source(i)%semiaErr)) then


 
            source(i)%extensionFlag = 1

            write(errstr,'(a, i0, a, f0.2, a, f0.2)') "Reclassified source", source(i)%number, &
                    " x=", source(i)%xpos, " y=", source(i)%ypos
            call message(AppMsg, Verbose)
            
            oldSourceNumber = source(i)%number
            source(i)%number = -neSources 


            do j = source(i)%imin, source(i)%imax                                          
                do k = source(i)%jmin, source(i)%jmax
                    if (plevelImage(j, k) .ne. oldSourcenumber) cycle
                    pLevelImage(j, k) = 0
                    if (elevelImage(j, k) .ne. 0) cycle
                    eLevelImage(j, k) = source(i)%number
                end do
            end do

            ! Add to the extended-source list 
            call addSourceToList(esourceList, source(i))
            nesources = nesources + 1

 
        else
            ! Add to point-source array
            npSources = npSources + 1
            source(npSources) = source(i)

        end if 
        
    end do

    n = count(source%extensionFlag .eq. 0)

    if (n .gt. 10) then
        data => allocateSingle(1_int16, n, "data")
        k = 0
        do i = 1, n
            if (source(i)%extensionFlag .eq. 0) then
                k = k + 1 
                data(k) = source(i)%semia
            end if 
        end do
   
  

        meanSemia = MedianValue1(data, int(n))
        write(errstr,'(a, f0.3)') "Recomputed median source semi-a=", meanSemia
        call message(AppMsg, Verbose)

        write(errstr,'(a, f0.3, a, f0.3)') "Minimum semi-a=", minval(data), " max=", maxval(data)
        call message(AppMsg, Verbose)
        deallocate(data)
        nullify(data)
    end if 


end subroutine reclassifyPointSources

!****************************************************
! Subroutine RealImage
!****************************************************
subroutine SaveBackgroundImage( image)
    
 
    real (single), pointer :: image(:, :)
    character(256) :: fileName
    type(dataSetT) :: backgroundImageSet, imageSet     
 

    fileName = trim(stringParameter("backgroundimage"))



    backgroundImageSet = DataSet(fileName, Create)

    
    call AddImageArray(backgroundImageset, "PRIMARY", image)


    fileName = trim(stringParameter("set"))

    imageSet = DataSet(fileName, read)



    call copyAllAttributes(attributable(imageSet), attributable(backgroundImageSet))
  

    call release(imageSet)
    call release(backgroundimageSet)

end subroutine SaveBackgroundImage

!****************************************
! Subroutine AddImageArray
!****************************************
subroutine AddImageArray(set, name, image)
    type (DataSetT), intent(in) :: set
 
    character(*), intent(in) :: name
 
    real(single), intent(in) :: image(:, :)
  
    type (ArrayT) :: arr
    real (single), dimension(:, :), pointer :: optr
 

    integer :: dim(0 : 1)
    dim(0) = size(image, 1)
    dim(1) = size(image, 2)
 

    arr = addArray(set, name, REAL32, dimensions = dim)
    optr=>real32Array2Data(arr)
 
    optr = image
   
    ! Release dataset
    call release(arr)

end subroutine AddImageArray

!*********************************************************************
! Should be called after all large extended-sources have been found
! Finds large scattered light features
!*********************************************************************
subroutine MaskScatteredLightRegions(image,             &
                                     tempLevelImagePtr, &
                                     plevelImage,       &
                                     x0, y0)
 
    real(single), pointer, dimension(:, :) :: image
    integer(int16), pointer, dimension(:, :) :: tempLevelimagePtr, pLevelImage
    real(single), intent(in) :: x0, y0

    type(SourceLinkedList) :: tempesourceList ! For extended sources 

    integer(int32) :: minNoPixels, maxNoPixels, nList
    real(single) :: height
    
    integer(int8) :: minCounts, extensionFlag
    real(single) :: xcentre, ycentre, xc, yc
    integer(int16) :: nesources, nesources1
    integer(int16) :: i, j, i0, j0, il, iu, jl, ju, dx
    integer(int16) :: nx, ny
    real(single), pointer :: tempBackgroundImagePtr(:, :) => null()
    real(single) :: mode, sigma, maxHeight   
    real(single), pointer :: pixelList(:) => null()


    xc = size(image, 1) / real(2 * binx)  ! The x-centre of the image
    yc = size(image, 2) / real(2 * binx)  ! The y-centre of the image



    xCentre = (992.0 - x0)  / real(binX) ! The x-centre of central-background scattered-light feature
    yCentre = (1063.0 - y0) / real(binY) 

 
    if ((xcentre - xc) .lt. 10.0 .or. (xcentre - xc) .gt. size(image, 1) - 10.0) return
    if ((ycentre - yc) .lt. 10.0 .or. (ycentre - yc) .gt. size(image, 2) - 10.0) return

    nx = size(image, 1)

    ny = size(image, 2)

    i0 = int(xcentre - 0.5)
    j0 = int(ycentre - 0.5)



    dx = 300 / binx ! Was 300


    il = max(int(0, int16), i0 - dx)
    iu = min(int(nx - 1_int16, int16), i0 + dx)
    jl = max(int(0, int16), j0 - dx)
    ju = min(int(ny - 1_int16, int16), j0 + dx)
   

    ! Compute the background mode in this central region
    pixelList => AllocateSingle(1, (iu - il  + 1) * (ju - jl + 1), "list")

    nlist = 0

    do j = jl, ju
        do i = il, iu

            if (image(i, j) .ge. 0.0 .and. .not. isNullValue(image(i, j)) ) then
                nList = nList + 1
                pixelList(nlist) = image(i, j)
            end if
        end do
    end do

    if(nList .lt. 100) then
        deallocate( pixelList )
        nullify( pixelList )
        return 

    end if
    
 
    !*********************************************************************
    !
    !*******************************************************************
    height = 3.0 ! was 3
    
    minnopixels = 1000 / binx
    maxNoPixels = 100000 / binx
    mincounts = 1
    nesources1 = 0
    nesources = abs(minval(templevelimageptr))
    call settitle( " " )


    where(templevelimageptr(il : iu, jl : ju) .gt. 0) templevelimageptr(il : iu, jl : ju) = 0
 
    tempBackgroundImagePtr => allocate2dSingle(0_int16, nx - 1_int16, 0_int16, ny - 1_int16, "tempBackgroundImagePtr")


    mode = BackgroundMode( nList, pixelList, 0.2, 2.0, sigma )  !

    deallocate( pixelList )
    nullify( pixelList )
    
    tempbackgroundImagePtr = mode
   
    maxHeight = mode + 3.0 * sigma


    do i = 1, 2
        call DetectExtendedSources(nx, ny, il, iu, jl, ju, image, image, &
                       tempbackgroundImagePtr, maxHeight, pLevelImage, tempLevelImagePtr,     &   
                       tempeSourceList, 0_int16, nESources1, minNoPixels, maxNoPixels,        &      
                       minCounts,  height, heightFunction5, IsExtendedSourceNoChecks1,        &
                       extensionFlag, .false.)

        
        where(tempLevelImagePtr(il : iu, jl : ju) .gt. 0) templevelImagePtr(il : iu, jl : ju) = 0
        if (nesources1 .gt. 0) exit

        height = 2.5
    end do

    deallocate(tempBackgroundImagePtr)
    nullify(tempBackgroundImagePtr)


    if (nesources1 .gt. nesources) then

        call DeleteSourceList(tempesourcelist)



        dx = 300 / binx ! Was 300


        il = max(int(0, int16), i0 - dx)
        iu = min(int(nx - 1_int16, int16), i0 + dx)
        jl = max(int(0, int16), j0 - dx)
        ju = min(int(ny - 1_int16, int16), j0 + dx)
   


        where(tempLevelImagePtr(il : iu, jl : ju) .lt. -nesources) 
            pmaskImage(il : iu, jl : ju) = .true.    
            emaskImage(il : iu, jl : ju) = .true.     
            tempLevelImagePtr(il : iu, jl : ju) = 0    
 
        end where


    

    end if


    where(emaskImage(il : iu, jl : ju) .and. templevelimagePtr(il : iu, jl : ju) .eq. 0) templevelimagePtr(il : iu, jl : ju) = 1


end subroutine MaskScatteredLightRegions

!*********************************************************************
! Should be called after all large extended-sources have been found
! Finds large scattered light features
!*********************************************************************
subroutine MaskScatteredLightRegions2(image,             &
                                     tempLevelImagePtr, &
                                     plevelImage,       &
                                     llevelImage, &
                                     x0, y0)
 
    real(single), pointer, dimension(:, :) :: image
    integer(int16), pointer, dimension(:, :) :: tempLevelimagePtr, pLevelImage
    integer(int16), pointer, dimension(:, :) :: lLevelImage
    real(single), intent(in) :: x0, y0

    type(SourceLinkedList) :: tempesourceList ! For extended sources 

    integer(int32) :: minNoPixels, maxNoPixels, nList
    real(single) :: height
    
    integer(int8) :: minCounts, extensionFlag
    real(single) :: xcentre, ycentre, xc, yc
    integer(int16) :: nesources, nesources1
    integer(int16) :: i, j, i0, j0, il, iu, jl, ju, dx
    integer(int16) :: nx, ny
    real(single), pointer :: tempBackgroundImagePtr(:, :) => null()
    real(single) :: mode, sigma, maxHeight   
    real(single), pointer :: pixelList(:) => null()

    !write(*,*) "L2842 ..."
    xc = size(image, 1) / real(2 * binx)  ! The x-centre of the image
    yc = size(image, 2) / real(2 * binx)  ! The y-centre of the image

    xCentre = (992.0 - x0)  / real(binX) ! The x-centre of central-background scattered-light feature
    yCentre = (1063.0 - y0) / real(binY) 

 
    if ((xcentre - xc) .lt. 10.0 .or. (xcentre - xc) .gt. size(image, 1) - 10.0) return
    if ((ycentre - yc) .lt. 10.0 .or. (ycentre - yc) .gt. size(image, 2) - 10.0) return

    nx = size(image, 1)

    ny = size(image, 2)

    i0 = int(xcentre - 0.5)
    j0 = int(ycentre - 0.5)


    dx = 300 / binx ! Was 300


    il = max(int(0, int16), i0 - dx)
    iu = min(int(nx - 1_int16, int16), i0 + dx)
    jl = max(int(0, int16), j0 - dx)
    ju = min(int(ny - 1_int16, int16), j0 + dx)
   

    ! Compute the background mode in this central region
    pixelList => AllocateSingle(1, (iu - il  + 1) * (ju - jl + 1), "list")

    nlist = 0

    do j = jl, ju
        do i = il, iu

            if (image(i, j) .ge. 0.0 .and. .not. isNullValue(image(i, j)) ) then
                nList = nList + 1
                pixelList(nlist) = image(i, j)
            end if
        end do
    end do

    if(nList .lt. 100) then
        deallocate( pixelList )
        nullify( pixelList )
        return 

    end if
    
 
    !*********************************************************************
    !
    !*******************************************************************
    height = 3.0 ! was 3
    
    minnopixels = 1000 / binx
    maxNoPixels = 100000 / binx
    mincounts = 1
    nesources1 = 0
    nesources = abs(minval(templevelimageptr))
    call settitle( " " )


    where(templevelimageptr(il : iu, jl : ju) .gt. 0) templevelimageptr(il : iu, jl : ju) = 0
 
    tempBackgroundImagePtr => allocate2dSingle(0_int16, nx - 1_int16, 0_int16, ny - 1_int16, &
         "tempBackgroundImagePtr")


    mode = BackgroundMode( nList, pixelList, 0.2, 2.0, sigma )  !

    deallocate( pixelList )
    nullify( pixelList )
    
    tempbackgroundImagePtr = mode
   
    maxHeight = mode + 3.0 * sigma


    do i = 1, 2
        call DetectExtendedSources(nx, ny, il, iu, jl, ju, image, image, &
                       tempbackgroundImagePtr, maxHeight, pLevelImage, tempLevelImagePtr,     &   
                       tempeSourceList, 0_int16, nESources1, minNoPixels, maxNoPixels,        &      
                       minCounts,  height, heightFunction5, IsExtendedSourceNoChecks1,        &
                       extensionFlag, .false.)
        
        where(tempLevelImagePtr(il : iu, jl : ju) .gt. 0) templevelImagePtr(il : iu, jl : ju) = 0
        if (nesources1 .gt. 0) exit

        height = 2.5
    end do

    deallocate(tempBackgroundImagePtr)
    nullify(tempBackgroundImagePtr)


    if (nesources1 .gt. nesources) then

        call DeleteSourceList(tempesourcelist)

        dx = 300 / binx ! Was 300

        il = max(int(0, int16), i0 - dx)
        iu = min(int(nx - 1_int16, int16), i0 + dx)
        jl = max(int(0, int16), j0 - dx)
        ju = min(int(ny - 1_int16, int16), j0 + dx)
   
        where(tempLevelImagePtr(il : iu, jl : ju) .lt. -nesources) 
            pmaskImage(il : iu, jl : ju) = .true.    
            emaskImage(il : iu, jl : ju) = .true.     
            tempLevelImagePtr(il : iu, jl : ju) = 0    
 
        end where


    

    end if


    where(emaskImage(il : iu, jl : ju) .and. templevelimagePtr(il : iu, jl : ju) .eq. 0) &
         templevelimagePtr(il : iu, jl : ju) = 1

    where(lLevelImage(il : iu, jl : ju) .gt. 0) &
         templevelimagePtr(il : iu, jl : ju) = -1

end subroutine MaskScatteredLightRegions2

!********************************************************************
!
!
!
!
!*******************************************************************
subroutine FindSmokeRings(pSourceList, npsources, x0, y0, &
                          image, eSourceList, neSources,  &
                          eLevelImage, plevelImage,       &
                          qualityImage)

 
    integer(int16), intent(inout) :: npSources
    integer, intent(in) :: x0, y0

    real(single), pointer, dimension(:, :) :: image

    type(SourceLinkedList), intent(inout) :: eSourceList ! For extended sources 
    integer(int16), intent(inout) :: neSources

    type(SourceLinkedList), intent(inout) :: pSourceList

    integer(int16), pointer, dimension(:, :) :: eLevelimage, plevelimage
    logical(boolean), pointer, dimension(:, :) :: qualityImage
 

    integer(int16) :: i, k, i0, j0, il, iu, jl, ju, nx, ny
    integer(int16) :: i0Source, j0Source  
    real(single) ::  xSmokeRing, ySmokeRing, rRing, r0, x, y, rStar, theta
    real(single) :: dx, dx1
    integer :: noSmokeRings   
    integer(int32) :: minCounts
    integer(int8), parameter :: noOffsets = 5
    real(single), parameter :: offset(noOffsets) = (/0.0, -2.5, 2.5, -5.0, 5.0/)  
    type(sourceParameters), pointer :: source(:) => null()

 

    if (npSources .lt. 1) return 
 
    write(errstr,'(a)') "Looking for smoke rings"
    call message(AppMsg, Verbose)




    dx = real(2 / binx)

    !*********************************************************************
    !
    !********************************************************************
  
    nx= ubound(image, 1)
    ny= ubound(image, 2)

!call saveImage(plevelimage, nx+1_int16, ny+1_int16, "plevelimage.fits", "background image")
!call fatal("stopped line 2297 displaying plevelimage.fits")


    call settitle( " " )


    !****************************************************
    ! Load the point sources from the linked list
    !****************************************************    
    source => SourceDataArray(psourceList, npSources)
    
    dx1 = 70 / binx ! was 70

    ! Loop through the sources to find any that may produce a smoke-ring


    noSmokeRings = 0
 
    do i = 1, npsources  

!if (i .ne. 4) cycle



!if (abs(source(i)%xpos - 95.0) .gt. 10.0) cycle

!write(*, *) "source ", i, " quality=", source(i)%qualityFlag, " x=", &
!             source(i)%xpos, " y=", source(i)%ypos

        if (.not. btest(source(i)%qualityFlag, 4)) cycle      

        ! The source is predicted to produce a smoke-ring 
        ! Predict where smoke-ring is
        ! Radial distance of star from centre of full-frame image
        ! xCentre and yCentre are defined in omconstants_mod (1024.5. 1024.5)


!write(*, *) "Checking for smoke ring from source ", i, " x=", source(i)%xpos, " y=", source(i)%ypos                 
               
        x = source(i)%xpos
        y = source(i)%ypos
        i0Source = nint(x - 0.5)
        j0Source = nint(y - 0.5)


        call ConvertCoordsToFullFrame(x, y, x0, y0, Binx, Biny)
    

        !***************************************************************
        ! Radial distance of star from centre and its position angle
        !***************************************************************
        rStar = RadialDistance( xCentre, yCentre, dble(x), dble(y) )
        if (rStar .lt. 60 / binx) cycle



        !****************************************************************
        ! The predicted position of the smoke ring
        !***************************************************************

        
        xSmokeRing = smokeRingXPosition(x, y) 
        ySmokeRing = smokeRingYPosition(x, y)

!write(*, *) "Line 2990 xstar=", x, " ystar=", y, " xsmoke=", xsmokering, " ysmoke=", ysmokering

        r0 = RadialDistance( xCentre, yCentre, dble(xSmokeRing), dble(ySmokeRing) )



        theta = PositionAngle( xCentre, yCentre, dble(xSmokeRing), dble(ySmokeRing) )

!write(*, *) "rstar=", rstar, " rring=", r0, " theta=", theta * radiansToDegrees

        do k = 1, noOffsets  

            rRing = r0 + offset(k)   
            if(abs(rRing - rStar) < real(20 / binx)) cycle

            xSmokeRing = xCentre + rRing * cos(theta)
            ySmokeRing = yCentre + rRing * sin(theta)
              
      
            !***********************************************************
            ! Set up detection bounds
            !***********************************************************

            call ConvertCoordsFromFullFrame(xsmokering, ysmokering, x0, y0, Binx, Biny)

!        write(*, *) "predicted smoke ring position - xsmokering=", xsmokering, " y=", ysmokering

            i0 = nint(xsmokering - 0.5)
            if (i0 < 0 .or. i0 > nx) cycle
            j0 = nint(ysmokering - 0.5)

            if (j0 < 0 .or. j0 > ny) cycle
  
              
!            write(*, *) "Line 2896 i0=", i0, " j0=", j0, " i0source=", i0Source, " j0source=", j0Source


            il = i0 - dx1

            !******************************************************************
            ! Check to see if smoke ring is close to the source
            !***************************************************************** 
        

            if (il .lt. 0) then
                il = 0
          
            else if (il > (nx - 40 / binx)) then 
                cycle 
            end if
 
     
            iu = il + 2 * dx1 
      

            if (iu .gt. nx) then
                iu = nx
            end if  
       

            jl = j0 - dx1

      
            if (jl .lt. 0) then
                jl = 0
          
            else if (jl > (ny - 40 / binx)) then 
                cycle 
            end if
 
            ju = jl + 2 * dx1
            if (ju .gt. ny) then
                ju = ny
            end if
     
 
            mincounts = (iu - il + 1) * (ju - jl + 1)

!write(*, *) "Line 2965 il=", il, " iu=", iu, " jl=", jl, " ju=", ju       


            if ( count(elevelImage(il : iu, jl : ju) .ne. 100) < mincounts / 2) cycle 

            if ( isSmokeRing(source(i)%number, image, plevelimage, elevelImage, &
                      qualityImage, il, iu, jl, ju, esourceList, neSources) ) then

                !****************************************************************************************
                !
                ! Found a smoke ring
                !
                !****************************************************************************************            



!write(*, *) "***********************************************************"
!write(*, *) "omdetect - find smoke rings- nesources=", nesources, " sources came from source ", i, &
! " x=", source(i)%xpos, " y=", source(i)%ypos
!write(*, *) "***********************************************************"
                noSmokeRings = noSmokeRings + 1
                where(elevelImage(il : iu, jl : ju) .lt. 0) 
                    pmaskimage(il : iu, jl : ju)  = .true.
                    emaskimage(il : iu, jl : ju) = .true.
                        
                end where
                    
!call saveImage(elevelimage, nx+1_int16, ny+1_int16, "elevelimage.fits", "background image")
!call fatal("stopped line 2949 displaying elevelimage.fits")

                !******************************************************************************
                !
                ! Record that a possible smoke-ring has been found attached to this star
                ! But remember to unset this later on
                !
                !******************************************************************************
                source(i)%qualityFlag = ibset(source(i)%qualityFlag, 2) 
                exit  

             end if    
         
         end do 
    end do 
      

    deallocate(source)
    nullify(source)
  

    
    write(errstr,'(a, i0)') "Number of possible smoke rings found=", noSmokeRings
    call message(AppMsg, Verbose)
 

!call saveImage(elevelimage, nx+1_int16, ny+1_int16, "elevelimage.fits", "background image")
!call fatal("stopped line 2988 displaying elevelimage.fits")

end subroutine FindSmokeRings


!*********************************************************
! If any sources have been picked up as extended in 
! the point-source list, transfer them to the extended source-list.
! Also modify the levelimage maps.
!
!*********************************************************
subroutine TransferESources(nPsources, pSourceList, &
                           plevelImage, nesources, eSourceList, elevelImage)

    ! Point-source variables
    integer(int16), intent(inout) :: nPsources
    type(SourceLinkedList), intent(inout) :: psourceList 
     integer(int16), pointer :: plevelImage(:, :)

    ! Extended-source variables
    integer(int16), intent(inout) :: nesources
    type(SourceLinkedList), intent(inout) :: esourceList  
    integer(int16), pointer :: elevelImage(:, :)
    integer(int16), pointer :: templevelImage(:, :) => null()
 
    integer :: nx, ny, n

    type(sourceParameters), pointer :: source(:) => null()
    integer :: i

!call checkLinkedList(nesources, eSourceList, "transferesources line 3138")

 
    if (npSources .eq. 0) return

    ! Load the point-sources
    source => SourceDataArray( pSourceList, npSources )

    !****************************************************************************
    ! Check to see that there is at least one sources in the point-source-list
    ! with an extension flag of one
    !****************************************************************************
    if ( count(source%extensionFlag .eq. 1) .eq. 0 ) then
        deallocate(source)
        nullify(source)
        return

    end if

  
    call DeleteSourceList( psourcelist ) 

    nx = ubound(plevelimage, 1)
    ny = ubound(plevelimage, 2)


    tempLevelImage => Allocate2DInt16(0, nx, 0, ny, "tempeLevelImage")
    n = npsources
    nPSources = 0

!write(*, *) "transferesources line 2987, nesources=", nesources

    !*******************************************************************************
    ! Loop through all the sources in the point-source list and
    ! transfer them to the extended source-list if the extension flag is 1
    !******************************************************************************
    do i = 1, n
!write(*, *) i, ") x=", source(i)%xpos, " y=", source(i)%ypos, " extensionflag=", source(i)%extensionFlag

        if (source(i)%extensionFlag .eq. 0) then ! Add to new point-source linked-list
        
            nPsources = npSources + 1
            where(plevelImage(source(i)%imin : source(i)%imax, source(i)%jmin : source(i)%jmax) .eq. source(i)%number) 
                templevelImage(source(i)%imin : source(i)%imax, source(i)%jmin : source(i)%jmax) = nPSources
            end where


            !*************************************************************
            ! Add source to the point-source list
            !*************************************************************
     
            source(i)%number = npSources
            call addSourceToList(pSourceList, source(i))

        else ! Add to extended-source linked-list
         
            neSources = neSources + 1 

            where(plevelImage(source(i)%imin : source(i)%imax, source(i)%jmin : source(i)%jmax) .eq. source(i)%number)
 
                  !*************************************************************************************
                  ! Add point-source to extended-source level image
                  !**************************************************************************************
         
                  elevelImage(source(i)%imin : source(i)%imax, source(i)%jmin : source(i)%jmax) = -neSources

                  !*************************************************************************************
                  ! Erase source from point-source level image
                  !**************************************************************************************
                  plevelImage(source(i)%imin : source(i)%imax, source(i)%jmin : source(i)%jmax) = 0
            end where
   
            source(i)%number = -neSources
            source(i)%noSourcePixels = count(elevelImage(source(i)%imin : source(i)%imax, &
                       source(i)%jmin : source(i)%jmax) .eq. source(i)%number)
   
            !*************************************************************
            ! Add source to the extended-source list
            !*************************************************************
            call addSourceToList(eSourceList, source(i))   
 
        end if

    end do
    deallocate(plevelImage)
    plevelimage => templevelimage


    if (n .ne. npsources) then
!        where(plevelImage > npsources) plevelImage = 0
    

        write(errstr,'(a, i0, a, i0)') "New number of likely point-sources=", npsources, " likely extended-sources=", nesources
        call message(AppMsg, Verbose)
    end if



!if (npsources > 0) then
!if (associated(source)) deallocate(source)
!write(*, *) "TRANSFERESOURCES ************* new list of point sources"
!    source => SourceDataArray( pSourceList, npSources )
!       do i = 1, npsources
!           write(*, *) i, ") x=", source(i)%xpos, " y=", source(i)%ypos, " extensionflag=", source(i)%extensionFlag
!       end do
!    if (associated(source)) deallocate(source)
!end if

!if (nesources > 0) then
!if (associated(source)) deallocate(source)
!    write(*, *) "************* new list of extended sources, nesources=", nesources
!    source => SourceDataArray( eSourceList, neSources )
!       do i = 1, nesources
!           write(*, *) i, ") x=", source(i)%xpos, " y=", source(i)%ypos, " extensionflag=", source(i)%extensionFlag
!       end do
!end if


    if(associated(source)) deallocate(source)
    nullify(source)




!call checkLinkedList(nesources, eSourceList, "transferesources line 3281")


  
end subroutine transferESources

!*********************************************************
!
! If any sources have been picked up as "point" in 
! the extended-source list, transfer them to the point source-list.
! Modify the levelimage maps as well.
!
!*********************************************************
subroutine TransferPSources(nPsources, pSourceList, &
                           plevelImage, nesources,  &
                           eSourceList, elevelImage)

    ! Point-source variables
    integer(int16), intent(inout) :: nPsources
    type(SourceLinkedList), intent(inout) :: psourceList 
     integer(int16), pointer :: plevelImage(:, :)

    ! Extended-source variables
    integer(int16), intent(inout) :: nesources
    type(SourceLinkedList), intent(inout) :: esourceList  
    integer(int16), pointer :: elevelImage(:, :)
  
 
    integer :: nx, ny, n

    type(sourceParameters), pointer :: source(:) => null()
    integer :: i
 
    if ( neSources .eq. 0 ) return

    ! Load the sources from the extended linked-list
    source => SourceDataArray(eSourceList, neSources)

    !****************************************************************************
    ! Check to see that there is at least one sources in the extended-list
    ! with an extension flag of zero
    !****************************************************************************
    
    if ( count(source%extensionFlag .eq. 0) .eq. 0 ) then
        deallocate(source)
        nullify(source)
        return

    end if


    !************************************************************************
    !
    !************************************************************************

    call DeleteSourceList( esourcelist ) ! Start afresh with a new extended-source list

    nx = ubound(plevelimage, 1)
    ny = ubound(plevelimage, 2)
  
    n = nEsources
    nESources = 0

    !*******************************************************************************
    ! Loop through all the sources in the point-source list and
    ! transfer them to the extended source-list if the extension flag is 1
    !******************************************************************************
    do i = 1, n

        if (source(i)%extensionFlag .eq. 0) then ! Transfer from extended source-list to point-source list
          
            nPsources = npSources + 1


            where(elevelImage(source(i)%imin : source(i)%imax, source(i)%jmin : source(i)%jmax) .eq. source(i)%number) 
                plevelImage(source(i)%imin : source(i)%imax, source(i)%jmin : source(i)%jmax) = nPSources
                !**********************************************************************************************
                ! Delete it from the extended level image
                !*********************************************************************************************
                elevelImage(source(i)%imin : source(i)%imax, source(i)%jmin : source(i)%jmax) = 0
        
            end where
            source(i)%number = npSources
            source(i)%noSourcePixels = 0
            call addSourceToList(pSourceList, source(i))


        else ! Add to extended-source linked-list
          
            neSources = neSources + 1 

            where(elevelImage(source(i)%imin : source(i)%imax, source(i)%jmin : source(i)%jmax) .eq. source(i)%number) 
                  elevelImage(source(i)%imin : source(i)%imax, source(i)%jmin : source(i)%jmax) = -neSources
  
            end where
   
            source(i)%number = -neSources
            call addSourceToList(eSourceList, source(i))   
 
        end if

    end do
    
    if (n .ne. npsources) then

        write(errstr,'(a, i0, a, i0)') "New number of likely point-sources=", npsources, " likely extended-sources=", nesources
        call message(AppMsg, Verbose)
    end if   

    deallocate(source)
    nullify(source)

    where(elevelImage < -nESources) elevelImage = 0


end subroutine transferPSources

!*********************************************************
!
! If any 2 extended sources overlap, merge them
! into a single source
!
!*********************************************************
subroutine MergeExtendedSources(nesources, eSourceList, &
          image, backgroundimage, elevelImage, minDr1)

    ! Extended-source variables

    integer(int16), intent(inout) :: nesources
    type(SourceLinkedList), intent(inout) :: esourceList  

    real(single), pointer, dimension(:, :) :: image, backgroundimage 
 
    integer(int16), pointer :: elevelImage(:, :)
 
    real(single), intent(in) :: minDr1
    integer(int16), pointer :: templevelImage(:, :) => null()
    real(single) :: dx, dy, radius, error, error1, error2
    real(single) :: d1, d2, d3, mindr, maxRadius, f1, f2, f3
    integer(int16) :: nx, ny, i0, j0, il, iu, jl, ju, dl
  
    type(sourceParameters), pointer :: source(:) => null()
    type(sourceParameters) :: tempSource

!    type(sourceParameters) :: Testsource

    integer :: i, j, k, n
   
    logical, pointer :: merged(:) => null()
    logical :: mergedP

   

!return  
!call checkLinkedList(nesources, eSourceList, "mergeextendedsources line 3450")

    if (neSources .lt. 2) return 

    maxRadius = real(100 / binx)
    dl = 4 / binX
 
    mergedP = .false.  ! Not yet merged any sources


    nx = ubound(elevelimage, 1)
    ny = ubound(elevelimage, 2)

    tempLevelImage => Allocate2DInt16(0_int16, nx, 0_int16, ny, "tempeLevelImage")
    allocate(merged(1 : nesources))
  

!*****************************************************************
! 
!*****************************************************************
!   write(*, *) "TESTS"
!   testSource%xpos = 100.0
!   testSource%ypos = 100.0  
!   testSource%semia = 10.0
!   testSource%semib = 5.0
!   testSource%theta = 0.0
!   write(*, *) "x=", testSource%xpos, " y=", testSource%ypos, &
!               " a=", testSource%semia, " b=", testSource%semib, " theta=", testSource%theta

 
  
!    d1 = sourceSeparation2(testsource, 110.0, 100.0)
!    write(*, *) "1) d1=", d1, " answer should be 1.0"

!   d1 = sourceSeparation2(testsource, 90.0, 100.0)
!    write(*, *) "2) d1=", d1, " answer should be 1.0"
     
!  d1 = sourceSeparation2(testsource, 100.0, 110.0)
!    write(*, *) "3) d1=", d1,  " answer should be 2.0"

!   d1 = sourceSeparation2(testsource, 100.0, 90.0)
!    write(*, *) "4) d1=", d1, " answer should be 2.0"

 
!  testSource%theta = 90.0

!    d1 = sourceSeparation2(testsource, 110.0, 100.0)
!    write(*, *) "5) d1=", d1, " answer should be 2.0"

!   d1 = sourceSeparation2(testsource, 90.0, 100.0)
!    write(*, *) "6) d1=", d1, " answer should be 2.0"

      
!  d1 = sourceSeparation2(testsource, 100.0, 110.0)
!    write(*, *) "7) d1=", d1, " answer should be 1.0"

!   d1 = sourceSeparation2(testsource, 100.0, 90.0)
!    write(*, *) "8) d1=", d1, " answer should be 1.0"


! testSource%theta = 180.0

!    d1 = sourceSeparation2(testsource, 110.0, 100.0)
!    write(*, *) "5) d1=", d1, " answer should be 1.0"

!   d1 = sourceSeparation2(testsource, 90.0, 100.0)
!    write(*, *) "6) d1=", d1, " answer should be 1.0"

      
!  d1 = sourceSeparation2(testsource, 100.0, 110.0)
!    write(*, *) "7) d1=", d1, " answer should be 2.0"

!   d1 = sourceSeparation2(testsource, 100.0, 90.0)
!    write(*, *) "8) d1=", d1, " answer should be 2.0"

 

!source => SourceDataArray(eSourceList, nESources)
 
!  do i = 1, neSources
!     write(*, *) i, ") x=", source(i)%xpos, " y=", source(i)%ypos, " number=", source(i)%number
!  end do
!deallocate(source)


    !****************************************************************
    ! Loop twice
    ! 
    !***************************************************************

    do k = 1, 2 ! was 2

        where(elevelimage .gt. 0_int16) 
            templevelimage = elevelimage
        elsewhere
            tempLevelImage = 0_int16
        end where

        n = neSources ! The original number of extended sources
       
        neSources = 0 ! The new number of extended sources

        !************************************************************   
        ! Load the extended-sources and then delete the linked-list
        !************************************************************
        source => SourceDataArray(eSourceList, n)

        call DeleteSourceList(esourcelist)  ! Delete source-list and start a new one

        merged = .false. ! No sources merged so far

!write(*, *) "Line 3555 nesources=", nesources, " n=", n, " k=", k

        !**********************************************************************************
        ! Loop through each source and check to see if it has a close neighbour
        ! that needs merging with it
        !**********************************************************************************
        do i = 1, n

           
!    write(*, *) i, ") line 2956 x=", source(i)%xpos, " y=", source(i)%ypos, &
!                  " extensionflag=", source(i)%extensionFlag 
 
                   
            if(merged(i)) cycle

            !************************************************************
            ! Get the maximum pixel value within the source
            !************************************************************
                
            f1 = maxval( image(source(i)%imin : source(i)%imax, source(i)%jmin : source(i)%jmax ), &
                    elevelimage(source(i)%imin : source(i)%imax, source(i)%jmin : source(i)%jmax)  &
                     .eq. source(i)%number)

            f1 = max(1.0, f1)

            neSources = neSources + 1 ! The new current source number
!            write(*, *) "Line 3587 nesources=", nesources


            !************************************************************
            ! 
            !
            !
            !***********************************************************
            where(elevelImage(source(i)%imin : source(i)%imax, &
                source(i)%jmin : source(i)%jmax) .eq. source(i)%number) 
                templevelImage(source(i)%imin : source(i)%imax, &
                               source(i)%jmin : source(i)%jmax) = -neSources
                                              
            end where

            tempSource = source(i)

            if( tempSource%extensionFlag .eq. 0) then ! Should never be the case
                tempSource%number = -nESources
                call addSourceToList(eSourceList, tempSource)   
                cycle
            end if
          
         
            error1 = source(i)%poserr


            do j = i + 1, n       
                if ( merged(j)) cycle

                if( source(j)%extensionFlag .eq. 0) cycle
                if( source(j)%rawCountRate / source(i)%rawCountRate .ge. 1.0) cycle 


                !************************************************************
                ! Get the maximum pixel value within the source
                !************************************************************
                f2 = maxval(image(source(j)%imin : source(j)%imax, source(j)%jmin : source(j)%jmax), &
                      elevelimage(source(j)%imin : source(j)%imax, source(j)%jmin : source(j)%jmax) &
                   .eq. source(j)%number)

                if( (f2 / f1) .ge. 5.0) cycle



           
!          write(*, *) j, ") line 3611 x=", source(j)%xpos, " y=", source(j)%ypos, &
!                  " extensionflag=", source(j)%extensionFlag 
 

                !*****************************************************************
                ! Compute the distance between source i and source j in terms
                ! of source(i)%semia, source(i)%semib and source(i)%theta
                !
                !*****************************************************************
                d1 = sourceSeparation2(source(i), source(j)%xPos, source(j)%yPos)
      


                !*****************************************************************
                ! Compute the distance between source j and source i in terms
                ! of source(j)%semia, source(j)%semib and source(j)%theta
                !
                !*****************************************************************
                d2 = sourceSeparation2(source(j), source(i)%xPos, source(i)%yPos)
      
                d3 = max(d1, d2)

                error2 = source(j)%poserr
               
                !****************************************************************
                !
                ! Compute the actual distance between them
                !
                !***************************************************************
                dx = source(i)%xpos - source(j)%xpos
                dy = source(i)%ypos - source(j)%ypos
                radius = sqrt(dx * dx + dy * dy)
                if(radius .gt. maxRadius) cycle

                error = 3.0 * sqrt(error1 * error1 + error2 * error2)
                


                if(abs(source(i)%theta - source(j)%theta) .lt. 20.0) then
                    minDr = max(3.0, minDr1)
                else
                    minDr = minDr1
                end if   

                if (error .gt. radius .or. d3 .lt. minDr) then ! Merge sources

!write(*, *) "Merging sources error=", error, " mindr=", mindr, " radius=", radius, " d3=", d3

    
!                    write(*, *) i, ")- x=", source(i)%xpos, " y=", source(i)%ypos, " eflag=", source(i)%extensionFlag, &
!                             " nopixels=", source(i)%noSourcePixels, " source%number=", source(i)%number, " d1=", d2
 

!                   write(*, *) j, ")- x=", source(j)%xpos, " y=", source(j)%ypos, " eflag=", source(j)%extensionFlag, &
!                                 " nopixels=", source(j)%noSourcePixels, source(j)%number, " d2=", d2
         
                    !***************************************************************
                    ! Set the source-number pixels on the new level image
                    !
                    !***************************************************************                
                    where(elevelImage(source(j)%imin : source(j)%imax, source(j)%jmin : &
                                      source(j)%jmax) .eq. source(j)%number) 
                        templevelImage(source(j)%imin : source(j)%imax, source(j)%jmin : &
                                      source(j)%jmax) = -neSources                         
                    end where

                    tempSource = source(i)                         
                    tempSource%number = -neSources
                    tempSource%imin = min(source(i)%imin, source(j)%imin)
                    tempSource%imax = max(source(i)%imax, source(j)%imax)
                    tempSource%jmin = min(source(i)%jmin, source(j)%jmin)
                    tempSource%jmax = max(source(i)%jmax, source(j)%jmax)
                    tempSource%noSourcePixels = source(i)%noSourcePixels + source(j)%noSourcePixels 
              
                    !****************************************************************
                    !
                    ! Compute extended-source parameters
                    !
                    !****************************************************************
                    call ComputeExtendedSrcParameters2(int(tempSource%number, int16), &
                          nx, ny, image, backgroundImage,                             &
                        templevelImage, tempSource, .true.)   ! True for Monte-carlo errors

                     
                    !*****************************************************************
                    !
                    ! Check the centre of the new source
                    !
                    !***************************************************************** 
                    i0 = max(0_int16, int(tempSource%xpos - 0.49999, int16))
                    i0 = min(nx, i0)
                    j0 = max(0_int16, int(tempSource%ypos - 0.49999, int16))
                    j0 = min(ny, j0)
                    il = max(0_int16, i0 - dl)
                    iu = min(nx, i0 + dl)
                    jl = max(0_int16, j0 - dl)
                    ju = min(ny, j0 + dl)

                    f3 = maxval(image(il : iu, jl : ju))
 
                    if(f3 .ge. max(f1, f2)) then   

!write(*, *) "line 3121 f1=", f1, " f2=", f2, " f3=", f3, " computed x=", tempSource%xpos, " y=", tempSource%ypos
 
                        mergedP = .true.

                        merged(i) = .true. ! Remember source has been merged
                        merged(j) = .true. ! ditto

                        write(errstr,'(a, i0, a, i0)') "Merging sources ", i, " and ", j
                        call message(AppMsg, Verbose)

                    else
                        !*********************************************
                        !
                        ! Not merged - remove source j from level image
                        !
                        !*********************************************                  
                        where(elevelImage(source(j)%imin : source(j)%imax, source(j)%jmin : &
                                      source(j)%jmax) .eq. source(j)%number) 
                            templevelImage(source(j)%imin : source(j)%imax, source(j)%jmin : &
                                      source(j)%jmax) = 0_int16                         
                        end where
                    end if  

!                    exit !
                end if        
            end do 
!            write(*, *) "Line 3698 nesources=", nesources   

            !*****************************************
            ! Add source to list
            !*****************************************  
            tempSource%number = -nESources
            call addSourceToList(eSourceList, tempSource)   
           
        end do
        if (associated(source) ) then   
            deallocate(source)
            nullify(source)
        end if
 

        if (n .eq. nesources) exit ! No sources merged

        elevelimage = templevelimage
 
        if(n .lt. 2) exit 
    end do


    if (mergedP) then

        write(errstr,'(a, i0)') "After merging, number of likely extended-sources=", nesources
        call message(AppMsg, Verbose)
    end if
!write(*, *) "Line 3708"

    deallocate(merged, tempLevelImage)
    nullify(merged, tempLevelImage)

!  write(*, *) "Line 3713"
!call checkLinkedList(nesources, eSourceList, "mergeextendedsources line 3714")


end subroutine mergeextendedSources




!*********************************************************
!
! If any 2 extended sources overlap, merge them
! into a single source
!
!*********************************************************
subroutine MergeSmokeRingSources(nesources, eSourceList, &
          image, backgroundimage, elevelImage, tolerance)

    ! Extended-source variables

    integer(int16), intent(inout) :: nesources
    type(SourceLinkedList), intent(inout) :: esourceList  

    real(single), pointer, dimension(:, :) :: image, backgroundimage 
 
    integer(int16), pointer :: elevelImage(:, :)
    real(single), intent(in) :: tolerance
 
    integer(int16), pointer :: templevelImage(:, :) => null()
    real(single) :: dx, dy, radius, error, error1, error2
    real(single) :: xPos, yPos, minRadius, maxAxialRatio, minDist   
    real(single) :: meanX, meanY, minX, maxX, minY, maxY, minDx, maxDx, maxSeparation
    real(single) :: xStart, yStart 
    integer(int16) :: nx, ny
    integer(int8), parameter :: maxNum = 10
    integer(int8) :: sourceIndex(maxNum), minNumPixels

    type(sourceParameters), pointer :: source(:) => null()
    type(sourceParameters) :: tempSource
 

    integer :: i, j, k, n, numToMerge, count
    logical, pointer :: merged(:) => null()
    logical :: mergedP

   
    if (neSources .lt. 3) return 

 

    minNumPixels = 200 / binx ! was 100 / binx
    mergedP = .false.  ! Not yet merged any sources

    minRadius = real(30 / binx)
    minDx = real(30 / binx)

    maxDx = real(100 / binx)

    minDist = real(60 / binx)
    maxSeparation = real(12 / binx)
    nx = ubound(elevelimage, 1)
    ny = ubound(elevelimage, 2)

    tempLevelImage => Allocate2DInt16(0_int16, nx, 0_int16, ny, "tempeLevelImage")
    allocate(merged(1 : nesources))
  


    where(elevelimage .gt. 0_int16) 
        templevelimage = elevelimage
    elsewhere
        tempLevelImage = 0_int16
    end where


    n = neSources
       
    neSources = 0


    !************************************************************   
    ! Load the extended-sources and then delete the linked-list
    !************************************************************
    source => SourceDataArray(eSourceList, n)
    call DeleteSourceList(esourcelist)  ! Delete source-list and start a new one
    merged = .false.



    !**********************************************************************************
    ! Loop through each source and check to see if it has a close neighbour
    ! that needs merging with it
    !**********************************************************************************

!write(*, *) "Merge smoke rings"     
    do i = 1, n

! if(i .ne. 30) cycle
!write(*, *) "Checking source ", i, " ) x=", source(i)%xpos, " y=", source(i)%ypos                   
        if(merged(i)) cycle

        neSources = neSources + 1 ! The new current source number


        !************************************************************
        ! 
        !***********************************************************
        where(elevelImage(source(i)%imin : source(i)%imax, &
            source(i)%jmin : source(i)%jmax) .eq. source(i)%number) 
            templevelImage(source(i)%imin : source(i)%imax, &
                           source(i)%jmin : source(i)%jmax) = -neSources
                                              
        end where
        if( source(i)%extensionFlag .eq. 0 .or. source(i)%noSourcePixels .gt. minNumPixels) then
            source(i)%number = -nESources
            call addSourceToList(eSourceList, source(i))   
            cycle
        end if
          
        tempSource = source(i)         
        error1 = source(i)%poserr

        numToMerge = 0  ! The number of sources to merge (not including source i)
        mergedP = .false.
        xStart = source(i)%xpos
        yStart = source(i)%ypos
        count = 0
        xPos = xStart
        yPos = yStart
        meanX = xPos
        meanY = yPos
        minX = xPos
        maxX = xPos
        minY = yPos
        maxY = yPos

        do j = 1, n
            if (i .eq. j) cycle
            if ( merged(j)) cycle
 
            if( source(j)%extensionFlag .eq. 0) cycle
            if( source(j)%noSourcePixels .gt. minNumPixels) cycle
          
        
            error2 = source(j)%poserr
               
            dx = xPos - source(j)%xpos
            dy = yPos - source(j)%ypos
            radius = sqrt(dx * dx + dy * dy)
            if(radius .gt. tolerance) cycle

            error = 2.0 * sqrt(error1 * error1 + error2 * error2)
            error = max(error, tolerance)

     
            if (error .gt. radius .or. radius .le. tolerance) then ! Merge sources
                numToMerge = numToMerge + 1
                sourceIndex(numToMerge) = j
                meanX = meanX + source(j)%xpos
                meanY = meanY + source(j)%ypos
                minX = min(minX, source(j)%xpos)
                maxX = max(maxX, source(j)%xpos)
                minY = min(minY, source(j)%ypos)
                maxY = max(maxY, source(j)%ypos)

       
                dx = xStart - source(j)%xpos
                dy = yStart- source(j)%ypos
                radius = sqrt(dx * dx + dy * dy)
    

                if( radius .le. minRadius) count = count + 1

                if(numToMerge .eq. maxNum) exit         
                xPos = source(j)%xpos
                yPos = source(j)%ypos
 

            end if
        end do

        !write(*, *) "Line 3925 numtomerge=", numToMerge, " count=", count
        if(numToMerge .lt. 3 .and. count .ne. 2) then

          
            !*****************************************
            ! Add source to list
            !*****************************************  

            source(i)%number = -nESources
            call addSourceToList(eSourceList, source(i))   
            cycle
        end if   
!write(*, *) "minx=", minx, " max=", maxx
!write(*, *) "miny=", miny, " max=", maxy


        if( ( maxX - minX) .lt. minDx .and. (maxY - minY) .lt. minDx .or. &
            (maxX - minX) .gt. maxDx .or. (maxY - minY) .gt. maxDx) then

 
            !*****************************************
            ! Add source to list
            !*****************************************  

            source(i)%number = -nESources
            call addSourceToList(eSourceList, source(i))   
            cycle
        end if   

        meanX = meanX / real(numToMerge + 1)
        meanY = meanY / real(numToMerge + 1)
 
!write(*, *) "meanx=", meanx, " meany=", meany
        !***********************************************
        ! Merge the sources
        !***********************************************

!write(*, *) i, ") x=", xpos, " y=", ypos, " numtomerge=", numtomerge, " count=", count

        do k = 1, numToMerge
            j = sourceIndex(k)
          

                
            where(elevelImage(source(j)%imin : source(j)%imax, source(j)%jmin : source(j)%jmax) .eq. source(j)%number) 
                      
                templevelImage(source(j)%imin : source(j)%imax, source(j)%jmin : source(j)%jmax) = -neSources
                         
            end where
                         
            source(i)%number = -neSources
            source(i)%imin = min(source(i)%imin, source(j)%imin)
            source(i)%imax = max(source(i)%imax, source(j)%imax)
            source(i)%jmin = min(source(i)%jmin, source(j)%jmin)
            source(i)%jmax = max(source(i)%jmax, source(j)%jmax)
            source(i)%noSourcePixels = source(i)%noSourcePixels + source(j)%noSourcePixels 
              
           
    
        end do
      
        !*****************************************************************
        ! Compute extended-source parameters
        !****************************************************************
        call ComputeExtendedSrcParameters2(int(source(i)%number, int16), &
                      nx, ny, image, backgroundImage, &
                      templevelImage, source(i), .true.)  ! True for Monte-carlo errors
!write(*, *) "a=", source(i)%semia, " b=", source(i)%semib
        if(   ( real(nx) + 0.5 - source(i)%xpos) .lt. minDist .or. &
                  ( real(ny) + 0.5 - source(i)%ypos) .lt. minDist .or. &
                 source(i)%xpos .lt. minDist .or. source(i)%ypos .lt. minDist) then
                maxAxialRatio = 2.0
        else
            maxAxialRatio = 2.0
        end if
   
!write(*, *) "Line 4003"
        if(source(i)%semia / source(i)%semib .le. maxAxialRatio .and. &
           abs(source(i)%xpos - meanX) .le. maxSeparation .and. &
           abs(source(i)%ypos - meanY) .le. maxSeparation .or. &
              isNullValue(source(i)%semia)) then
            do k = 1, numToMerge
                j = sourceIndex(k)
                merged(i) = .true. ! Remember source has been merged

                merged(j) = .true. ! ditto
            end do       
            source(i)%qualityFlag = ibset(source(i)%qualityFlag, 2) 
        else
            source(i) = tempSource
        end if
        
        !*****************************************
        ! Add source to list
        !*****************************************  

        source(i)%number = -nESources

        !******************************************************************************
        ! Record that a possible smoke-ring has been found attached to this star
        ! But remember to unset this later on
        !******************************************************************************
      
       
        call addSourceToList(eSourceList, source(i))   
      

    end do


    if (associated(source) ) then   
        deallocate(source)
        nullify(source)
    end if

    elevelimage = templevelimage
 
    if ( n .lt. nesources) then

        write(errstr,'(a, i0)') "After merging smoke-ring sources, number of likely extended-sources=", nesources
        call message(AppMsg, Verbose)
    end if

    deallocate(merged, tempLevelImage)
    nullify(merged, tempLevelImage)


end subroutine mergeSmokeRingSources

!*********************************************************
!
! If any 2 sources lie within a radius tolerance then
! merge into a single source
!*********************************************************
subroutine MergePointSources(npsources, pSourceList, image, &
          backgroundimage, plevelImage, tolerance)

 

    integer(int16), intent(inout) :: npsources
    type(SourceLinkedList), intent(inout) :: psourceList  

    real(single), pointer, dimension(:, :) :: image, backgroundimage 
 
    integer(int16), pointer :: plevelImage(:, :)
    real(single), intent(in) :: tolerance
   
    integer(int16), pointer :: templevelImage(:, :) => null()
    real(single) :: dx, dy, radius, error, error1, error2
   
    integer(int16) :: nx, ny

    type(sourceParameters), pointer :: source(:) => null()
    

    integer :: i, j, k, n
    logical, pointer :: merged(:) => null()
    logical :: mergedP    


    if (npSources .lt. 2) return
   
    mergedP = .false.
    nx = ubound(plevelimage, 1)
    ny = ubound(plevelimage, 2)




    tempLevelImage => Allocate2DInt16(0_int16, nx, 0_int16, ny, "tempeLevelImage")
    allocate(merged(1 : npsources))
  


    do k = 1, 3 ! Have 3 iterations at the most
  

        tempLevelImage = 0_int16  ! The new source map
        
        n = npSources             ! The number of sources before merging
       
   
        ! Load the extended-sources
        source => SourceDataArray(pSourceList, n)
        call DeleteSourceList(psourcelist)
        merged = .false.          ! Not merged anything yet

       
        nPSources = 0
        !****************************************************************
        ! Loop through all the point-sources to see if any
        ! has a very close neighbour
        !*****************************************************************
        do i = 1, n

            if(merged(i)) cycle           
            npSources = npSources + 1  ! The new source number

            !*****************************************************************
            ! Transfer the source pixels for source i to the new source-map
            !*****************************************************************           

            where(plevelImage(source(i)%imin : source(i)%imax,      &
                source(i)%jmin : source(i)%jmax) .eq. source(i)%number) 
                templevelImage(source(i)%imin : source(i)%imax, &
                source(i)%jmin : source(i)%jmax) = npSources
                                              
            end where
            source(i)%number = npsources 

            error1 = source(i)%poserr
         
            !*******************************************************************
            ! See if this source has any close neighbours
            !
            !******************************************************************
            do j = i + 1, n
                if (merged(j)) cycle  ! Skip if merged
                error2 = source(j)%poserr
            
                dx = source(i)%xpos - source(j)%xpos
                dy = source(i)%ypos - source(j)%ypos
                radius = sqrt(dx * dx + dy * dy)
                error = 2.0 * sqrt(error1 * error1 + error2 * error2)
                error = max(error, tolerance)

                if (error .gt. radius) then ! Merge sources

                    !******************************************************************
                    ! Merging two sources
                    !******************************************************************

                    merged(i) = .true.
                    merged(j) = .true.
                    write(errstr,'(a, i0, a, i0)') "Merging point sources ", i, " and ", j
!write(*, *) "i=", i, " x=", source(i)%xpos, " y=", source(i)%ypos
!write(*, *) "j=", j, " x=", source(j)%xpos, " y=", source(j)%ypos


                    call message(AppMsg, Verbose)
                    
                 
                    !*****************************************************************
                    ! Transfer the source pixels for source j to the new source-map
                    !*****************************************************************               
                    where(plevelImage(source(j)%imin : source(j)%imax, &
                                     source(j)%jmin : source(j)%jmax) .eq. source(j)%number) 
                      
                      templevelImage(source(j)%imin : source(j)%imax, &
                                     source(j)%jmin : source(j)%jmax) = npSources
                         
                    end where

                    !*****************************************************************
                    ! The source parameters for the merged source
                    !*****************************************************************
                    source(i)%xpos = (source(i)%xpos + source(j)%xpos) / 2.0
                    source(i)%ypos = (source(i)%ypos + source(j)%ypos) / 2.0


!write(*, *) "Line 4347 merged source x=", source(i)%xpos, " y=", source(i)%ypos


                    source(i)%imin = min(source(i)%imin, source(j)%imin)
                    source(i)%imax = max(source(i)%imax, source(j)%imax)
                    source(i)%jmin = min(source(i)%jmin, source(j)%jmin)
                    source(i)%jmax = max(source(i)%jmax, source(j)%jmax)
                    source(i)%noSourcePixels = count( templevelImage(source(i)%imin : source(i)%imax, &
                              source(i)%jmin : source(i)%jmax) == npSources) 
 

                    call ComputeExtendedSrcParameters2(int(source(i)%number, int16), nx, ny, image, backgroundImage, &
                            templevelImage, source(i), .true.)  ! True for Monte-carlo errors
                

!write(*, *) "Line 4366 merged source x=", source(i)%xpos, " y=", source(i)%ypos


                    !**********************************************************************
                    ! Classify the source - point (0) or extended (1)
                    !*********************************************************************
                    call ClassifySource(source(i), image, backgroundImage)

                    if (source(i)%extensionFlag .eq. 0_int16) then                   
                        source(i)%noSourcePixels = 0_int16

                    end if

                    mergedP = .true.

 

!write(*, *) "merged source has ", source(i)%noSourcePixels, " pixels number=", source(i)%number
!call saveImage(plevelimage, nx+1_int16, ny+1_int16, "plevelimage.fits", "level image")
!call saveImage(templevelimage, nx+1_int16, ny+1_int16, "templevelimage.fits", "level image")
!call fatal("stopped line 3931 displaying plevelimage.fits and templevelimage.fits")


                   exit ! Move on to next source in the list
                
                end if        
            end do 
              
            !***********************************************
            !
            ! Add source to list
            !
            !***********************************************
               
            call addSourceToList(pSourceList, source(i))   
           

        end do
   
        plevelimage = templevelimage
        if(associated(source)) then   
            deallocate(source)
            nullify(source)
        end if

        if (n .eq. npsources) exit
    end do

    if (mergedP) then

        write(errstr,'(a, i0)') "After merging, new number of likely point-sources=", npsources
        call message(AppMsg, Verbose)
    end if

    deallocate(merged, tempLevelImage)
    nullify(merged, tempLevelImage)
  
end subroutine mergePointSources

!***********************************************************
!
!**********************************************************
subroutine DilateELevelImage( neSources, image, sourceList )
    integer(int16), intent(inout) :: neSources
    integer(int16), pointer  :: image(:, :)

    type(SourceLinkedList) :: sourceList 
    type(sourceParameters), pointer :: sourceData(:) =>  null()

    integer(int16) :: dx

  
    integer(int16) :: nx, ny, i, j, k, l, n
    integer(int16) :: il, iu, jl, ju
    
    integer(int16) :: il1, iu1, jl1, ju1

    if (neSources == 0) return

    ! Load the data
    sourceData => SourceDataArray( sourceList, neSources )
 
    nx = ubound(image, 1)
    ny = ubound(image, 2)
    dx = 8 / binx


    do n = 1, neSources ! Loop through all the extended sources

    
        il1 = max(0_int16, sourceData(n)%imin - dx)
        iu1 = min(nx, sourceData(n)%imax + dx)


        jl1 = max(0_int16, sourceData(n)%jmin - dx)
        ju1 = min(ny, sourceData(n)%jmax + dx)

   
        do j = jl1, ju1
            jl = max(0_int16, j - dx)
            ju = min(ny, j + dx)

            do i = il1, iu1
       
                if (image(i, j) .ne. 0) cycle 

                il = max(0_int16, i - dx)
                iu = min(nx, i + dx)

                kloop : do k = jl, ju

                    do l = il, iu

                        if (image(l, k) .lt.  0_int16) then


                            image(i, j) = 1_int16
                            emaskImage(i, j) = .true.
                            exit kLoop
                        end if
                    end do 
                end do kloop
      
            end do
        end do
    end do

    deallocate(sourceData)




end subroutine DilateELevelImage

!**********************************************************************
!
!
!*********************************************************************


function AnnulusMedian(x0, y0, innerRadius, outerRadius, imagePtr) result(median)
    real(single), intent(in) :: x0, y0, innerRadius, outerRadius
    real(single), pointer :: imagePtr(:, :)
    real(single) :: median, x, y, radius, dx, dy
    integer :: i, j, nx, ny, il, iu, jl, ju, noPixels

    real(single), pointer :: counts(:) => null()

    nx = ubound(imagePtr, 1)
    ny = ubound(imagePtr, 2)

    x = x0 - outerRadius
    
    il = floor(x) - 1
    il = max(0, il)

    x = x0 + outerRadius
    
    iu = min(nx, ceiling(x) + 1)


    y = y0 - outerRadius
    
    jl = floor(y) - 1
    jl = max(0, jl)

    y = y0 + outerRadius
    
    ju = min(ny, ceiling(y) + 1)

 
    counts => allocateSingle(1, 1000, "counts")
 
    noPixels = 0   
iLoop : do i = il, iu
        x = real(i)
        dx = x - x0
!write(*, *) "i=", i, " x0=", x0, " dx=", dx
        do j = jl, ju

            if( isNullValue(imagePtr(i, j)) ) cycle
            y = real(j)
            dy = y - y0
!write(*, *) "j=", j, " y0=", y0, " dy=", dy

            radius = sqrt(dx * dx + dy * dy)
            if (radius .ge. innerRadius .and. radius .le. outerRadius) then 

!          write(*, *) "radius=", radius, " inner=", innerradius, " outer=", outerradius
              
                noPixels = noPixels + 1 
                counts(noPixels) = imagePtr(i, j)  
                if (noPixels == 1000) exit iLoop
            end if
                    
            
        end do
    end do iLoop 


    if (noPixels > 3) then
!        median = MedianValue1(counts, noPixels)
         median = sum(dble(counts)) / dble(noPixels)

    else
        median = -1.0
    end if
 

    deallocate(counts)


end function AnnulusMedian



!*******************************************************
!
!
! 
!*******************************************************
logical function isSmokeRing( sourceNumber, image, &
        plevelImage, elevelImage,                  &
        qualityImage,                              &
        il, iu, jl, ju,                            &
        eSourceList, neSources)
    integer(int32), intent(in) :: sourceNumber
    real(single), pointer :: image(:, :)
    integer(int16), pointer, dimension(:, :) :: plevelImage, elevelImage
    logical(boolean), pointer, dimension(:, :) :: qualityImage


    integer(int16), intent(in) :: il, iu, jl, ju
    type(SourceLinkedList), intent(inout) :: eSourceList
    integer(int16), intent(inout) :: neSources
    integer(int16) :: nx, ny, neSources1
    integer(int8) :: iter
    integer :: noPixels, n1, noSourcePixels 
    type(SourceLinkedList) :: tempeSourceList
    type(sourceParameters), pointer :: sourceData(:) =>  null()
  
  
    real(single) :: background, sigma, height, maxHeight
    real(single), pointer :: tempImage(:, :) => null()
    real(single), pointer :: backgroundImage(:, :) => null()
    
    integer(int16), pointer, dimension(:, :) :: tempelevelImage => null(), tempplevelImage => null()
    integer(int16) :: j, k, il1, iu1, jl1, ju1 
    integer(int8) :: dx1
    real(single) :: dx, dy, radius, radius3



    nx = iu - il 
    ny = ju - jl 
  
    isSmokeRing = .false.
      
    noPixels = (iu - il + 1) * (ju - jl + 1)

    n1 = count(plevelImage(il : iu, jl : ju) > 0 .and. plevelImage(il : iu, jl : ju) .ne. sourceNumber) 


    if (  (real( n1)  / real(noPixels) ) > 0.05) then
        return
    end if


    backgroundImage => allocate2dsingle(0_int16, nx, 0_int16, ny, "backgroundimage")
  

    tempImage => allocate2dsingle(0_int16, nx, 0_int16, ny, "tempimage")
    tempElevelImage => allocate2dInt16(0_int16, nx, 0_int16, ny, "tempELevelimage")
    tempPlevelImage => allocate2dInt16(0_int16, nx, 0_int16, ny, "tempPlevelimage")


    tempImage = image(il : iu, jl : ju)
    where(plevelImage(il : iu, jl : ju) .ne. 0)  tempElevelImage = 1


    tempPlevelImage = 0


    !*************************************************************
    !
    ! Smooth the image slightly
    !
    !************************************************************
    call lpdfilter( tempImage, qualityImage )


    background =  minval( tempImage, tempImage >= 0.0)

    backgroundImage = background
    sigma = max(1.0, sqrt(background))   



!call SaveImage(tempimage, nx+1_int16, ny+1_int16, "tempimage.fits", "Background image")
!call fatal("stopped at line 3650 displaying tempimage.fits")

    
    nx = nx + 1
    ny = ny + 1

!write(*, *) "il=", il, " iu=", iu, " jl=", jl, " ju=", ju, " background=", background


    do iter = 1, 20
    

        height = max(2.0, 0.5 * real(iter) * sigma)
              

        maxHeight = background + height + 20.0 * sigma


        noSourcePixels = count(tempElevelImage == 0 .and. &
                 tempimage >= (background + height) .and. &
                 tempimage <= maxheight)
           
        if ( noSourcePixels < 200 / binx .or. noSourcePixels > int(real(noPixels) * 0.5) ) then
            cycle
        end if

    
        nESources1 = 0

        where(tempElevelimage < 0) tempElevelimage = 0


        call DetectExtendedSources(nx, ny,                                         &
                   tempImage, tempImage, backgroundImage,                          &
                   maxHeight, tempplevelImage, tempeLevelImage,                    &   !
                   tempeSourceList, 1_int16, neSources1, 200 / binx, noPixels,     &      
                   1_int8,  height, heightFunction2, IsValidSmokeRing,             &
                   1_int8,                                                         &   ! The extension flag 
                   .false.)            

        where(tempElevelimage > 1) tempElevelimage = 0

        if ( neSources1 .eq. 1) then
                
            isSmokeRing = .true. ! Found a smoke-ring


            ! Load the data
            sourceData => SourceDataArray( tempEsourceList, neSources1 )

            nesources = nesources + 1

            nx = nx - 1
            ny = ny - 1 

            dx1 = 10 / binx
            il1 = max( 0_int16, sourceData(1)%imin - dx1)
            iu1 = min( nx, sourceData(1)%imax + dx1)
            jl1 = max( 0_int16, sourceData(1)%jmin - dx1)
            ju1 = min( ny, sourceData(1)%jmax + dx1)

            radius3 = real(36 / binx)

            do j = jl1, ju1
                dy =  abs(sourceData(1)%ypos - 1.0 - real(j))
                if (dy > radius3) cycle
                do k = il1, iu1
                    dx = abs(sourceData(1)%xpos -1.0 - real(k))
                    if (dx > radius3) cycle 
                    radius = sqrt(dx * dx + dy * dy)
                    if (radius .le. radius3 .and. tempelevelImage(k, j) == 0) then 
                    tempelevelimage(k, j) = 1
          
                
                    end if            
                end do

            end do 
 

            sourceData(1)%xpos = sourceData(1)%xpos + real(il)
            sourceData(1)%ypos = sourceData(1)%ypos + real(jl)


            sourceData(1)%imin = sourceData(1)%imin + il
            sourceData(1)%imax = sourceData(1)%imax + il
            sourceData(1)%jmin = sourceData(1)%jmin + jl
            sourceData(1)%jmax = sourceData(1)%jmax + jl
            sourceData(1)%number = -nesources
            
            !**********************************************    
            !
            ! Add to the extended-source list 
            !
            !*********************************************
            call addSourceToList(esourceList, sourceData(1))
            
            where( tempElevelImage < 0 .and. elevelImage(il : iu, jl : ju) .eq. 0) 
                elevelImage(il : iu, jl : ju) = -nesources
                emaskImage(il : iu, jl : ju) = .true.
                pmaskImage(il : iu, jl : ju) = .true.

            end where
   

            where( tempElevelImage == 1 .and. elevelImage(il : iu, jl : ju) .eq. 0) 
                elevelImage(il : iu, jl : ju) = 1

                emaskImage(il : iu, jl : ju) = .true.
                pmaskImage(il : iu, jl : ju) = .true.
                
            end where
   
            deallocate(sourceData)

            deallocate(tempImage, backgroundimage, tempElevelImage, tempPlevelimage)

            return
      
        endif

    end do   
    

    deallocate(tempImage, backgroundimage, tempElevelImage, tempPlevelimage)
  
end function isSmokeRing

!*********************************************
!
!
!********************************************
Subroutine LookForReadOutStreaks1(source, image)
    type(sourceParameters), pointer :: source(:)

    real(single), pointer, dimension(:, :) :: image
    

    integer :: nMin, i0, i, n1, n2, n3, n4, n5, n6
    integer :: nx, ny, nStreaks
    real(single) :: width
  

    if (.not. associated(source)) return

    write(errstr,'(a)') "Checking for read-out streaks"
    call message(AppMsg, Verbose)

    nx = ubound(image, 1)
    ny = ubound(image, 2)

    nMin = int( 0.60 * (ny + 1)) 
    width = real(10 / binx)
    nStreaks = 0  
    i0 = 0
    do i = 0, nx

        if ( i0 < 6) then 

            
            n4 = count(image(i0, :) > image(i0 + 1, :) ) 
            n5 = count(image(i0, :) > image(i0 + 2, :) ) 
            n6 = count(image(i0, :) > image(i0 + 3, :) )
            n1 = count(image(i0, :) > image(i0 + 4, :) ) 
            n2 = count(image(i0, :) > image(i0 + 5, :) ) 
            n3 = count(image(i0, :) > image(i0 + 6, :) )
!write(*, *) "i0=", i0            
!write(*, *) "n1=", n1, " n2=", n2, " n3=", n3, " n4=", n4, " n5=", n5, " n6=", n6

        else if ( i0 > (nx - 6)) then


          
                      

            n4 = count(image(i0, :) > image(i0 - 4, :) ) 
            n5 = count(image(i0, :) > image(i0 - 5, :) ) 
            n6 = count(image(i0, :) > image(i0 - 6, :) )


            n3 = ny
            n4 = ny
            n5 = ny



!write(*, *) "i0=", i0            
!write(*, *) "n1=", n1, " n2=", n2, " n3=", n3, " n4=", n4, " n5=", n5, " n6=", n6

        else ! Not near an edge

            n1 = count(image(i0, :) > image(i0 - 2, :) ) 
            n2 = count(image(i0, :) > image(i0 - 3, :) ) 
            n3 = count(image(i0, :) > image(i0 - 4, :) )

!            i0 = i0 + 1
            n4 = count(image(i0, :) > image(i0 + 2, :) ) 
            n5 = count(image(i0, :) > image(i0 + 3, :) ) 
            n6 = count(image(i0, :) > image(i0 + 4, :) )

!write(*, *) "i0=", i0            
!write(*, *) "n1=", n1, " n2=", n2, " n3=", n3, " n4=", n4, " n5=", n5, " n6=", n6


        end if
        if (n1 > nMin .and. n2 > nMin .and. n3 > nMin .and. n4 > nMin &
                      .and. n5 > nMin .and. n6 > nMin) then




!write(*, *) "i0=", i0, " nmin=", nmin            
!write(*, *) "n1=", n1, " n2=", n2, " n3=", n3, " n4=", n4, " n5=", n5, " n6=", n6




            write(errstr,'(a, i0, a, f0.2, a, f0.2)') &
                      "Possible read-out streak centred on column ", i0
            call message(AppMsg, Verbose)

            where(abs(source%xpos - real(i0) + 1.0) .lt. width              &
                 .and. source%significance .lt. 50.0) source%qualityFlag =  &
                  ibset(source%qualityFlag, 1)  ! BIT 1 set for sources on read-out streaks

            nStreaks = nStreaks + 1
            i0 = i0 + 10
          
        else
            i0 = i0 + 1
        end if

        if ( (nx - i0) < 6) exit
  

    end do

    write(errstr,'(a, i0)') "Number of possible read-out streaks found=", nstreaks

    call message(AppMsg, Verbose)

    write(errstr,'(a, i0)') "Number of source flags set=", count(source%qualityFlag .eq. 2 .or. source%qualityFlag .eq. 3)

    call message(AppMsg, Verbose)

end Subroutine LookForReadOutStreaks1

!***************************************************************
!
!
!**************************************************************
subroutine setSourceRings(source, nSources)
    type(sourceParameters), pointer :: source(:)
    integer(int16), intent(in) :: nSources
    integer :: i, j
    real(single) :: dx, dy, radius
    real(single) :: width
    if (nSources == 0) return
     
    width = real(38 / binx)


    do i = 1, nSources
        if ( btest(source(i)%qualityFlag, 2) ) then ! Then has been detected as a smoke ring 

               

            do j = i - 1, 1, -1
                if ( btest(source(j)%qualityFlag, 4) ) cycle 
                dx = source(i)%xpos - source(j)%xpos
                dy = source(i)%ypos - source(j)%ypos
                radius = sqrt(dx * dx + dy * dy)

                if (radius .le. width) then

                    ! Remember that it is probably a smoke ring     
                    source(j)%qualityFlag = ibset(source(j)%qualityFlag, 2) 
  
                end if
 
            end do 
            do j = i + 1, 1
                if ( btest(source(j)%qualityFlag, 4) ) cycle 
                dx = source(i)%xpos - source(j)%xpos
                dy = source(i)%ypos - source(j)%ypos
                radius = sqrt(dx * dx + dy * dy)

                if (radius .le. width) then

                    ! Remember that it is probably a smoke ring     
                    source(j)%qualityFlag = ibset(source(j)%qualityFlag, 2) 
  
                end if
 
            end do 

        end if
    end do

end subroutine setSourceRings

!*************************************************************
!
!***********************************************************
!subroutine checklevelimage(plevelimage, plist, npsources, elevelimage, elist, nesources)
!    integer(int16), pointer :: plevelImage(:, :)
!    type(SourceLinkedList), intent(inout) :: plist 
!    integer(int16), intent(in) :: npsources

!    integer(int16), pointer :: elevelImage(:, :)
!    type(SourceLinkedList), intent(inout) :: elist 
!    integer(int16), intent(in) :: nesources


!    type(sourceParameters), pointer :: source(:) =>  null()  
!    integer(int16) :: nx, ny, i

   
 
    ! Load the data

!    source => SourceDataArray( plist, npSources )
   
!    do i = 1, npsources
!        write(*, *) i, ") x=", source(i)%xpos, " y=", source(i)%ypos, &
!          " number=", source(i)%number, " eflag=", source(i)%extensionFlag, &
!         " imin=", source(i)%imin, " imax=", source(i)%imax, &
!         " jmin=", source(i)%jmin, " jmax=", source(i)%jmax

!    end do

!    deallocate(source)


!    source => SourceDataArray( elist, neSources )
!    write(*, *) "Extended source list" 
!    do i = 1, nesources
!        write(*, *) i, ") x=", source(i)%xpos, " y=", source(i)%ypos, &
!          " number=", source(i)%number, " eflag=", source(i)%extensionFlag, &
!         " imin=", source(i)%imin, " imax=", source(i)%imax, &
!         " jmin=", source(i)%jmin, " jmax=", source(i)%jmax

!    end do

!    deallocate(source)




!    nx = size(plevelimage, 1)
!    ny = size(plevelimage, 2)
   
!call saveImage(plevelimage, nx, ny, "plevelimage.fits", "background image")
!call saveImage(elevelimage, nx, ny, "elevelimage.fits", "background image")

!call fatal("stopped line 5333 displaying plevelimage.fits and elevelimage.fits")

!end subroutine checklevelimage






!*************************************************************
!
! FOR TEST PURPOSES ONLY
!
!***********************************************************
!subroutine displaySourceList(list, nsources)
!    integer(int16), intent(inout) :: nsources
!    type(SourceLinkedList), intent(inout) :: list     
!    type(sourceParameters), pointer :: source(:) => null()
   



!    if(nsources .eq. 0) return

    !****************************************************
    ! Load the point sources from the linked list
    !****************************************************    
!    source => SourceDataArray( list, nSources)

 

   
!    call displaySources(nsources, source)


!    deallocate(source)
!    nullify(source)

!end subroutine displaySourceList
!*************************************************
!
!************************************************
!subroutine displaySources(nsources, source)
!    integer(int16), intent(in) :: nSources
!    type(sourceParameters), pointer :: source(:)
!    integer :: i
 

!   write(*, *) "Displaying source list"


!    do i = 1, nsources

!        if (isNullValue(source(i)%xpos) .or. source(i)%xpos < 0.0 .or. &
!            isNullValue(source(i)%ypos) .or. source(i)%ypos < 0.0) then


    
!        write(*, *) i, ") x=", source(i)%xpos, " y=", source(i)%ypos, &
!              " number=", source(i)%number, &
!              " imin=", source(i)%imin, " imax=", source(i)%imax, &
!              " jmin=", source(i)%jmin, " jmax=", source(i)%jmax, &
!              " extension flag=", source(i)%extensionFlag, " no pixels=", &
!              source(i)%noSourcePixels, " rawcountrate=", source(i)%rawCountrate

!        call fatal("stopped in omdetect at line 5230")
!        end if
!    end do

!end subroutine displaySources




!*********************************************************
!
!
! Carry out various checks on the point-sources 
!
!*********************************************************
!subroutine checkPointSources( nSources, source, image, levelimage)
subroutine checkPointSources( nSources, source, image, levelimage, llevelImage)

    
    integer(int16), intent(inout) :: nSources
    
    type(sourceParameters), pointer :: source(:)
 
    real(single), pointer, dimension(:, :) :: image


    integer(int16), pointer, dimension(:, :) :: levelimage
    integer(int16), pointer, dimension(:, :) :: llevelimage

    integer(int16) :: i, j, isrc, jsrc
    integer(int16) :: ilow, ihigh, jlow, jhigh
    logical :: onScattered
    logical(bool), pointer :: delete(:) => null()
    integer(int16), pointer, dimension(:, :) :: templevelimage => null()
    

    real(single) :: dx, dy, radius
    real(single) :: tolerance, f0, minRadius
    real(single) :: constant
    integer(int16) :: number, nx, ny


    if(nSources .lt. 2) return

    tolerance = min(40.0, real(60 / binx))
    minRadius = real(12 / binx)

    constant = real(30 / binx) / log(0.01)

  
    allocate( delete(1 : nSources))

!write(*, *) "Checking point sources"
     
    delete = .false.
    !write(*,*) "L5166 ---------"
    ilow = lbound(llevelimage, 1)
    jlow = lbound(llevelimage, 2)
    ihigh = ubound(llevelimage, 1)
    jhigh = ubound(llevelimage, 2)

    do i = 1, nSources

        if(delete(i))  cycle


        !********************************************************
        ! Check scattered light features for point sources
        !********************************************************
        !write(*,*) "L5180 i=", i, " xPos=", source(i)%xPos, &
        !     " ypos=", source(i)%yPos, " imin=",source(i)%imin, &
        !     " imax=", source(i)%imax, " jmin=", source(i)%jmin, &
        !     " jmax=", source(i)%jmax, " qflag=", source(i)%qualityFlag
       
       
       onScattered=.false.
       do isrc=source(i)%imin, source(i)%imax
          do jsrc=source(i)%jmin, source(i)%jmax
             if (isrc .ge. ilow .and. isrc .le. ihigh &
                  .and. jsrc .ge. jlow .and. jsrc .le. jhigh) then
                if (llevelImage(isrc, jsrc) .ne. 0) then
                   onScattered=.true.
                   cycle
                endif
             endif
          enddo
          if (onScattered) cycle
       enddo
       if (onScattered) then ! set the flag for the underlying extended source
          source(i)%qualityFlag = ibset(source(i)%qualityFlag, 8)
       endif
       !write(*,*) "    L5202 qFlag=", source(i)%qualityFlag


        !*********************************************************
        !
        !
        ! 
        !
        !*********************************************************

        if ( .not. btest(source(i)%qualityFlag, 4) .or. source(i)%rawCountRate .lt. 60.0) cycle 
       
        f0 = source(i)%rawCountRate

!        write(*, *) "*******************************************************************************" 
!        write(*, *) "Checking sources near source ", i, " x=", source(i)%xpos, " y=", source(i)%ypos, &
!                    " rawcountrate=", f0  
!        write(*, *) "*******************************************************************************" 
     
        do j = 1, nsources
            if (i .eq. j) cycle

            if(delete(j)) cycle
   
            if( source(j)%rawCountRate .gt. 20.0) cycle 
            dx = abs(source(i)%xpos - source(j)%xpos)
            if (dx > tolerance) cycle








            dy = abs(source(i)%ypos - source(j)%ypos)
            if (dy > tolerance) cycle
            radius = sqrt(dx * dx + dy * dy)
            if(radius > tolerance) cycle

            if ( .not. checkSourceCirclePixels( source(j), image, real(6 / binX) ) )then



              

          !write(*, *) "Deleting source ", j, " x=", &
          !          source(j)%xpos, " y=", source(j)%ypos

                delete(j) = .true.
            end if 

      
                      
        end do

    end do

    if( count( delete ) .eq. 0)  then

         !*********************************************************
         !
         ! Nothing to change
         !
         !*********************************************************
         deallocate( delete )
         nullify(delete)
         return
    end if 

    nx = ubound(levelimage, 1)
    ny = ubound(levelimage, 2)

    tempLevelImage => allocate2dInt16( 0_int16, nx, 0_int16, ny, "tempLevelImage" )

    number = 0

    do i = 1, nsources

        if ( .not. delete(i)) then
            number = number + 1
            call reNumberLevelImagePixels(source(i), levelImage, tempLevelImage, number)

            source(number) = source(i)

            source(number)%number = number

        end if

    end do

    deallocate(levelImage)
    levelImage => tempLevelImage


    deallocate(delete)
    nullify(delete)
  
    if (number .lt. nSources) then
        write(errstr,'(a, i0)') "Number of point sources deleted=", (nSources - number)
        call message(AppMsg, Verbose)
    end if

    nSources = number

!write(*, *) "After deleting"
!call displaySources(nsources, source)

end subroutine CheckPointSources


!*********************************************************
!
!
! Carry out various checks on the extended-sources 
!
!
!*********************************************************
!subroutine checkExtendedSources( nSources, psourceList, &
!        nESources, esource, levelimage)
subroutine checkExtendedSources( nSources, psourceList, &
        nESources, esource, levelimage, llevelImage)
    

    integer(int16), intent(inout) :: nSources, neSources
    type(SourceLinkedList), intent(inout) :: psourceList ! For extended sources 
 

    type(sourceParameters), pointer, dimension(:) :: source, eSource
 
  

    integer(int16), pointer, dimension(:, :) :: levelimage
    integer(int16), pointer, dimension(:, :) :: llevelimage

    integer(int16) :: i, j, isrc, jsrc
    integer(int16) :: ilow, ihigh, jlow, jhigh
    logical :: onScattered

    logical, pointer :: delete(:) => null()
    integer(int16), pointer, dimension(:, :) :: templevelimage => null()
    

    real(single) :: dx, dy, radius
    real(single) :: tolerance, tolerance1, f0, f1, f2, ratio
    integer(int16) :: number, nx, ny
  

    !write(*,*) "L5308 ---------"
    if(nSources .eq. 0 .or. nEsources .eq. 0) return

 
    source => SourceDataArray(pSourceList, nSources)
    
    tolerance = min(40.0, real(60 / binx))
    tolerance1 = real(12 / binx)    
    
    allocate( delete(1 : nESources))
   
    delete = .false.

    ilow = lbound(llevelimage, 1)
    jlow = lbound(llevelimage, 2)
    ihigh = ubound(llevelimage, 1)
    jhigh = ubound(llevelimage, 2)

    do i = 1, nSources  ! Loop through all the point-sources
        

        !*********************************************************
        ! 
        !*********************************************************

             
        f0 = source(i)%rawCountRate


        !write(*, *) "L5333 Checking source ", i, " x=", source(i)%xpos, " y=", source(i)%ypos, &
        !            " sedmia=", source(i)%semia, " b=", source(i)%semib, " theta=", source(i)%theta, &
        !            " rawcountrate=", f1  

        do j = 1, nesources
                       

            if(delete(j)) cycle

            dx = source(i)%xpos - esource(j)%xpos
            if (abs(dx) > tolerance) cycle

            dy = source(i)%ypos - esource(j)%ypos
            if (abs(dy) > tolerance) cycle
            radius = sqrt(dx * dx + dy * dy)

            if(radius .le. tolerance .and. btest(source(i)%qualityFlag, 4) ) then

                f1 = f0 * (tolerance1 / radius) ** 2
                f2 = esource(j)%rawCountRate
             

 !               write(*, *) "Comparing source ", j, " x=", &
 !                   esource(j)%xpos, " y=", esource(j)%ypos, " radius=", radius, &
 !                   " rawcountrate=", f2, " ratio=", (f2 / f1)  

              
                if ( (f2 / f1) .lt. 0.1) then



              
!                write(*, *) "Line 5231 Deleting source ", j, " number=", esource(j)%number, " x=", &
!                    esource(j)%xpos, " y=", esource(j)%ypos, " radius=", radius, &
!                    " rawcountrate=", f2, " ratio=", (f2 / f1)  

       
                    delete(j) = .true.
          
                   
  
                end if
            else if(radius .le. tolerance1) then !  (12/binx)
                f2 = esource(j)%rawCountRate
                ratio = f2 / f0  
!write(*, *) "Line 5241 Checking extended-source ", j, " x=", esource(j)%xpos, &
!                " y=", esource(j)%ypos, " radius=", radius, &
!                    " rawcountrate=", f2, " ratio=", (f2 / f0)  

              
                if(ratio .lt. 0.1) then                     
                    delete(j) = .true.
                else if(ratio .le. 0.25) then
                   
                    esource(j)%extensionFlag = 0
                end if
                

            end if     
                      
        end do
        !********************************************************
        ! Check scattered light features for point sources
        !********************************************************
       !write(*,*) "L5405 i=", i, " xPos=", source(i)%xPos, &
       !     " ypos=", source(i)%yPos, " imin=",source(i)%imin, &
       !     " imax=", source(i)%imax, " jmin=", source(i)%jmin, &
       !     " jmax=", source(i)%jmax, " qflag=", source(i)%qualityFlag
       
       
       !onScattered=.false.
       !do isrc=source(i)%imin, source(i)%imax
       !   do jsrc=source(i)%jmin, source(i)%jmax
       !      if (isrc .ge. ilow .and. isrc .le. ihigh &
       !           .and. jsrc .ge. jlow .and. jsrc .le. jhigh) then
       !         if (llevelImage(isrc, jsrc) .ne. 0) then
       !            onScattered=.true.
       !            cycle
       !         endif
       !      endif
       !   enddo
       !   if (onScattered) cycle
       !enddo
       !if (onscattered) then ! set the flag for the underlying extended source
       !   source(i)%qualityFlag = ibset(source(i)%qualityFlag, 8)
       !endif
       !write(*,*) "    L5427 qFlag=", source(i)%qualityFlag


    end do ! Finished looping through the point-sources
    !write(*,*) "L5431 ........................."

    !********************************************************
    ! Check scattered light features extended sources
    !********************************************************

    do j = 1, nesources
                       

       if(delete(j)) cycle
       
       !dx = source(i)%xpos - esource(j)%xpos
       !if (abs(dx) > tolerance) cycle
       !
       !dy = source(i)%ypos - esource(j)%ypos
       !if (abs(dy) > tolerance) cycle
       !radius = sqrt(dx * dx + dy * dy)
       
       ! checking if it is within a scattered light feature
       !write(*,*) "L5450 j=", j, " xPos=", esource(j)%xPos, &
       !     " ypos=", esource(j)%yPos, " imin=",esource(j)%imin, &
       !     " imax=", esource(j)%imax, " jmin=", esource(j)%jmin, &
       !     " jmax=", esource(j)%jmax, " qflag=", esource(j)%qualityFlag
       
       
       onScattered=.false.
       do isrc=esource(j)%imin, esource(j)%imax
          do jsrc=esource(j)%jmin, esource(j)%jmax
             if (isrc .ge. ilow .and. isrc .le. ihigh &
                  .and. jsrc .ge. jlow .and. jsrc .le. jhigh) then
                if (llevelImage(isrc, jsrc) .ne. 0) then
                   onScattered=.true.
                   cycle
                endif
             endif
          enddo
          if (onScattered) cycle
       enddo
       if (onscattered) then ! set the flag for the underlying extended source
          esource(j)%qualityFlag = ibset(esource(j)%qualityFlag, 8)
       endif
       !write(*,*) "    L5472 qFlag=", esource(j)%qualityFlag
    end do
    !****************************

    if( count( delete ) .eq. 0)  then
       !write(*,*) "L5398"
         !*********************************************************
         ! Nothing to change
         !*********************************************************
         deallocate( source, delete )
         return
    end if 

    nx = ubound(levelimage, 1)
    ny = ubound(levelimage, 2)
 

    !write(*,*) "L5410 nx=", nx, " ny=", ny

    tempLevelImage => allocate2dInt16( 0_int16, nx, 0_int16, ny, "tempLevelImage" )
    tempLevelImage = 0
    number = 0
    do i = 1, nesources

       write(*,*) "L5417 i=", i, " delete(i)=", delete(i), " ....."

        if ( .not. delete(i)) then  ! Keeping this source
            number = number + 1
 
             
            call reNumberLevelImagePixels(esource(i), levelImage, tempLevelImage, -number)
            
            esource(number) = esource(i)

            esource(number)%number = -number
      
            
            
        end if
        

    end do

    if (number .lt. neSources) then
        write(errstr,'(a, i0)') "Number of extended sources deleted=", (neSources - number)
        call message(AppMsg, Verbose)
    end if

    neSources = number
    deallocate(levelImage)
    levelImage => tempLevelImage

    deallocate(delete, source)



end subroutine CheckExtendedSources


!*********************************************************
!
!
! Checks each point-source for a close extended-source
! and if one exists removes it if its total counts
! are less than a fraction of the extended-source counts
!
!
!*********************************************************
subroutine filterPointSources(image,       &
                              npsources,   &
                              neSources,   &
                              pSourceList, &
                              eSourceList, &
                              pLevelImage, &
                              eLevelImage, &
                              minRadius,   &
                              radiusRatio, ratio)
                         

    integer(int16), intent(inout) :: npsources, neSources
    type(SourceLinkedList), intent(inout) :: psourceList, eSourceList
  
    real(single), pointer, dimension(:, :) :: image 
 
    integer(int16), pointer, dimension(:, :) :: plevelImage, elevelImage

    real(single), intent(in) :: minRadius, radiusRatio
    real(single), intent(in) :: ratio

    integer(int16), pointer :: templevelImage(:, :) => null()


    type(sourceParameters), pointer, dimension(:) :: source=>null()
    
    type(sourceParameters), pointer, dimension(:) :: esource=>null()
  

    integer :: i, j, nNew
    integer(int16) :: nx, ny, number
    integer(int16) :: iMin, iMax, jMin, jMax
    real(single) :: radius, distance
    real(double) :: eCounts, pCounts, eMax, pMax

 
    if (npSources .eq. 0 .or. neSources .eq. 0) return

    write(errstr,'(a)') "Checking point sources for close extended-sources"
    call message(AppMsg, Verbose)


    nx = ubound(pLevelImage, 1)
    ny = ubound(pLevelImage, 2)

    tempLevelImage => Allocate2DInt16(0_int16, nx, 0_int16, ny, "tempeLevelImage")
   
  
    tempLevelImage = 0_int16  ! The new source map
        
    nNew = 0
   
    ! Load the extended-sources
    source => SourceDataArray(pSourceList, npSources)
    esource => SourceDataArray(eSourceList, neSources)

    call DeleteSourceList(psourcelist)
    
 
    !****************************************************************
    ! Loop through all the point-sources to see if any
    ! has a very close neighbour
    !*****************************************************************
    nNew = 0
   



 iLoop : do i = 1, npSources

        do j = 1, nESources
                      
            if (isNullValue( eSource(j)%semia) ) cycle 

!            write(*, *) i, ") checking extended source x=", esource(j)%xpos, " y=", esource(j)%ypos


            !*********************************************************************
            ! Compute the distnace between esource(j) and source(i) in terms
            ! of the esource(j)%semia, esource(j)%semib and esource(j)%theta
            !
            !*********************************************************************
            distance = sourceSeparation2(esource(j), source(i)%xPos, source(i)%yPos, radius)


!            write(*, *) "radius=", radius, " radius ratio=", distance

            if(distance .gt. radiusRatio .and. radius .gt. minRadius) cycle
 
          
            iMin = source(i)%imin
            iMax = source(i)%imax
            jMin = source(i)%jmin
            jMax = source(i)%jmax

            number = source(i)%number
    
            pCounts = sum( image(iMin : iMax, jMin : jMax), &
                      plevelImage(iMin : iMax, jMin : jMax) .eq. number ) 

            pMax = maxval(image(iMin : iMax, jMin : jMax), &         
                          plevelImage(iMin : iMax, jmin : jMax) .eq. number ) 

            iMin = eSource(j)%imin
            iMax = eSource(j)%imax
            jMin = eSource(j)%jmin
            jMax = eSource(j)%jmax

            number = eSource(j)%number
            eCounts = sum(image(iMin : iMax, jMin : jMax), &
                          elevelImage(iMin : iMax, jMin : jMax) .eq. number) ! The extended source counts

            eMax = maxval(image(iMin : iMax, jMin : jMax), &
                          elevelImage(iMin : iMax, jMin : jMax) .eq. number) ! The maximum extended source pixel counts


!  write(*, *) "Source ", i, " has an extended neighbour- ecounts=", ecounts, " pcounts=", pcounts, &
!           " pmax=", pmax, " emax=", emax, " ratio=", ratio

            if(pCounts .lt. ratio * eCounts .and. pMax .lt. eMax) then ! Remove this source

                write(errstr,'(a, i0, a, f0.2, a, f0.2)') "Deleting point-source ", i, " x=", &
                       source(i)%xpos, " y=", source(i)%ypos
                call message(AppMsg, Verbose)
 

                cycle iLoop

            end if  

        end do
              
        !***********************************************
        ! Add source to list
        !***********************************************
        nNew = nNew + 1
        source(i)%number = nnew       
        call addSourceToList(pSourceList, source(i))   
           
        where(plevelImage(source(i)%imin : source(i)%imax,      &
                source(i)%jmin : source(i)%jmax) .eq. source(i)%number) 

                templevelImage(source(i)%imin : source(i)%imax, &
                source(i)%jmin : source(i)%jmax) = nNew
                                              
        end where
    
    end do iLoop


    if (nNew .lt. nPsources) then

        write(errstr,'(a, i0)') "Number of point-sources deleted=", (nPsources - nNew)
        call message(AppMsg, Verbose)
    end if

    nPSources = nNew     
    plevelimage = templevelimage
   
    deallocate(source)
    nullify(source)

    
!*************************************
! 
!*************************************
!   write(*, *) "Test 1"

!   source1%xpos = 10.0
!   source1%ypos = 10.0
!   source1%semia = 5.0
!   source1%semib = 5.0
!   source1%theta = 90.0


!   source2%xpos = 20.0
!   source2%ypos = 10.0

!   source2%semia = 2.5
!   source2%semib = 2.5
!   source2%theta = 90.0

!      distance =  sourceSeparation(source1, source2, 1.0)



!   write(*, *) "source separation=", distance, " answer should be 2.5"


 
end subroutine filterPointSources
!***********************************************************
!
!
!**********************************************************
subroutine resetConfusionFlags(nSources, source)
    integer(int16), intent(in) :: nSources
    type(sourceParameters), pointer, dimension(:) :: source
  
    integer(int16) :: i, j
    real(single) :: dx, dy, radius
    real(single) :: r1, r2

    write(errstr,'(a)') "Checking source confusion flags"
    call message(AppMsg, Verbose)


!write(*, *) "Number of point source confusion flags set before=", &
!  count(source%confusionFlag > 0 .and. source%extensionFlag .eq. 0)
    source%confusionFlag = 0

    r1 = real(24 / binx)
    r2 = real(12 / binx)

    do i = 1, nSources
        if(source(i)%extensionFlag .eq. 1) cycle        
   
      
        do j = i + 1, nSources
            
            if(source(j)%extensionFlag .eq. 1) cycle        
   
            dx = source(i)%xpos - source(j)%xpos
            if(abs(dx) .gt. r1) cycle

            dy = source(i)%ypos - source(j)%ypos
            if(abs(dy) .gt. r1) cycle
            radius = sqrt(dx * dx + dy * dy)
            if(radius .le. r1) then

                if(radius .le. r2) then

                    
                    !**************************************************************
                    !
                    !**************************************************************
!                    source(i)%confusionFlag = 0
!                    source(j)%confusionFlag = 0

                    source(i)%confusionFlag = ibset(source(i)%confusionFlag, 1) 
                    source(j)%confusionFlag = ibset(source(j)%confusionFlag, 1) 
       
                else
                    !*************************************************************
                    !
                    !*************************************************************
                    if (.not. btest(source(i)%confusionFlag, 1 )) then
                        source(i)%confusionFlag = ibset(source(i)%confusionFlag, 0) 
                        source(j)%confusionFlag = ibset(source(j)%confusionFlag, 0) 
                    end if 

                end if 
            end if
        end do


    end do

!write(*, *) "Number of point source confusion flags set after=", &
!  count(source%confusionFlag > 0 .and.source%extensionFlag .eq. 0)

end subroutine resetConfusionFlags



!************************************************************************
! Check for highly elongated sources along read-out streaks
!************************************************************************
subroutine removeSpuriousStreakSources(nSources, source, levelImage)

    integer(int16), intent(inout) :: nSources
    type(sourceParameters), pointer, dimension(:) :: source
    integer(int16), pointer, dimension(:, :) :: levelimage

    integer(int16) :: i, number, nx, ny
    real(single) :: minWidth
    integer(int16), pointer, dimension(:, :) :: templevelimage => null()

    minWidth = real(4 / binx)

    nx = ubound(levelimage, 1)
    ny = ubound(levelimage, 2)


    tempLevelImage => allocate2dInt16( 0_int16, nx, 0_int16, ny, "tempLevelImage" )

    number = 0 ! The new number of sources
    do i = 1, nSources ! Loop through all the sources

!     if ( btest(source(i)%qualityFlag, 1) ) then
      
!write(*, *) i, ") x=", source(i)%xpos, " y=", source(i)%ypos, &
!   " a=", source(i)%semia, " b=", source(i)%semib, " theta=", source(i)%theta, &
!   " qflag=", source(i)%qualityFlag, " sig=", source(i)%significance
!end if
        !***********************************************************
        ! Check semi-major axis and quality flag
        !***********************************************************

        if( isNullValue(source(i)%semiA) .or.          &  ! Copy straight to new list
            .not. btest(source(i)%qualityFlag, 1) .or. &
            source(i)%significance .gt. 20.0)  then

            number = number + 1
        
      
            call reNumberLevelImagePixels(source(i), levelImage, tempLevelImage, int(source(i)%number, int16))

            source(number) = source(i)

          
        else  ! On a read-oust streak
            if ( abs(source(i)%theta - 90.0) .gt. 20.0 .or.    &
                   source(i)%semib .gt. minWidth         .or.   & 
                   ( source(i)%semia / source(i)%semib ) .lt. 2.0) then 

                number = number + 1
        
      
                call reNumberLevelImagePixels(source(i), levelImage, tempLevelImage, int(source(i)%number, int16))

                source(number) = source(i)

                
            end if 
        end if
    end do

    nSources = number
    deallocate(levelImage)
    nullify(levelImage)

    levelImage => tempLevelImage
end subroutine removeSpuriousStreakSources

!*****************************************************************
! For test purposes only
!****************************************************************
!subroutine checkLinkedList(nSources, sourceList, string)
!    integer(int16), intent(inout) :: nSources
!    type(SourceLinkedList), intent(inout) :: sourceList
!    character(*), intent(in) :: string
!    type(sourceParameters), pointer, dimension(:) :: source => null()

!    integer(int16) :: n 
!    n = nSources
!    source => SourceDataArray(sourceList, nSources)

!    if (n .ne. nSources) then
!        write(*, *) "n=", n, " nsources=", nsources 
!        call fatal("Something wrong with linked list", trim(string))
!    end if

!end subroutine checkLinkedList


!*******************************************
!
! subroutine GetExposureImage
! Gets the 2-d image from the fits file
!
!*******************************************
subroutine GetExposureImage(dataSet, image) 

    type(dataSetT) :: dataSet
 
    real(single), pointer, dimension(: , :) :: image
    integer(int16) :: nx, ny    
    integer(int8), dimension(:, :),  pointer :: int8ImageP     ! Integer8 array pointer.
    integer(int32), dimension(:, :), pointer :: int32ImageP    ! Integer32 array pointer.
    integer(int16), dimension(:, :), pointer :: int16ImageP    ! Integer16 array pointer.
    real(single), dimension(:, :),   pointer :: real32ImageP   ! Real32 array pointer. 
    real(double), dimension(:, :),   pointer :: real64ImageP   ! Real64 array pointer.
    integer :: dim(2)
    type(ArrayT) :: inArray
    integer :: nDim
 
    if( associated(image) ) then
        deallocate(image)
        nullify(image)
    end if


    inArray = array(dataSet, "EXPOSURE", read)

    nDim  = numberOfDimensions(inArray)             ! How many dimensions?
    ! If it's not an image then crash out
    if (nDim .ne. 2) call fatal('notImage','input file not an image')

    ! Get the number of bins.
    dim = dimensions(inArray)
    nx = dim(1)
    ny = dim(2)
    if (nx <= 0) call fatal("Bad image", "nx <=0")
    if (ny <= 0) call fatal("Bad image", "ny <=0")


 
    image => Allocate2DSingle(0,  nx - 1,  0 , ny - 1, "image")

    ! Convert the image to real, if necessary
    select case(arrayDataType(inArray))
        case(integer8)   
            int8ImageP => int8Array2Data(inArray)
            image = real(int8ImageP)
        case(integer16)
            int16ImageP => int16Array2Data(inArray)
            image = real(int16ImageP)    
        case(integer32) 
            int32ImageP => int32Array2Data(inArray)
            image = real(int32ImageP)
        case(real32)
         
            real32ImageP => real32Array2Data(inArray)
            image = real32ImageP       
        case(real64)
            real64ImageP => real64Array2Data(inArray)
            image = real(real64ImageP)
    end select

    call release(inArray)

end subroutine GetExposureImage

!*****************************************************************
!
!
!
!****************************************************************
subroutine setRADECColumns( dataSetH, nSource, source) 
    type(dataSetT), intent(in) :: dataSetH
  
    integer(int16), intent(in) :: nSource
    type(sourceParameters), pointer :: source(:)
    
    real(double) :: cdelt1, cdelt2, crpix1, crpix2, crval1, crval2
    real(double) :: dx, dy
    integer :: i
    type(DataSetT) :: dataSetH1
    type(tableT) :: tableH

    if (.not. associated(source) .or. nSource .eq. 0) return


    if( .not. hasAttribute(dataSetH, "CDELT1")) then
         return
    end if  

    cdelt1 = real64Attribute( dataSetH, "CDELT1")  ! RA plate-scale (degrees and negative)
    cdelt2 = real64Attribute( dataSetH, "CDELT2")  ! Dec plate-scale (degrees)
    crpix1 = real64attribute( dataSetH, "CRPIX1")
    crpix2 = real64attribute( dataSetH, "CRPIX2")
    crval1 = real64attribute( dataSetH, "CRVAL1")
    crval2 = real64attribute( dataSetH, "CRVAL2")
    
   
   
   


  
    do i = 1, nSource
        dx = (dble(source(i)%xpos) - crpix1) * cdelt1
        dy = (dble(source(i)%ypos) - crpix2) * cdelt2
        call convert(dx, dy, crval1, crval2, source(i)%rightAscension, source(i)%declination )


    end do

    dataSetH1 = dataSet( trim(stringParameter( "outset" ) ), modify)

    tableH = table( dataSetH1, "SRCLIST", modify)
 
!    source%xpos = 0.0
!    source%ypos = 0.0

!    call AddReal32Column( tableH, source%xpos,  int(nSource, int32), &
!               "XPOS", "pixels", "", nullValue)
 
    
!    call AddReal32Column( tableH, source%ypos,  int(nSource, int32), &
!               "YPOS", "pixels", "", nullValue)
 




   call AddReal64Column( tableH, source%rightAscension,  int(nSource, int32), &
               "RA", "degrees", "J2000.0 Right Ascension", nullValue)
 
    
    call AddReal64Column( tableH, source%declination,  int(nSource, int32), &
               "DEC", "degrees", "J2000.0 Declination", nullValue)
 
    
    call release(dataSetH1)

 end subroutine setRADECColumns

!********************************************************************
!
!
!
!
!********************************************************************
subroutine convert(x, y, ra0, dec0, ra, dec)

    real(double), intent(in) :: x, y, ra0, dec0
    real(double), intent(out) :: ra, dec


    call sla_dtp2s(x * degreesToRadians, y * degreesToRadians,  &
               ra0 * degreesToRadians, dec0 * degreesToRadians, &
               ra, dec)

    ra = ra * radiansToDegrees
    dec = dec * radiansToDegrees


end subroutine convert
!******************************************************************
!
!
!
!
!*****************************************************************
subroutine loadTheImage( dataSetH, nx, ny, image)
    type(DataSetT) :: dataSetH
    real(single), pointer :: image(:, :)
    integer(int16), intent(out) :: nx, ny
    real(single), pointer :: exposureImage(:, :) => null()
  
    integer :: count1
    integer :: i, j
    real(single), pointer :: counts(:) => null()
  
    if ( associated(image) ) then
        deallocate( image )
        nullify( image ) 
    end if

    call GetImage( dataSetH, nx, ny, image)

    if(ommodeType .eq. "MOSAICED") then 
        
        call GetExposureImage( dataSetH, exposureImage) 

        counts => allocateSingle(1, int(nx, int32) * int(ny, int32), "counts")
      
    
        count1 = 0
      
        do j = 0, ny - 1
            do i = 0, nx - 1
                if( .not. isNullValue(exposureImage(i, j) ) &
                    .and. exposureImage(i, j) .gt. 0.0 ) then
                    count1 = count1 + 1
                    counts(count1) = exposureImage(i, j)
                                       
                end if
            end do
        end do
       
        exposureConstant = MedianValue( counts, count1)

        do j = 0, ny - 1
            do i = 0, nx - 1
                if( .not. isNullValue(image(i, j) )) then
                    image(i, j) = dble(image(i, j)) * dble(exposureConstant)  
 
                end if
            end do
        end do

        
        deallocate(counts, exposureImage)
        nullify(counts, exposureImage)
       

    end if


end subroutine loadTheImage
!******************************************************************************
!
! Find out how many different exposures there are in the exposure image
!
!******************************************************************************
subroutine checkExposureImage(dataSetH)
    type(dataSetT), intent(in) :: dataSetH 
   
    real(single), pointer :: exposureImage(:, :) => null()
    real(single) :: fMin, fMax, percentage
    integer :: numExposures
    integer :: i, j, count1, numGoodPixels
    if(ommodeType .ne. "MOSAICED") return 
  

    write(errstr,'(a)') "Checking exposure image"
    call message(AppMsg, Verbose)
 

    call GetExposureImage( dataSetH, exposureImage) 

    !****************************************************
    !
    ! Count the number of non-null pixels
    !
    !****************************************************

    numGoodPixels = 0
  
    nx = size(exposureImage, 1)
    ny = size(exposureImage, 2)

    do j = lbound(exposureImage, 2), ubound(exposureImage, 2)

        do i = lbound(exposureImage, 1), ubound(exposureImage, 1)
            if (.not. isNullValue(exposureImage(i, j) ) ) then
                numGoodPixels = numGoodPixels + 1 

            end if 
        end do
    end do

    percentage = 100.0 * dble(numGoodPixels) / dble(nx * ny)
    write(errstr,'(a, i0, a, f0.1, a)') "Number of non-null exposure pixels=", numGoodPixels, &
             " percentage of total pixels=", percentage, "%"
    call message(AppMsg, Verbose)
 

    !*********************************************************************
    !
    ! Count the number of different pixel values on the exposure image
    !
    !********************************************************************


    nuMExposures = 0
    fMin = 0.1
    fMax = 1e8


  !  write(*, *) "Line 63415"
    
    do i = 1, 100000
                 
        fMin = minVal(exposureImage, mask=exposureImage > fmin .and. exposureImage <= fMax)
        if(fMin > fMax) exit
      !  write(*, *) i, ") fmin=", fMin            

        count1 = count( abs(exposureImage - fMin) < 0.0001)
        if(count1 > 0) then
            numExposures = numExposures + 1
            percentage = 100.0 * dble(count1) / dble(numGoodPixels)
            write(errstr,'(a, f0.1, a, i0, a, f0.1, a)') "Exposure ", fMin, " number of pixels=", &
                          count1, " percentage of good image pixels=", percentage, "%"


         
            fmin = fmin + 10.0
        else
            exit
        end if 
            
    end do



    deallocate(exposureImage)
    nullify(exposureImage)

    write(errstr,'(a, i0)') "Number of different exposures on exposure image=", numExposures
    call message(AppMsg, Verbose)
 


end subroutine checkExposureImage




!************************************************************************
! Check for hot pixels
!************************************************************************
subroutine checkForHotPixels(image, qualityImage)

    real(single), pointer :: image(:, :)
    logical(boolean), pointer :: qualityImage(:, :)
    integer(int16) :: i, j, k, l, nx, ny
    integer(int8), parameter :: dx = 1
    integer(int16) :: count1, count2
    integer :: numHotPixels
    real(single) :: f0


    nx = ubound(image, 1)
    ny = ubound(image, 2)
    numHotPixels = 0

    do i = 0, nx
        do j = 0, ny 
            if(qualityImage(i, j) ) cycle
            if(image(i, j) < 10.0) cycle
            count1 = 0
            count2 = 0
            f0 = image(i, j) * 0.1  
            do k = max(0_int16, i - dx), min(nx, i + dx)
                do l = max(0_int16, j - dx), min(ny, j + dx)
                     if( k == i .and. l == j) cycle

                     if (.not. isNullValue(image(k, l)) ) then   
                         count1 = count1 + 1                        
                         if(image(k, l) .ge. f0) then
                             count2 = count2 + 1
                         end if
                     end if

                end do
            end do

            
            if(count1 > 0 ) then
                
                if(count2 < 1) then
                    qualityImage(i, j) = .true.
                    numHotPixels = numHotPixels + 1

                end if

            end if
        end do
    end do


!write(*, *) "Checkforhotpixels number found=", numHotPixels
!write(*, *) "num pixels>30000.0=", count(image > 30000.0)


end subroutine checkForHotPixels
!*************************************************************
!
!
!************************************************************
subroutine modifyQualityImage(image, qualityImage) 
    real(single), pointer :: image(:, :)
    logical(boolean), pointer :: qualityImage(:, :)
    integer(int16) :: nx, ny, i, j


    nx = ubound(image, 1)
    ny = ubound(image, 2)
 
    do i = 0, nx
        do j = 0, ny
            if( image(i, j) .lt. 0.0 .or. isNullValue(image(i, j) ) ) then
                qualityImage(i, j) = .true.
            else
                qualityImage(i, j) = .false.
            end if

        end do
    end do

end subroutine modifyQualityImage

end module omdetect_mod 









